@@ -20,22 +20,19 @@
 #include "mutt_curses.h"
 #include "mutt_menu.h"
 #include "buffy.h"
+#include "mapping.h"
 #include "sort.h"
 #include "mailbox.h"
 
 #include <stdlib.h>
 #include <dirent.h>
 #include <string.h>
+#include <ctype.h>
 #include <unistd.h>
 #include <sys/stat.h>
 #include <pwd.h>
 #include <grp.h>
 
-/* HP-UX and ConvexOS don't have this macro */
-#ifndef S_ISLNK
-#define S_ISLNK(x) (((x) & S_IFMT) == S_IFLNK ? 1 : 0)
-#endif
-
 struct folder_file
 {
   mode_t mode;
@@ -53,13 +50,20 @@ struct browser_state
 };
 
 static struct mapping_t FolderHelp[] = {
-  { "Exit",  OP_EXIT },
-  { "Chdir", OP_CHANGE_DIRECTORY },
-  { "Mask",  OP_ENTER_MASK },
-  { "Help",  OP_HELP },
+  { N_("Exit"),  OP_EXIT },
+  { N_("Chdir"), OP_CHANGE_DIRECTORY },
+  { N_("Mask"),  OP_ENTER_MASK },
+  { N_("Help"),  OP_HELP },
   { NULL }
 };
 
+typedef struct folder_t
+{
+  const char *name;
+  const struct stat *f;
+  int new;
+} FOLDER;
+
 static char LastDir[_POSIX_PATH_MAX] = "";
 
 /* Frees up the memory allocated for the local-global variables.  */
@@ -80,7 +84,7 @@ static int browser_compare_subject (const void *a, const void *b)
   struct folder_file *pa = (struct folder_file *) a;
   struct folder_file *pb = (struct folder_file *) b;
 
-  int r = strcmp (pa->name, pb->name);
+  int r = mutt_strcmp (pa->name, pb->name);
 
   return ((BrowserSort & SORT_REVERSE) ? -r : r);
 }
@@ -137,46 +141,159 @@ static int link_is_dir (const char *path)
     return (-1);
 }
 
+static const char *
+folder_format_str (char *dest, size_t destlen, char op, const char *src,
+		   const char *fmt, const char *ifstring, const char *elsestring,
+		   unsigned long data, format_flag flags)
+{
+  char fn[SHORT_STRING], tmp[SHORT_STRING], permission[11];
+  char date[16], *t_fmt;
+  time_t tnow;
+  FOLDER *folder = (FOLDER *) data;
+  struct passwd *pw;
+  struct group *gr;
+
+  switch (op)
+  {
+    case 'd':
+      if (folder->f != NULL)
+      {
+	tnow = time (NULL);
+	t_fmt = tnow - folder->f->st_mtime < 31536000 ? "%b %d %H:%M" : "%b %d  %Y";
+	strftime (date, sizeof (date), t_fmt, localtime (&folder->f->st_mtime));
+	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
+	snprintf (dest, destlen, tmp, date);
+      }
+      else
+      {
+	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
+	snprintf (dest, destlen, tmp, "");
+      }
+      break;
+    case 'f':
+      strfcpy (fn, folder->name, sizeof(fn));
+      if (folder->f != NULL)
+      {
+	strcat (fn, S_ISLNK (folder->f->st_mode) ? "@" : 
+		(S_ISDIR (folder->f->st_mode) ? "/" : 
+		 ((folder->f->st_mode & S_IXUSR) != 0 ? "*" : "")));
+      }
+      snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
+      snprintf (dest, destlen, tmp, fn);
+      break;
+    case 'F':
+      if (folder->f != NULL)
+      {
+	sprintf (permission, "%c%c%c%c%c%c%c%c%c%c",
+		 S_ISDIR(folder->f->st_mode) ? 'd' : (S_ISLNK(folder->f->st_mode) ? 'l' : '-'),
+		 (folder->f->st_mode & S_IRUSR) != 0 ? 'r': '-',
+		 (folder->f->st_mode & S_IWUSR) != 0 ? 'w' : '-',
+		 (folder->f->st_mode & S_ISUID) != 0 ? 's' : (folder->f->st_mode & S_IXUSR) != 0 ? 'x': '-',
+		 (folder->f->st_mode & S_IRGRP) != 0 ? 'r' : '-',
+		 (folder->f->st_mode & S_IWGRP) != 0 ? 'w' : '-',
+		 (folder->f->st_mode & S_ISGID) != 0 ? 's' : (folder->f->st_mode & S_IXGRP) != 0 ? 'x': '-',
+		 (folder->f->st_mode & S_IROTH) != 0 ? 'r' : '-',
+		 (folder->f->st_mode & S_IWOTH) != 0 ? 'w' : '-',
+		 (folder->f->st_mode & S_ISVTX) != 0 ? 't' : (folder->f->st_mode & S_IXOTH) != 0 ? 'x': '-');
+	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
+	snprintf (dest, destlen, tmp, permission);
+      }
+      else
+      {
+#ifdef USE_IMAP
+	if (strchr(folder->name, '{'))
+	{
+	  snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
+	  snprintf (dest, destlen, tmp, "IMAP");
+	}                                        
+#endif
+      }
+      break;
+    case 'g':
+      if (folder->f != NULL)
+      {
+	if ((gr = getgrgid (folder->f->st_gid)))
+	{
+	  snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
+	  snprintf (dest, destlen, tmp, gr->gr_name);
+	}
+	else
+	{
+	  snprintf (tmp, sizeof (tmp), "%%%sld", fmt);
+	  snprintf (dest, destlen, tmp, folder->f->st_gid);
+	}
+      }
+      else
+      {
+	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
+	snprintf (dest, destlen, tmp, "");
+      }
+      break;
+    case 'l':
+      if (folder->f != NULL)
+      {
+	snprintf (tmp, sizeof (tmp), "%%%sd", fmt);
+	snprintf (dest, destlen, tmp, folder->f->st_nlink);
+      }
+      else
+      {
+	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
+	snprintf (dest, destlen, tmp, "");
+      }
+      break;
+    case 'N':
+      snprintf (tmp, sizeof (tmp), "%%%sc", fmt);
+      snprintf (dest, destlen, tmp, folder->new ? 'N' : ' ');
+      break;
+    case 's':
+      if (folder->f != NULL)
+      {
+	snprintf (tmp, sizeof (tmp), "%%%sld", fmt);
+	snprintf (dest, destlen, tmp, (long) folder->f->st_size);
+      }
+      else
+      {
+	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
+	snprintf (dest, destlen, tmp, "");
+      }
+      break;
+    case 'u':
+      if (folder->f != NULL)
+      {
+	if ((pw = getpwuid (folder->f->st_uid)))
+	{
+	  snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
+	  snprintf (dest, destlen, tmp, pw->pw_name);
+	}
+	else
+	{
+	  snprintf (tmp, sizeof (tmp), "%%%sld", fmt);
+	  snprintf (dest, destlen, tmp, folder->f->st_uid);
+	}
+      }
+      else
+      {
+	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
+	snprintf (dest, destlen, tmp, "");
+      }
+      break;
+  }
+  return (src);
+}
+
 static void add_folder (MUTTMENU *m, struct browser_state *state,
 			const char *name, const struct stat *s, int new)
 {
   char buffer[_POSIX_PATH_MAX + SHORT_STRING];
-  struct passwd *pw;
-  struct group *gr;
-  char date[16];
-  char *fmt;
-  time_t tnow;
+  FOLDER folder;
+
+  folder.name = name;
+  folder.f = s;
+  folder.new = new;
+  mutt_FormatString (buffer, sizeof (buffer), NONULL(FolderFormat),
+		     folder_format_str, (unsigned long) &folder,
+		     M_FORMAT_ARROWCURSOR);
 
-  pw = getpwuid (s->st_uid);
-  gr = getgrgid (s->st_gid);
-  
-  tnow = time (NULL);
-  fmt = tnow - s->st_mtime < 31536000 ? "%b %d %H:%M" : "%b %d  %Y";
-  strftime (date, sizeof (date), fmt, localtime (&s->st_mtime));
-  
-  snprintf (buffer, sizeof (buffer), "%c %c%c%c%c%c%c%c%c%c%c %2d %-8.8s %-8.8s %8ld %s %s%s",
- 	    new ? 'N' : ' ',
- 	    S_ISDIR(s->st_mode) ? 'd' : (S_ISLNK(s->st_mode) ? 'l' : '-'),
- 	    
- 	    (s->st_mode & S_IRUSR) != 0 ? 'r': '-',
- 	    (s->st_mode & S_IWUSR) != 0 ? 'w' : '-',
- 	    (s->st_mode & S_ISUID) != 0 ? 's' : (s->st_mode & S_IXUSR) != 0 ? 'x': '-',
- 	    
- 	    (s->st_mode & S_IRGRP) != 0 ? 'r' : '-',
- 	    (s->st_mode & S_IWGRP) != 0 ? 'w' : '-',
- 	    (s->st_mode & S_ISGID) != 0 ? 's' : (s->st_mode & S_IXGRP) != 0 ? 'x': '-',
- 	    
- 	    (s->st_mode & S_IROTH) != 0 ? 'r' : '-',
- 	    (s->st_mode & S_IWOTH) != 0 ? 'w' : '-',
- 	    (s->st_mode & S_ISVTX) != 0 ? 't' : (s->st_mode & S_IXOTH) != 0 ? 'x': '-',
- 	    s->st_nlink,
- 	    pw ? pw->pw_name : "(null)",
- 	    gr ? gr->gr_name : "(null)",
- 	    (long) s->st_size,
- 	    date,
- 	    name,
- 	    S_ISLNK (s->st_mode) ? "@" : (S_ISDIR (s->st_mode) ? "/" : ((s->st_mode & S_IXUSR) != 0 ? "*" : "")));
-  
   if (state->entrylen == state->entrymax)
   {
     /* need to allocate more space */
@@ -186,9 +303,12 @@ static void add_folder (MUTTMENU *m, struct browser_state *state,
       m->data = state->entry;
   }
 
+  if (s != NULL)
+  {
   (state->entry)[state->entrylen].mode = s->st_mode;
   (state->entry)[state->entrylen].mtime = s->st_mtime;
   (state->entry)[state->entrylen].size = s->st_size;
+  }
   (state->entry)[state->entrylen].name = safe_strdup (name);
   (state->entry)[state->entrylen].desc = safe_strdup (buffer);
 
@@ -221,7 +341,7 @@ static int examine_directory (MUTTMENU *menu, struct browser_state *state,
 
   if (!S_ISDIR (s.st_mode))
   {
-    mutt_error ("%s is not a directory", d);
+    mutt_error (_("%s is not a directory."), d);
     return (-1);
   }
 
@@ -237,12 +357,12 @@ static int examine_directory (MUTTMENU *menu, struct browser_state *state,
 
   while ((de = readdir (dp)) != NULL)
   {
-    if (strcmp (de->d_name, ".") == 0)
+    if (mutt_strcmp (de->d_name, ".") == 0)
       continue;    /* we don't need . */
     
-    if (prefix && *prefix && strncmp (prefix, de->d_name, strlen (prefix)) != 0)
+    if (prefix && *prefix && mutt_strncmp (prefix, de->d_name, mutt_strlen (prefix)) != 0)
       continue;
-    if (regexec (Mask.rx, de->d_name, 0, NULL, 0) != 0)
+    if (!((regexec (Mask.rx, de->d_name, 0, NULL, 0) == 0) ^ Mask.not))
       continue;
 
     snprintf (buffer, sizeof (buffer), "%s/%s", d, de->d_name);
@@ -254,7 +374,7 @@ static int examine_directory (MUTTMENU *menu, struct browser_state *state,
       continue;
     
     tmp = Incoming;
-    while (tmp && strcmp (buffer, tmp->path))
+    while (tmp && mutt_strcmp (buffer, tmp->path))
       tmp = tmp->next;
     add_folder (menu, state, de->d_name, &s, (tmp) ? tmp->new : 0);
   }
@@ -277,6 +397,13 @@ static int examine_mailboxes (MUTTMENU *menu, struct browser_state *state)
 
   do
   {
+#ifdef USE_IMAP
+    if (tmp->path[0] == '{')
+    {
+      add_folder (menu, state, tmp->path, NULL, tmp->new);
+      continue;
+    }
+#endif
     if (lstat (tmp->path, &s) == -1)
       continue;
 
@@ -284,7 +411,7 @@ static int examine_mailboxes (MUTTMENU *menu, struct browser_state *state)
 	(! S_ISLNK (s.st_mode)))
       continue;
     
-    strfcpy (buffer, tmp->path, sizeof (buffer));
+    strfcpy (buffer, NONULL(tmp->path), sizeof (buffer));
     mutt_pretty_mailbox (buffer);
 
     add_folder (menu, state, buffer, &s, tmp->new);
@@ -309,16 +436,22 @@ static void init_menu (struct browser_state *state, MUTTMENU *menu, char *title,
 {
   char path[_POSIX_PATH_MAX];
 
-  menu->current = 0;
-  menu->top = 0;
   menu->max = state->entrylen;
+
+  if(menu->current >= menu->max)
+    menu->current = menu->max - 1;
+  if (menu->current < 0)
+    menu->current = 0;
+  if (menu->top > menu->current)
+    menu->top = 0;
+
   if (buffy)
-    snprintf (title, titlelen, "Mailboxes [%d]", mutt_buffy_check (0));
+    snprintf (title, titlelen, _("Mailboxes [%d]"), mutt_buffy_check (0));
   else
   {
     strfcpy (path, LastDir, sizeof (path));
     mutt_pretty_mailbox (path);
-    snprintf (title, titlelen, "Directory [%s], File mask: %s",
+    snprintf (title, titlelen, _("Directory [%s], File mask: %s"),
 	      path, Mask.pattern);
   }
   menu->redraw = REDRAW_FULL;
@@ -326,7 +459,7 @@ static void init_menu (struct browser_state *state, MUTTMENU *menu, char *title,
 
 void mutt_select_file (char *f, size_t flen, int buffy)
 {
-  char buf[STRING];
+  char buf[_POSIX_PATH_MAX];
   char prefix[_POSIX_PATH_MAX] = "";
   char helpstr[SHORT_STRING];
   char title[STRING];
@@ -340,7 +473,7 @@ void mutt_select_file (char *f, size_t flen, int buffy)
   if (*f)
   {
     mutt_expand_path (f, flen);
-    for (i = strlen (f) - 1; i > 0 && f[i] != '/' ; i--);
+    for (i = mutt_strlen (f) - 1; i > 0 && f[i] != '/' ; i--);
     if (i > 0)
     {
       if (f[0] == '/')
@@ -371,7 +504,7 @@ void mutt_select_file (char *f, size_t flen, int buffy)
     killPrefix = 1;
   }
   else if (!LastDir[0])
-    strfcpy (LastDir, Maildir, sizeof (LastDir));
+    strfcpy (LastDir, NONULL(Maildir), sizeof (LastDir));
 
   *f = 0;
 
@@ -402,7 +535,7 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 
 	if (!state.entrylen)
 	{
-	  mutt_error ("No files match the file mask");
+	  mutt_error _("No files match the file mask");
 	  break;
 	}
 
@@ -427,9 +560,9 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	    /* save the old directory */
 	    strfcpy (OldLastDir, LastDir, sizeof (OldLastDir));
 
-	    if (strcmp (state.entry[menu->current].name, "..") == 0)
+	    if (mutt_strcmp (state.entry[menu->current].name, "..") == 0)
 	    {
-	      if (strcmp ("..", LastDir + strlen (LastDir) - 2) == 0)
+	      if (mutt_strcmp ("..", LastDir + mutt_strlen (LastDir) - 2) == 0)
 		strcat (LastDir, "/..");
 	      else
 	      {
@@ -452,7 +585,7 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	      mutt_expand_path (LastDir, sizeof (LastDir));
 	    }
 	    else
-	      sprintf (LastDir + strlen (LastDir), "/%s",
+	      sprintf (LastDir + mutt_strlen (LastDir), "/%s",
 		       state.entry[menu->current].name);
 
 	    destroy_state (&state);
@@ -468,10 +601,12 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	      strfcpy (LastDir, OldLastDir, sizeof (LastDir));
 	      if (examine_directory (menu, &state, LastDir, prefix) == -1)
 	      {
-		strfcpy (LastDir, Homedir, sizeof (LastDir));
+		strfcpy (LastDir, NONULL(Homedir), sizeof (LastDir));
 		return;
 	      }
 	    }
+	    menu->current = 0; 
+	    menu->top = 0; 
 	    init_menu (&state, menu, title, sizeof (title), buffy);
 	    break;
 	  }
@@ -493,10 +628,21 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	mutt_menuDestroy (&menu);
 	return;
 
+      case OP_BROWSER_TELL:
+        if(state.entrylen)
+	  mutt_message(state.entry[menu->current].name);
+        break;
+      
       case OP_CHANGE_DIRECTORY:
 
 	strfcpy (buf, LastDir, sizeof (buf));
-	if (mutt_get_field ("Chdir to: ", buf, sizeof (buf), M_FILE) == 0 &&
+	{/* add '/' at the end of the directory name */
+	int len=mutt_strlen(LastDir);
+	if (sizeof (buf) > len)
+	  buf[len]='/';
+	}
+
+	if (mutt_get_field (_("Chdir to: "), buf, sizeof (buf), M_FILE) == 0 &&
 	    buf[0])
 	{
 	  buffy = 0;	  
@@ -508,17 +654,21 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	      strfcpy (LastDir, buf, sizeof (LastDir));
 	      destroy_state (&state);
 	      if (examine_directory (menu, &state, LastDir, prefix) == 0)
+	      {
+		menu->current = 0; 
+		menu->top = 0; 
 		init_menu (&state, menu, title, sizeof (title), buffy);
+	      }
 	      else
 	      {
-		mutt_error ("Error scanning directory.");
+		mutt_error _("Error scanning directory.");
 		destroy_state (&state);
 		mutt_menuDestroy (&menu);
 		return;
 	      }
 	    }
 	    else
-	      mutt_error ("%s is not a directory.", buf);
+	      mutt_error (_("%s is not a directory."), buf);
 	  }
 	  else
 	    mutt_perror (buf);
@@ -529,17 +679,25 @@ void mutt_select_file (char *f, size_t flen, int buffy)
       case OP_ENTER_MASK:
 
 	strfcpy (buf, Mask.pattern, sizeof (buf));
-	if (mutt_get_field ("File Mask: ", buf, sizeof (buf), 0) == 0)
+	if (mutt_get_field (_("File Mask: "), buf, sizeof (buf), 0) == 0)
 	{
 	  regex_t *rx = (regex_t *) safe_malloc (sizeof (regex_t));
-	  int err;
+	  char *s = buf;
+	  int not = 0, err;
 
 	  buffy = 0;
 	  /* assume that the user wants to see everything */
 	  if (!buf[0])
 	    strfcpy (buf, ".", sizeof (buf));
+	  SKIPWS (s);
+	  if (*s == '!')
+	  {
+	    s++;
+	    SKIPWS (s);
+	    not = 1;
+	  }
 
-	  if ((err = REGCOMP (rx, buf, REG_NOSUB | mutt_which_case (buf))) != 0)
+	  if ((err = REGCOMP (rx, s, REG_NOSUB)) != 0)
 	  {
 	    regerror (err, rx, buf, sizeof (buf));
 	    regfree (rx);
@@ -553,17 +711,23 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	    safe_free ((void **) &Mask.rx);
 	    Mask.pattern = safe_strdup (buf);
 	    Mask.rx = rx;
+	    Mask.not = not;
 
 	    destroy_state (&state);
 	    if (examine_directory (menu, &state, LastDir, NULL) == 0)
 	      init_menu (&state, menu, title, sizeof (title), buffy);
 	    else
 	    {
-	      mutt_error ("Error scanning directory.");
+	      mutt_error _("Error scanning directory.");
 	      mutt_menuDestroy (&menu);
 	      return;
 	    }
 	    killPrefix = 0;
+	    if (!state.entrylen)
+	    {
+	      mutt_error _("No files match the file mask");
+	      break;
+	    }
 	  }
 	}
 	MAYBE_REDRAW (menu->redraw);
@@ -572,54 +736,49 @@ void mutt_select_file (char *f, size_t flen, int buffy)
       case OP_SORT:
       case OP_SORT_REVERSE:
 
-	{
-	  int reverse = 0;
-
-	  move (LINES - 1, 0);
-	  if (i == OP_SORT_REVERSE)
+        {
+	  int resort = 1;
+	  int reverse = (i == OP_SORT_REVERSE);
+	  
+	  switch (mutt_multi_choice ((reverse) ?
+	      _("Reverse sort by (d)ate, (a)lpha, si(z)e or do(n)'t sort? ") :
+	      _("Sort by (d)ate, (a)lpha, si(z)e or do(n)'t sort? "),
+	      _("dazn")))
 	  {
-	    reverse = SORT_REVERSE;
-	    addstr ("Reverse ");
-	  }
-	  addstr ("Sort by (d)ate, (a)lpha, si(z)e or do(n)'t sort? ");
-	  clrtoeol ();
-
-	  while ((i = mutt_getch ()) != EOF && i != 'a' && i != 'd' && i != 'z'
-		 && i != 'n')
-	  {
-	    if (i == ERR || CI_is_return (i))
+	    case -1: /* abort */
+	      resort = 0;
 	      break;
-	    else
-	      BEEP ();
-	  }
 
-	  if (i != EOF)
+            case 1: /* (d)ate */
+	      BrowserSort = SORT_DATE;
+	      break;
+
+            case 2: /* (a)lpha */
+	      BrowserSort = SORT_SUBJECT;
+	      break;
+
+            case 3: /* si(z)e */
+	      BrowserSort = SORT_SIZE;
+	      break;
+
+            case 4: /* do(n)'t sort */
+	      BrowserSort = SORT_ORDER;
+	      resort = 0;
+	      break;
+	  }
+	  if (resort)
 	  {
-	    switch (i)
-	    {
-	      case 'a': 
-	        BrowserSort = reverse | SORT_SUBJECT;
-		break;
-	      case 'd':
-	        BrowserSort = reverse | SORT_DATE;
-		break;
-	      case 'z': 
-	        BrowserSort = reverse | SORT_SIZE;
-		break;
-	      case 'n': 
-	        BrowserSort = SORT_ORDER;
-		break;
-	    }
+	    BrowserSort |= reverse ? SORT_REVERSE : 0;
 	    browser_sort (&state);
 	    menu->redraw = REDRAW_FULL;
 	  }
+	  break;
 	}
 
-	break;
+      case OP_TOGGLE_MAILBOXES:
+	buffy = 1 - buffy;
 
       case OP_CHECK_NEW:
-
-	buffy = 1 - buffy;
 	destroy_state (&state);
 	prefix[0] = 0;
 	killPrefix = 0;
@@ -636,7 +795,7 @@ void mutt_select_file (char *f, size_t flen, int buffy)
       case OP_BROWSER_NEW_FILE:
 
 	snprintf (buf, sizeof (buf), "%s/", LastDir);
-	if (mutt_get_field ("New file name: ", buf, sizeof (buf), M_FILE) == 0)
+	if (mutt_get_field (_("New file name: "), buf, sizeof (buf), M_FILE) == 0)
 	{
 	  strfcpy (f, buf, flen);
 	  destroy_state (&state);
@@ -645,6 +804,38 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	}
 	MAYBE_REDRAW (menu->redraw);
 	break;
+
+      case OP_BROWSER_VIEW_FILE:
+	if (!state.entrylen)
+	{
+	  mutt_error _("No files match the file mask");
+	  break;
+	}
+
+        if (S_ISDIR (state.entry[menu->current].mode) ||
+	    (S_ISLNK (state.entry[menu->current].mode) &&
+	    link_is_dir (state.entry[menu->current].name)))
+	{
+	  mutt_error _("Can't view a directory");
+	  break;
+	} 
+	else
+	{
+	  BODY *b;
+	  char buf[_POSIX_PATH_MAX];
+
+	  snprintf (buf, sizeof (buf), "%s/%s", LastDir,
+		    state.entry[menu->current].name);
+	  b = mutt_make_file_attach (buf);
+	  if (b != NULL)
+	  {
+	    mutt_view_attachment (NULL, b, M_REGULAR);
+	    mutt_free_body (&b);
+	    menu->redraw = REDRAW_FULL;
+	  }
+	  else
+	    mutt_error _("Error trying to view file");
+	}
     }
   }
   /* not reached */
