@@ -20,11 +20,15 @@
 #include "mutt_curses.h"
 #include "mutt_menu.h"
 #include "mailbox.h"
+#include "mapping.h"
 #include "sort.h"
 #include "buffy.h"
 
 
 
+#ifdef _PGPPATH
+#include "pgp.h"
+#endif
 
 
 
@@ -46,32 +50,40 @@
 #define CHECK_MSGCOUNT if (!Context) \
 	{ \
 	  	mutt_flushinp (); \
-		mutt_error ("No mailbox is open."); \
+		mutt_error _("No mailbox is open."); \
 		break; \
 	} \
 	else if (!Context->msgcount) \
 	{ \
 	  	mutt_flushinp (); \
-		mutt_error ("There are no messages."); \
+		mutt_error _("There are no messages."); \
 		break; \
 	}
 
 #define CHECK_READONLY if (Context->readonly) \
 			{ \
 			  	mutt_flushinp (); \
-				mutt_error ("Mailbox is read-only."); \
+				mutt_error _("Mailbox is read-only."); \
 				break; \
 			}
 
+#define CHECK_ATTACH if(option(OPTATTACHMSG)) \
+		     {\
+			mutt_flushinp (); \
+			mutt_error ("Function not permitted in attach-message mode."); \
+			break; \
+		     }
+
 #define CURHDR Context->hdrs[Context->v2r[menu->current]]
 #define OLDHDR Context->hdrs[Context->v2r[menu->oldcurrent]]
+#define UNREAD(h) mutt_thread_contains_unread (Context, h)
 
 extern const char *ReleaseDate;
 
 void index_make_entry (char *s, size_t l, MUTTMENU *menu, int num)
 {
-  format_flag flag = M_FORMAT_MAKEPRINT;
-  int reverse = Sort & SORT_REVERSE, edgemsgno;
+  format_flag flag = M_FORMAT_MAKEPRINT | M_FORMAT_ARROWCURSOR | M_FORMAT_INDEX;
+  int edgemsgno, reverse = Sort & SORT_REVERSE;
   HEADER *tmp, *h = Context->hdrs[Context->v2r[num]];
 
   if ((Sort & SORT_MASK) == SORT_THREADS && h->tree)
@@ -84,18 +96,19 @@ void index_make_entry (char *s, size_t l, MUTTMENU *menu, int num)
     {
       if (reverse)
       {
-       if (menu->top + menu->pagelen > menu->max)
-         edgemsgno = Context->v2r[menu->max - 1];
-       else
-         edgemsgno = Context->v2r[menu->top + menu->pagelen - 1];
+	if (menu->top + menu->pagelen > menu->max)
+	  edgemsgno = Context->v2r[menu->max - 1];
+	else
+	  edgemsgno = Context->v2r[menu->top + menu->pagelen - 1];
       }
       else
-       edgemsgno = Context->v2r[menu->top];
+	edgemsgno = Context->v2r[menu->top];
 
       for (tmp = h->parent; tmp; tmp = tmp->parent)
       {
-	if ((reverse && tmp->msgno > edgemsgno)
-	    || (!reverse && tmp->msgno < edgemsgno))
+	/* if no ancestor is visible on current screen, provisionally force
+	 * subject... */
+	if (reverse ? tmp->msgno > edgemsgno : tmp->msgno < edgemsgno)
 	{
 	  flag |= M_FORMAT_FORCESUBJ;
 	  break;
@@ -107,8 +120,8 @@ void index_make_entry (char *s, size_t l, MUTTMENU *menu, int num)
       {
 	for (tmp = h->prev; tmp; tmp = tmp->prev)
 	{
-	  if ((reverse && tmp->msgno > edgemsgno)
-	      || (!reverse && tmp->msgno < edgemsgno))
+	  /* ...but if a previous sibling is available, don't force it */
+	  if (reverse ? tmp->msgno > edgemsgno : tmp->msgno < edgemsgno)
 	    break;
 	  else if (tmp->virtual >= 0)
 	  {
@@ -125,7 +138,13 @@ void index_make_entry (char *s, size_t l, MUTTMENU *menu, int num)
 
 int index_color (int index_no)
 {
-  return Context->hdrs[Context->v2r[index_no]]->pair;
+  HEADER *h = Context->hdrs[Context->v2r[index_no]];
+
+  if (h->pair)
+    return h->pair;
+
+  mutt_set_header_color (Context, h);
+  return h->pair;
 }
 
 static int ci_next_undeleted (int msgno)
@@ -193,43 +212,61 @@ static int mx_toggle_write (CONTEXT *ctx)
 
   if (ctx->readonly)
   {
-    mutt_error ("Cannot toggle write on a readonly mailbox!");
+    mutt_error _("Cannot toggle write on a readonly mailbox!");
     return -1;
   }
 
   if (ctx->dontwrite)
   {
     ctx->dontwrite = 0;
-    mutt_message ("Changes to folder will be written on folder exit.");
+    mutt_message _("Changes to folder will be written on folder exit.");
   }
   else
   {
     ctx->dontwrite = 1;
-    mutt_message ("Changes to folder will not be written.");
+    mutt_message _("Changes to folder will not be written.");
   }
 
   return 0;
 }
 
+static void resort_index (MUTTMENU *menu)
+{
+  int i;
+  const HEADER *current = CURHDR;
+
+  mutt_sort_headers (Context, 0);
+  /* Restore the current message */
+  for (i = 0; i < Context->vcount; i++)
+  {
+    if (Context->hdrs[Context->v2r[i]] == current)
+    {
+      menu->current = i;
+      break;
+    }
+  }
+  menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
+}
+
 struct mapping_t IndexHelp[] = {
-  { "Quit",  OP_QUIT },
-  { "Del",   OP_DELETE },
-  { "Undel", OP_UNDELETE },
-  { "Save",  OP_SAVE },
-  { "Mail",  OP_MAIL },
-  { "Reply", OP_REPLY },
-  { "Group", OP_GROUP_REPLY },
-  { "Help",  OP_HELP },
+  { N_("Quit"),  OP_QUIT },
+  { N_("Del"),   OP_DELETE },
+  { N_("Undel"), OP_UNDELETE },
+  { N_("Save"),  OP_SAVE },
+  { N_("Mail"),  OP_MAIL },
+  { N_("Reply"), OP_REPLY },
+  { N_("Group"), OP_GROUP_REPLY },
+  { N_("Help"),  OP_HELP },
   { NULL }
 };
 
 /* This function handles the message index window as well as commands returned
  * from the pager (MENU_PAGER).
  */
-void mutt_index_menu (void)
+int mutt_index_menu (void)
 {
   char buf[LONG_STRING], helpstr[SHORT_STRING];
-  int op = OP_NULL;                 /* function to execute */
+  int op = OP_NULL;
   int done = 0;                /* controls when to exit the "event" loop */
   int i = 0, j;
   int tag = 0;                 /* has the tag-prefix command been pressed? */
@@ -240,6 +277,8 @@ void mutt_index_menu (void)
   char *cp;                    /* temporary variable. */
   int index_hint;   /* used to restore cursor position */
   int do_buffy_notify = 1;
+  int close = 0; /* did we OP_QUIT or OP_EXIT out of this menu? */
+  int attach_msg = option(OPTATTACHMSG);
 
   menu = mutt_new_menu ();
   menu->menu = MENU_MAIN;
@@ -250,7 +289,8 @@ void mutt_index_menu (void)
   menu->current = ci_first_message ();
   menu->help = mutt_compile_help (helpstr, sizeof (helpstr), MENU_MAIN, IndexHelp);
   
-  mutt_buffy_check(1); /* force the buffy check after we enter the folder */
+  if (!attach_msg) 
+    mutt_buffy_check(1); /* force the buffy check after we enter the folder */
 
   FOREVER
   {
@@ -259,15 +299,16 @@ void mutt_index_menu (void)
     menu->max = Context ? Context->vcount : 0;
     oldcount = Context ? Context->msgcount : 0;
 
-    if (Context)
+    if (Context && !attach_msg)
     {
+      int check;
       /* check for new mail in the mailbox.  If nonzero, then something has
        * changed about the file (either we got new mail or the file was
        * modified underneath us.)
        */
       index_hint = (Context->vcount) ? CURHDR->index : 0;
 
-      if ((i = mx_check_mailbox (Context, &index_hint)) < 0)
+      if ((check = mx_check_mailbox (Context, &index_hint)) < 0)
       {
 	if (!Context->path)
 	{
@@ -278,8 +319,11 @@ void mutt_index_menu (void)
 
 	set_option (OPTSEARCHINVALID);
       }
-      else if (i == M_NEW_MAIL ||  i == M_REOPENED)
+      else if (check == M_NEW_MAIL || check == M_REOPENED)
       {
+	/* store pointers to the newly added messages */
+	HEADER  **save_new = NULL;
+
 	/* take note of the current message */
 	if (oldcount)
 	{
@@ -294,32 +338,73 @@ void mutt_index_menu (void)
 	 * they will be visible in the limited view */
         if (Context->pattern)
 	{
-          #define this_body Context->hdrs[i]->content
-	  if (oldcount)
-          for (i = oldcount; i < Context->msgcount; i++)
+#define THIS_BODY Context->hdrs[j]->content
+	  if (oldcount || check == M_REOPENED)
+	    for (j = (check == M_REOPENED) ? 0 : oldcount; j < Context->msgcount; j++)
 	  {
             if (mutt_pattern_exec (Context->limit_pattern,
 	                           M_MATCH_FULL_ADDRESS, 
-	                           Context, Context->hdrs[i]))
+				    Context, Context->hdrs[j]))
             {
-	     Context->hdrs[i]->virtual = Context->vcount;
-	     Context->v2r[Context->vcount] = i;
-	     Context->vcount++;
-	     Context->vsize+=this_body->length + this_body->offset -
-                             this_body->hdr_offset;
+		Context->hdrs[j]->virtual = Context->vcount;
+		Context->v2r[Context->vcount] = j;
+		Context->hdrs[j]->limited = 1;
+		Context->vcount++;
+		Context->vsize += THIS_BODY->length + THIS_BODY->offset - THIS_BODY->hdr_offset;
             }
 	  }
-          #undef this_body
+#undef THIS_BODY
+        }
+
+	/* save the list of new messages */
+	if (oldcount && check != M_REOPENED)
+	{
+	  save_new = (HEADER **) safe_malloc (sizeof (HEADER *) * (Context->msgcount - oldcount));
+	  for (j = oldcount; j < Context->msgcount; j++)
+	    save_new[j-oldcount] = Context->hdrs[j];
         }
 
 	/* if the mailbox was reopened, need to rethread from scratch */
-	mutt_sort_headers (Context, (i == M_REOPENED));
+	mutt_sort_headers (Context, (check == M_REOPENED));
+
+	/* uncollapse threads with new mail */
+	if ((Sort & SORT_MASK) == SORT_THREADS)
+	{
+	  if (check == M_REOPENED)
+	  {
+	    HEADER *h;
+
+	    h = Context->tree;
+	    Context->collapsed = 0;
+
+	    while (h)
+	    {
+	      mutt_uncollapse_thread (Context, h);
+	      h = h->next;
+	    }
+	    mutt_set_virtual (Context);
+	  }
+	  else if (oldcount)
+	  {
+	    for (j = 0; j < Context->msgcount - oldcount; j++)
+	    {
+	      int k;
+	      
+	      for (k = 0; k < Context->msgcount; k++)
+	      {
+		HEADER *h = Context->hdrs[k];
+		if (h == save_new[j] && (!Context->pattern || h->limited))
+		  mutt_uncollapse_thread (Context, h);
+	      }
+	    }
+	    FREE (&save_new);
+	    mutt_set_virtual (Context);
+	  }
+	}
 
 	menu->current = -1;
 	if (oldcount)
 	{
-	  int j;
-
 	  /* restore the current message to the message it was pointing to */
 	  for (j = 0; j < Context->vcount; j++)
 	    if (Context->hdrs[Context->v2r[j]]->index == menu->oldcurrent)
@@ -333,11 +418,11 @@ void mutt_index_menu (void)
 	  menu->current = ci_first_message ();
 
 	/* notify the user of new mail */
-	if (i == M_REOPENED)
-	  mutt_error ("Mailbox was externally modified.  Flags may be wrong.");
+	if (check == M_REOPENED)
+	  mutt_error _("Mailbox was externally modified.  Flags may be wrong.");
 	else
 	{
-	  mutt_message ("New mail in this mailbox.");
+	  mutt_message _("New mail in this mailbox.");
 	  if (option (OPTBEEPNEW))
 	    beep ();
 	}
@@ -351,17 +436,20 @@ void mutt_index_menu (void)
       }
     }
 
-    /* check for new mail in the incoming folders */
-    oldcount = newcount;
-    if ((newcount = mutt_buffy_check (0)) != oldcount)
-      menu->redraw |= REDRAW_STATUS;
-    if (do_buffy_notify)
+    if (!attach_msg)
     {
-      if (mutt_buffy_notify () && option (OPTBEEPNEW))
-	beep ();
+     /* check for new mail in the incoming folders */
+     oldcount = newcount;
+     if ((newcount = mutt_buffy_check (0)) != oldcount)
+       menu->redraw |= REDRAW_STATUS;
+     if (do_buffy_notify)
+     {
+       if (mutt_buffy_notify () && option (OPTBEEPNEW))
+ 	beep ();
+     }
+     else
+       do_buffy_notify = 1;
     }
-    else
-      do_buffy_notify = 1;
 
     mutt_curs_set (0);
 
@@ -407,21 +495,12 @@ void mutt_index_menu (void)
 	move (menu->current - menu->top + menu->offset, COLS - 1);
       mutt_refresh ();
 
-      if (Timeout > 0)
-      {
-	timeout (Timeout * 1000); /* milliseconds */      
-	op = mutt_getch ();
-	timeout (-1); /* restore blocking operation */
-	if (op != -1)
-	{
-	  mutt_ungetch (op);
-	  op = km_dokey (MENU_MAIN);
-	}
-      }
-      else
-	op = km_dokey (MENU_MAIN);
-      mutt_curs_set (1);
+      op = km_dokey (MENU_MAIN);
 
+      dprint(4, (debugfile, "mutt_index_menu[%d]: Got op %d\n", __LINE__, op));
+
+      mutt_curs_set (1);
+      
 #if defined (USE_SLANG_CURSES) || defined (HAVE_RESIZETERM)
       if (Signals & S_SIGWINCH)
       {
@@ -443,13 +522,13 @@ void mutt_index_menu (void)
       {
 	if (!Context)
 	{
-	  mutt_error ("No mailbox is open.");
+	  mutt_error _("No mailbox is open.");
 	  continue;
 	}
 
 	if (!Context->tagged)
 	{
-	  mutt_error ("No tagged messages.");
+	  mutt_error _("No tagged messages.");
 	  continue;
 	}
 	tag = 1;
@@ -527,15 +606,15 @@ void mutt_index_menu (void)
       case OP_JUMP:
 
 	CHECK_MSGCOUNT;
-	mutt_ungetch (LastKey);
+        if (isdigit (LastKey)) mutt_ungetch (LastKey, 0);
 	buf[0] = 0;
-	if (mutt_get_field ("Jump to message: ", buf, sizeof (buf), 0) != 0 ||
-	    !buf[0])
+	if (mutt_get_field (_("Jump to message: "), buf, sizeof (buf), 0) != 0
+	    || !buf[0])
 	  break;
 
 	if (! isdigit ((unsigned char) buf[0]))
 	{
-	  mutt_error ("Argument must be a message number.");
+	  mutt_error _("Argument must be a message number.");
 	  break;
 	}
 
@@ -568,10 +647,10 @@ void mutt_index_menu (void)
 	    menu->redraw = REDRAW_MOTION;
 	  }
 	  else
-	    mutt_error ("That message is not visible.");
+	    mutt_error _("That message is not visible.");
 	}
 	else
-	  mutt_error ("Invalid message number.");
+	  mutt_error _("Invalid message number.");
 
 	break;
 
@@ -583,13 +662,15 @@ void mutt_index_menu (void)
 
 	CHECK_MSGCOUNT;
 	CHECK_READONLY;
-	mutt_pattern_func (M_DELETE, "Delete messages matching: ", CURHDR);
+	CHECK_ATTACH;
+	mutt_pattern_func (M_DELETE, _("Delete messages matching: "));
 	menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
 	break;
 
 #ifdef USE_POP
       case OP_MAIN_FETCH_MAIL:
 
+	CHECK_ATTACH;
 	mutt_fetchPopMail ();
 	menu->redraw = REDRAW_FULL;
 	break;
@@ -604,11 +685,12 @@ void mutt_index_menu (void)
       case OP_MAIN_SHOW_LIMIT:
         CHECK_MSGCOUNT;
 	if (!Context->pattern)
-	   mutt_message ("No limit pattern is in effect.");
+	   mutt_message _("No limit pattern is in effect.");
 	else
 	{
 	   char buf[STRING];
-	   snprintf (buf, sizeof(buf), "Limit: %s",Context->pattern);
+	   /* i18n: ask for a limit to apply */
+	   snprintf (buf, sizeof(buf), _("Limit: %s"),Context->pattern);
            mutt_message ("%s", buf);
 	}
         break;
@@ -617,7 +699,7 @@ void mutt_index_menu (void)
 
 	CHECK_MSGCOUNT;
 	menu->oldcurrent = Context->vcount ? CURHDR->index : -1;
-	if (mutt_pattern_func (M_LIMIT, "Limit to messages matching: ", CURHDR) == 0)
+	if (mutt_pattern_func (M_LIMIT, _("Limit to messages matching: ")) == 0)
 	{
 	  if (menu->oldcurrent >= 0)
 	  {
@@ -640,7 +722,14 @@ void mutt_index_menu (void)
 
       case OP_QUIT:
 
-	if (query_quadoption (OPT_QUIT, "Quit Mutt?") == M_YES)
+	close = op;
+	if (attach_msg)
+	{
+	 done = 1;
+	 break;
+	}
+
+	if (query_quadoption (OPT_QUIT, _("Quit Mutt?")) == M_YES)
 	{ 
 	  if (!Context || mx_close_mailbox (Context) == 0)
 	    done = 1;
@@ -674,16 +763,7 @@ void mutt_index_menu (void)
 	{
 	  if (Context && Context->msgcount)
 	  {
-	    menu->oldcurrent = CURHDR->index;
-	    mutt_sort_headers (Context, 0);
-
-	    /* try to restore the current message */
-	    for (i = 0; i < Context->vcount; i++)
-	    {
-	      if (Context->hdrs[Context->v2r[i]]->index == menu->oldcurrent)
-		menu->current = i;
-	    }
-	    menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
+	    resort_index (menu);
 	    set_option (OPTSEARCHINVALID);
 	  }
 	}
@@ -708,13 +788,15 @@ void mutt_index_menu (void)
 	  if (mx_sync_mailbox (Context) == 0)
 	  {
 	    if (Context->vcount != oldvcount)
-	    {
 	      menu->current -= dcount;
-	      if (menu->current < 0 || menu->current >= Context->vcount)
-		menu->current = ci_first_message ();
-	    }
 	    set_option (OPTSEARCHINVALID);
 	  }
+	  
+	  /* do a sanity check even if mx_sync_mailbox failed.
+	   */
+
+	  if (menu->current < 0 || menu->current >= Context->vcount)
+	    menu->current = ci_first_message ();
 	}
 
 	/* check for a fatal error, or all messages deleted */
@@ -749,7 +831,7 @@ void mutt_index_menu (void)
       case OP_MAIN_TAG_PATTERN:
 
 	CHECK_MSGCOUNT;
-	mutt_pattern_func (M_TAG, "Tag messages matching: ", CURHDR);
+	mutt_pattern_func (M_TAG, _("Tag messages matching: "));
 	menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
 	break;
 
@@ -757,14 +839,14 @@ void mutt_index_menu (void)
 
 	CHECK_MSGCOUNT;
 	CHECK_READONLY;
-	if (mutt_pattern_func (M_UNDELETE, "Undelete messages matching: ", CURHDR) == 0)
+	if (mutt_pattern_func (M_UNDELETE, _("Undelete messages matching: ")) == 0)
 	  menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
 	break;
 
       case OP_MAIN_UNTAG_PATTERN:
 
 	CHECK_MSGCOUNT;
-	if (mutt_pattern_func (M_UNTAG, "Untag messages matching: ", CURHDR) == 0)
+	if (mutt_pattern_func (M_UNTAG, _("Untag messages matching: ")) == 0)
 	  menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
 	break;
 
@@ -774,7 +856,7 @@ void mutt_index_menu (void)
 
       case OP_MAIN_CHANGE_FOLDER:
       
-	if (option (OPTREADONLY))
+	if (attach_msg || option (OPTREADONLY))
 	  op = OP_MAIN_CHANGE_FOLDER_READONLY;
 
 	/* fallback to the readonly case */
@@ -782,9 +864,9 @@ void mutt_index_menu (void)
       case OP_MAIN_CHANGE_FOLDER_READONLY:
 
         if (op == OP_MAIN_CHANGE_FOLDER)
-          cp = "Open mailbox";
+          cp = _("Open mailbox");
         else
-          cp = "Open mailbox in read-only mode";
+          cp = _("Open mailbox in read-only mode");
 
 	buf[0] = '\0';
 	mutt_buffy (buf);
@@ -800,7 +882,7 @@ void mutt_index_menu (void)
 	mutt_expand_path (buf, sizeof (buf));
 	if (mx_get_magic (buf) <= 0)
 	{
-	  mutt_error ("%s is not a mailbox.", buf);
+	  mutt_error (_("%s is not a mailbox."), buf);
 	  break;
 	}
 
@@ -853,6 +935,14 @@ void mutt_index_menu (void)
 
 	unset_option (OPTNEEDRESORT);
 
+	if ((Sort & SORT_MASK) == SORT_THREADS && CURHDR->collapsed)
+	{
+	  mutt_uncollapse_thread (Context, CURHDR);
+	  mutt_set_virtual (Context);
+	  if (option (OPTUNCOLLAPSEJUMP))
+	    menu->current = mutt_thread_next_unread (Context, CURHDR);
+	}
+ 
 	if ((op = mutt_display_message (CURHDR)) == -1)
 	{
 	  unset_option (OPTNEEDRESORT);
@@ -861,16 +951,7 @@ void mutt_index_menu (void)
 
 	if (option (OPTNEEDRESORT) && Context && Context->msgcount)
 	{
-	  menu->oldcurrent = CURHDR->index;
-	  mutt_sort_headers (Context, 0);
-
-	  /* try to restore the current message */
-	  for (i = 0; i < Context->vcount; i++)
-	  {
-	    if (Context->hdrs[Context->v2r[i]]->index == menu->oldcurrent)
-	      menu->current = i;
-	  }
-	  menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
+	  resort_index (menu);
 	}
 
 	menu->menu = MENU_PAGER;
@@ -879,14 +960,21 @@ void mutt_index_menu (void)
 
       case OP_EXIT:
 
+	close = op;
+	if (menu->menu == MENU_MAIN && attach_msg)
+	{
+	 done = 1;
+	 break;
+	}
+
 	if ((menu->menu == MENU_MAIN)
 	    && (query_quadoption (OPT_QUIT, 
-				  "Exit Mutt without saving?") == M_YES))
+				  _("Exit Mutt without saving?")) == M_YES))
 	{
 	  if (Context)
 	  {
 	    mx_fastclose_mailbox (Context);
-	    safe_free ((void **)&Context);
+	    safe_free ((void **) &Context);
 	  }
 	  done = 1;
 	}
@@ -898,14 +986,14 @@ void mutt_index_menu (void)
 	if (menu->current >= Context->vcount - 1)
 	{
 	  if (menu->menu == MENU_MAIN)
-	    mutt_error ("You are on the last message.");
+	    mutt_error _("You are on the last message.");
 	  break;
 	}
 	if ((menu->current = ci_next_undeleted (menu->current)) == -1)
 	{
 	  menu->current = menu->oldcurrent;
 	  if (menu->menu == MENU_MAIN)
-	    mutt_error ("No undeleted messages.");
+	    mutt_error _("No undeleted messages.");
 	}
 	else if (menu->menu == MENU_PAGER)
 	{
@@ -922,7 +1010,7 @@ void mutt_index_menu (void)
 	if (menu->current >= Context->vcount - 1)
 	{
 	  if (menu->menu == MENU_MAIN)
-	    mutt_error ("You are on the last message.");
+	    mutt_error _("You are on the last message.");
 	  break;
 	}
 	menu->current++;
@@ -940,14 +1028,14 @@ void mutt_index_menu (void)
 	CHECK_MSGCOUNT;
 	if (menu->current < 1)
 	{
-	  mutt_error ("You are on the first message.");
+	  mutt_error _("You are on the first message.");
 	  break;
 	}
 	if ((menu->current = ci_previous_undeleted (menu->current)) == -1)
 	{
 	  menu->current = menu->oldcurrent;
 	  if (menu->menu == MENU_MAIN)
-	    mutt_error ("No undeleted messages.");
+	    mutt_error _("No undeleted messages.");
 	}
 	else if (menu->menu == MENU_PAGER)
 	{
@@ -963,7 +1051,7 @@ void mutt_index_menu (void)
 	CHECK_MSGCOUNT;
 	if (menu->current < 1)
 	{
-	  if (menu->menu == MENU_MAIN) mutt_error ("You are on the first message.");
+	  if (menu->menu == MENU_MAIN) mutt_error _("You are on the first message.");
 	  break;
 	}
 	menu->current--;
@@ -980,13 +1068,27 @@ void mutt_index_menu (void)
       case OP_SAVE:
       case OP_DECODE_COPY:
       case OP_DECODE_SAVE:
-
+#ifdef _PGPPATH
+      case OP_DECRYPT_COPY:
+      case OP_DECRYPT_SAVE:
+#endif
 	CHECK_MSGCOUNT;
-	if (mutt_save_message (tag ? NULL : CURHDR,
-			       (op == OP_SAVE || op == OP_DECODE_SAVE),
-			       (op == OP_DECODE_SAVE || op == OP_DECODE_COPY),
+        if (mutt_save_message (tag ? NULL : CURHDR,
+#ifdef _PGPPATH
+			       (op == OP_DECRYPT_SAVE) ||
+#endif
+			       (op == OP_SAVE) || (op == OP_DECODE_SAVE),
+			       (op == OP_DECODE_SAVE) || (op == OP_DECODE_COPY),
+#ifdef _PGPPATH
+			       (op == OP_DECRYPT_SAVE) || (op == OP_DECRYPT_COPY) ||
+#endif
+			       0,
 			       &menu->redraw) == 0 &&
-	    (op == OP_SAVE || op == OP_DECODE_SAVE))
+	    (op == OP_SAVE || op == OP_DECODE_SAVE
+#ifdef _PGPPATH
+	     || op == OP_DECRYPT_SAVE
+#endif
+	     ))
 	{
 	  if (tag)
 	    menu->redraw |= REDRAW_INDEX;
@@ -1015,12 +1117,13 @@ void mutt_index_menu (void)
 	menu->current = -1;
 	for (j = 0; j != Context->vcount; j++)
 	{
+#define CUR Context->hdrs[Context->v2r[i]]
 	  if (op == OP_MAIN_NEXT_NEW || op == OP_MAIN_NEXT_UNREAD)
 	  {
 	    i++;
 	    if (i > Context->vcount - 1)
 	    {
-	      mutt_message ("Search wrapped to top.");
+	      mutt_message _("Search wrapped to top.");
 	      i = 0;
 	    }
 	  }
@@ -1029,27 +1132,41 @@ void mutt_index_menu (void)
 	    i--;
 	    if (i < 0)
 	    {
-	      mutt_message ("Search wrapped to bottom.");
+	      mutt_message _("Search wrapped to bottom.");
 	      i = Context->vcount - 1;
 	    }
 	  }
 
-	  if (! Context->hdrs[Context->v2r[i]]->deleted &&
-	      ! Context->hdrs[Context->v2r[i]]->read)
+	  if (CUR->collapsed)
 	  {
-	    if (op == OP_MAIN_NEXT_UNREAD || op == OP_MAIN_PREV_UNREAD ||
-		! Context->hdrs[Context->v2r[i]]->old)
+	    if ((op == OP_MAIN_NEXT_UNREAD || op == OP_MAIN_PREV_UNREAD) &&
+		UNREAD (CUR))
+	    {
+	      menu->current = i;
+	      break;
+	    }
+	    if ((op == OP_MAIN_NEXT_NEW || op == OP_MAIN_PREV_NEW) &&
+		UNREAD (CUR) == 1)
+	    {
+	      menu->current = i;
+	      break;
+	    }
+	  }
+	  else if ((!CUR->deleted && !CUR->read))
+	  {
+	    if (op == OP_MAIN_NEXT_UNREAD || op == OP_MAIN_PREV_UNREAD || !CUR->old)
 	    {
 	      menu->current = i;
 	      break;
 	    }
 	  }
 	}
+#undef CUR
 	if (menu->current == -1)
 	{
 	  menu->current = menu->oldcurrent;
-	  mutt_error ((op == OP_MAIN_NEXT_NEW || op == OP_MAIN_PREV_NEW) ? 
-		      "No new messages." : "No unread messages.");
+	  mutt_error ("%s%s.", (op == OP_MAIN_NEXT_NEW || op == OP_MAIN_PREV_NEW) ? _("No new messages") : _("No unread messages"),
+		      Context->pattern ? _(" in this limited view") : "");
 	}
 	else if (menu->menu == MENU_PAGER)
 	{
@@ -1136,12 +1253,6 @@ void mutt_index_menu (void)
       case OP_MAIN_PREV_SUBTHREAD:
 
 	CHECK_MSGCOUNT;
-	if (Context->msgcount != Context->vcount)
-	{
-	  mutt_error ("No threads in limit mode.");
-	  break;
-	}
-
 	switch (op)
 	{
 	  case OP_MAIN_NEXT_THREAD:
@@ -1165,9 +1276,9 @@ void mutt_index_menu (void)
 	{
 	  menu->current = menu->oldcurrent;
 	  if (op == OP_MAIN_NEXT_THREAD || op == OP_MAIN_NEXT_SUBTHREAD)
-	    mutt_error ("No more threads.");
+	    mutt_error _("No more threads.");
 	  else
-	    mutt_error ("You are on the first thread.");
+	    mutt_error _("You are on the first thread.");
 	}
 	else if (menu->menu == MENU_PAGER)
 	{
@@ -1203,12 +1314,88 @@ void mutt_index_menu (void)
 	}
 	break;
 
-	/* --------------------------------------------------------------------
-	 * These functions are invoked directly from the internal-pager
-	 */
+      case OP_MAIN_COLLAPSE_THREAD:
+	CHECK_MSGCOUNT;
+	
+	if (CURHDR->collapsed)
+	{
+	  menu->current = mutt_uncollapse_thread (Context, CURHDR);
+	  mutt_set_virtual (Context);
+	  if (option (OPTUNCOLLAPSEJUMP))
+	    menu->current = mutt_thread_next_unread (Context, CURHDR);
+	}
+	else if (option (OPTCOLLAPSEUNREAD) || !UNREAD (CURHDR))
+	{
+	  menu->current = mutt_collapse_thread (Context, CURHDR);
+	  mutt_set_virtual (Context);
+	}
+	else
+	{
+	  mutt_error _("Thread contains unread messages.");
+	  break;
+	}
+
+	menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
+
+       break;
+
+      case OP_MAIN_COLLAPSE_ALL:
+        CHECK_MSGCOUNT;
+
+        if ((Sort & SORT_MASK) != SORT_THREADS)
+        {
+	  mutt_error _("Threading is not enabled.");
+	  break;
+	}
+
+        {
+	  HEADER *h, *base;
+	  int final;
+	  
+	  if (CURHDR->collapsed)
+	    final = mutt_uncollapse_thread (Context, CURHDR);
+	  else if (option (OPTCOLLAPSEUNREAD) || !UNREAD (CURHDR))
+	    final = mutt_collapse_thread (Context, CURHDR);
+	  else
+	    final = CURHDR->virtual;
+	  
+	  base = Context->hdrs[Context->v2r[final]];
+	  
+	  h = Context->tree;
+	  Context->collapsed = !Context->collapsed;
+	  while (h)
+	  {
+	    if (h->collapsed != Context->collapsed)
+	    {
+	      if (h->collapsed)
+		mutt_uncollapse_thread (Context, h);
+	      else if (option (OPTCOLLAPSEUNREAD) || !UNREAD (h))
+		mutt_collapse_thread (Context, h);
+	    }
+	    h = h->next;
+	  }
+	  
+	  mutt_set_virtual (Context);
+	  for (j = 0; j < Context->vcount; j++)
+	  {
+	    if (Context->hdrs[Context->v2r[j]]->index == base->index)
+	    {
+	      menu->current = j;
+	      break;
+	    }
+	  }
+	  
+	  menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
+	  break;
+	}
+      
+      /* --------------------------------------------------------------------
+       * These functions are invoked directly from the internal-pager
+       */
 
       case OP_BOUNCE_MESSAGE:
 
+	CHECK_ATTACH;
 	CHECK_MSGCOUNT;
 	ci_bounce_message (tag ? NULL : CURHDR, &menu->redraw);
 	break;
@@ -1220,6 +1407,7 @@ void mutt_index_menu (void)
 	break;
 
       case OP_QUERY:
+	CHECK_ATTACH;
 	mutt_query_menu (NULL, 0);
 	MAYBE_REDRAW (menu->redraw);
 	break;
@@ -1283,20 +1471,12 @@ void mutt_index_menu (void)
 
       case OP_ENTER_COMMAND:
 
+	CurrentMenu = MENU_MAIN;
 	mutt_enter_command ();
 	mutt_check_rescore (Context);
 	if (option (OPTNEEDRESORT) && Context && Context->msgcount)
 	{
-	  menu->oldcurrent = CURHDR->index;
-	  mutt_sort_headers (Context, 0);
-
-	  /* try to restore the current message */
-	  for (i = 0; i < Context->vcount; i++)
-	  {
-	    if (Context->hdrs[Context->v2r[i]]->index == menu->oldcurrent)
-	      menu->current = i;
-	  }
-	  menu->redraw = REDRAW_INDEX | REDRAW_STATUS;
+	  resort_index (menu);
 	}
 	if (option (OPTFORCEREDRAWINDEX))
 	  menu->redraw = REDRAW_FULL;
@@ -1304,27 +1484,48 @@ void mutt_index_menu (void)
 	unset_option (OPTFORCEREDRAWPAGER);
 	break;
 
+      case OP_EDIT_MESSAGE:
+
+	CHECK_MSGCOUNT;
+	CHECK_READONLY;
+	CHECK_ATTACH;
+        
+        set_option(OPTUSEHEADERDATE);
+	ci_send_message (SENDEDITMSG, NULL, NULL, Context, CURHDR);
+        unset_option(OPTUSEHEADERDATE);
+	menu->redraw = REDRAW_FULL;
+	break;
+
       case OP_FORWARD_MESSAGE:
 
 	CHECK_MSGCOUNT;
+	CHECK_ATTACH;
 	ci_send_message (SENDFORWARD, NULL, NULL, Context, tag ? NULL : CURHDR);
 	menu->redraw = REDRAW_FULL;
 	break;
 
 
 
+#ifdef _PGPPATH
+      case OP_FORGET_PASSPHRASE:
+
+	mutt_forget_passphrase ();
+	break;
+#endif /* _PGPPATH */
 
 
 
       case OP_GROUP_REPLY:
 
 	CHECK_MSGCOUNT;
+	CHECK_ATTACH;
 	ci_send_message (SENDREPLY|SENDGROUPREPLY, NULL, NULL, Context, tag ? NULL : CURHDR);
 	menu->redraw = REDRAW_FULL;
 	break;
 
       case OP_LIST_REPLY:
 
+	CHECK_ATTACH;
 	CHECK_MSGCOUNT;
 	ci_send_message (SENDREPLY|SENDLISTREPLY, NULL, NULL, Context, tag ? NULL : CURHDR);
 	menu->redraw = REDRAW_FULL;
@@ -1332,7 +1533,8 @@ void mutt_index_menu (void)
 
       case OP_MAIL:
 
-	ci_send_message (0, NULL, NULL, NULL, NULL);
+	CHECK_ATTACH;
+	ci_send_message (0, NULL, NULL, Context, NULL);
 	menu->redraw = REDRAW_FULL;
 	break;
 
@@ -1342,6 +1544,22 @@ void mutt_index_menu (void)
 
 
 
+#ifdef _PGPPATH
+      case OP_MAIL_KEY:
+	
+	CHECK_ATTACH;
+	ci_send_message (SENDKEY, NULL, NULL, NULL, NULL);
+	menu->redraw = REDRAW_FULL;
+	break;
+      
+      case OP_EXTRACT_KEYS:
+      
+        CHECK_MSGCOUNT;
+        pgp_extract_keys_from_messages(tag ? NULL : CURHDR);
+        menu->redraw = REDRAW_FULL;
+        break;
+      
+#endif /* _PGPPATH */
 
 
 
@@ -1353,6 +1571,7 @@ void mutt_index_menu (void)
 
 	CHECK_MSGCOUNT;
 	mutt_pipe_message (tag ? NULL : CURHDR);
+	MAYBE_REDRAW (menu->redraw);
 	break;
 
       case OP_PRINT:
@@ -1383,12 +1602,14 @@ void mutt_index_menu (void)
 
       case OP_RECALL_MESSAGE:
 
+	CHECK_ATTACH;
 	ci_send_message (SENDPOSTPONED, NULL, NULL, Context, NULL);
 	menu->redraw = REDRAW_FULL;
 	break;
 
       case OP_REPLY:
 
+	CHECK_ATTACH;
 	CHECK_MSGCOUNT;
 	ci_send_message (SENDREPLY, NULL, NULL, Context, tag ? NULL : CURHDR);
 	menu->redraw = REDRAW_FULL;
@@ -1496,6 +1717,7 @@ void mutt_index_menu (void)
   }
 
   mutt_menuDestroy (&menu);
+  return (close);
 }
 
 void mutt_set_header_color (CONTEXT *ctx, HEADER *curhdr)
@@ -1513,12 +1735,3 @@ void mutt_set_header_color (CONTEXT *ctx, HEADER *curhdr)
    }
   curhdr->pair = ColorDefs[MT_COLOR_NORMAL];
 }
-
-void mutt_cache_index_colors (CONTEXT *ctx)
-{
- int i;
-
- if (ctx)
-   for (i = 0; i < ctx->msgcount; i++)
-     mutt_set_header_color (ctx, ctx->hdrs[i]);
-}
