@@ -33,7 +33,7 @@
  * SLcurses_waddnstr() can't take a "const char *", so this is only
  * declared "static" (sigh)
  */
-static char EditorHelp[] = "\
+static char EditorHelp[] = N_("\
 ~~		insert a line begining with a single ~\n\
 ~b users	add users to the Bcc: field\n\
 ~c users	add users to the Cc: field\n\
@@ -51,7 +51,7 @@ static char EditorHelp[] = "\
 ~w file		write message to file\n\
 ~x		abort changes and quit editor\n\
 ~?		this message\n\
-.		on a line by itself ends input\n";
+.		on a line by itself ends input\n");
 
 static char **
 be_snarf_data (FILE *f, char **buf, int *bufmax, int *buflen, int offset,
@@ -64,8 +64,8 @@ be_snarf_data (FILE *f, char **buf, int *bufmax, int *buflen, int offset,
   tmp[sizeof (tmp) - 1] = 0;
   if (prefix)
   {
-    strfcpy (tmp, Prefix, sizeof (tmp));
-    tmplen = strlen (tmp);
+    strfcpy (tmp, NONULL(Prefix), sizeof (tmp));
+    tmplen = mutt_strlen (tmp);
     p = tmp + tmplen;
     tmplen = sizeof (tmp) - tmplen;
   }
@@ -74,7 +74,7 @@ be_snarf_data (FILE *f, char **buf, int *bufmax, int *buflen, int offset,
   while (bytes > 0)
   {
     if (fgets (p, tmplen - 1, f) == NULL) break;
-    bytes -= strlen (p);
+    bytes -= mutt_strlen (p);
     if (*bufmax == *buflen)
       safe_realloc ((void **)&buf, sizeof (char *) * (*bufmax += 25));
     buf[(*buflen)++] = safe_strdup (tmp);
@@ -129,9 +129,9 @@ static int be_barf_file (const char *path, char **buf, int buflen)
 static void be_free_memory (char **buf, int buflen)
 {
   while (buflen-- > 0)
-    free (buf[buflen]);
+    FREE (&buf[buflen]);
   if (buf)
-    free (buf);
+    FREE (&buf);
 }
 
 static char **
@@ -149,11 +149,14 @@ be_include_messages (char *msg, char **buf, int *bufmax, int *buflen,
       n--;
 
       /* add the attribution */
-      mutt_make_string (tmp, sizeof (tmp)-1, Attribution, Context->hdrs[n]);
-      strcat (tmp, "\n");
+      if (Attribution)
+      {
+	mutt_make_string (tmp, sizeof (tmp) - 1, Attribution, Context, Context->hdrs[n]);
+	strcat (tmp, "\n");
+      }
 
       if (*bufmax == *buflen)
-	safe_realloc ((void **)&buf, sizeof (char *) * (*bufmax += 25));
+	safe_realloc ((void **) &buf, sizeof (char *) * (*bufmax += 25));
       buf[(*buflen)++] = safe_strdup (tmp);
 
       bytes = Context->hdrs[n]->content->length;
@@ -172,7 +175,7 @@ be_include_messages (char *msg, char **buf, int *bufmax, int *buflen,
       buf[(*buflen)++] = safe_strdup ("\n");
     }
     else
-      printw ("%d: invalid message number.\n", n);
+      printw (_("%d: invalid message number.\n"), n);
     msg = NULL;
   }
   return (buf);
@@ -306,7 +309,7 @@ int mutt_builtin_editor (const char *path, HEADER *msg, HEADER *cur)
 
   be_edit_header (msg->env, 0);
 
-  addstr ("(End message with a . on a line by itself)\n");
+  addstr (_("(End message with a . on a line by itself)\n"));
 
   buf = be_snarf_file (path, buf, &bufmax, &buflen, 0);
 
@@ -320,10 +323,10 @@ int mutt_builtin_editor (const char *path, HEADER *msg, HEADER *cur)
     }
     addch ('\n');
 
-    if (tmp[0] == EscChar[0] && tmp[1] != EscChar[0])
+    if (EscChar && tmp[0] == EscChar[0] && tmp[1] != EscChar[0])
     {
       /* remove trailing whitespace from the line */
-      p = tmp + strlen (tmp) - 1;
+      p = tmp + mutt_strlen (tmp) - 1;
       while (p >= tmp && ISSPACE (*p))
 	*p-- = 0;
 
@@ -333,7 +336,7 @@ int mutt_builtin_editor (const char *path, HEADER *msg, HEADER *cur)
       switch (tmp[1])
       {
 	case '?':
-	  addstr (EditorHelp);
+	  addstr (_(EditorHelp));
 	  break;
 	case 'b':
 	  msg->env->bcc = mutt_parse_adrlist (msg->env->bcc, p);
@@ -353,35 +356,39 @@ int mutt_builtin_editor (const char *path, HEADER *msg, HEADER *cur)
 	  if (Context)
 	  {
 	    if (!*p && cur)
-	    {
+ 	    {
 	      /* include the current message */
-	      p = tmp + strlen (tmp) + 1;
-	      snprintf (tmp + strlen (tmp), sizeof (tmp) - strlen (tmp), " %d",
+	      p = tmp + mutt_strlen (tmp) + 1;
+	      snprintf (tmp + mutt_strlen (tmp), sizeof (tmp) - mutt_strlen (tmp), " %d",
 								cur->msgno + 1);
 	    }
 	    buf = be_include_messages (p, buf, &bufmax, &buflen,
 				       (tolower (tmp[1]) == 'm'),
-				       (isupper (tmp[1])));
+				       (isupper ((unsigned char) tmp[1])));
 	  }
 	  else
-	    addstr ("No mailbox.\n");
+	    addstr (_("No mailbox.\n"));
 	  break;
 	case 'p':
 	  addstr ("-----\n");
-	  addstr ("Message contains:\n");
+	  addstr (_("Message contains:\n"));
 	  be_print_header (msg->env);
 	  for (i = 0; i < buflen; i++)
 	    addstr (buf[i]);
-	  addstr ("(continue)\n");
+	  addstr (_("(continue)\n"));
 	  break;
 	case 'q':
 	  done = 1;
 	  break;
 	case 'r':
 	  if (*p)
-	    buf = be_snarf_file (p, buf, &bufmax, &buflen, 1);
+          {
+	    strncpy(tmp, p, sizeof(tmp));
+	    mutt_expand_path(tmp, sizeof(tmp));
+	    buf = be_snarf_file (tmp, buf, &bufmax, &buflen, 1);
+          }
 	  else
-	    addstr ("missing filename.\n");
+	    addstr (_("missing filename.\n"));
 	  break;
 	case 's':
 	  safe_free ((void **) &msg->env->subject);
@@ -396,13 +403,13 @@ int mutt_builtin_editor (const char *path, HEADER *msg, HEADER *cur)
 	  {
 	    buflen--;
 	    strfcpy (tmp, buf[buflen], sizeof (tmp));
-	    tmp[strlen (tmp)-1] = 0;
-	    free (buf[buflen]);
+	    tmp[mutt_strlen (tmp)-1] = 0;
+	    FREE (&buf[buflen]);
 	    buf[buflen] = NULL;
 	    continue;
 	  }
 	  else
-	    addstr ("No lines in message.\n");
+	    addstr (_("No lines in message.\n"));
 	  break;
 
 	case 'e':
@@ -414,13 +421,13 @@ int mutt_builtin_editor (const char *path, HEADER *msg, HEADER *cur)
 	    bufmax = buflen = 0;
 
 	    if (option (OPTEDITHDRS))
-	      mutt_edit_headers (Visual, path, msg, NULL, 0);
+	      mutt_edit_headers (NONULL(Visual), path, msg, NULL, 0);
 	    else
-	      mutt_edit_file (Visual, path);
+	      mutt_edit_file (NONULL(Visual), path);
 
 	    buf = be_snarf_file (path, buf, &bufmax, &buflen, 0);
 
-	    addstr ("(continue)\n");
+	    addstr (_("(continue)\n"));
 	  }
 	  break;
 	case 'w':
@@ -431,15 +438,15 @@ int mutt_builtin_editor (const char *path, HEADER *msg, HEADER *cur)
 	  done = 1;
 	  break;
 	default:
-	  printw ("%s: unknown editor command (~? for help)\n", tmp);
+	  printw (_("%s: unknown editor command (~? for help)\n"), tmp);
 	  break;
       }
     }
-    else if (strcmp (".", tmp) == 0)
+    else if (mutt_strcmp (".", tmp) == 0)
       done = 1;
     else
     {
-      strcat (tmp, "\n");
+      strncat (tmp, "\n", sizeof(tmp)); tmp[sizeof(tmp) - 1] = '\0';
       if (buflen == bufmax)
 	safe_realloc ((void **)&buf, sizeof (char *) * (bufmax += 25));
       buf[buflen++] = safe_strdup (tmp[1] == '~' ? tmp + 1 : tmp);
