@@ -20,6 +20,9 @@
 
 
 
+#ifdef _PGPPATH
+#include "pgp.h"
+#endif
 
 
 
@@ -125,7 +128,7 @@ void mutt_edit_headers (const char *editor,
   /* Read the temp file back in */
   ifp = fopen (path, "r");
   ofp = safe_fopen (body, "w");
-  n = mutt_read_rfc822_header (ifp, NULL);
+  n = mutt_read_rfc822_header (ifp, NULL, 1);
   while ((i = fread (buffer, 1, sizeof (buffer), ifp)) > 0)
     fwrite (buffer, 1, i, ofp);
   fclose (ofp);
@@ -155,9 +158,9 @@ void mutt_edit_headers (const char *editor,
      * not, remove the references: field later so that we can generate a new
      * message based upon this one.
      */
-    if (strncasecmp ("in-reply-to:", cur->data, 12) == 0)
+    if (mutt_strncasecmp ("in-reply-to:", cur->data, 12) == 0)
       in_reply_to = 1;
-    else if (fcc && strncasecmp ("fcc:", cur->data, 4) == 0)
+    else if (fcc && mutt_strncasecmp ("fcc:", cur->data, 4) == 0)
     {
       p = cur->data + 4;
       SKIPWS (p);
@@ -168,7 +171,7 @@ void mutt_edit_headers (const char *editor,
       }
       keep = 0;
     }
-    else if (strncasecmp ("attach:", cur->data, 7) == 0)
+    else if (mutt_strncasecmp ("attach:", cur->data, 7) == 0)
     {
       BODY *body;
       BODY *parts;
@@ -186,7 +189,7 @@ void mutt_edit_headers (const char *editor,
 	else
 	  strfcpy (path, p, sizeof (path));
 	mutt_expand_path (path, sizeof (path));
-	if ((body = mutt_make_attach (path)))
+	if ((body = mutt_make_file_attach (path)))
 	{
 	  body->description = safe_strdup (q);
 	  for (parts = msg->content; parts->next; parts = parts->next) ;
@@ -195,7 +198,7 @@ void mutt_edit_headers (const char *editor,
 	else
 	{
 	  mutt_pretty_mailbox (path);
-	  mutt_error ("%s: unable to attach file", path);
+	  mutt_error (_("%s: unable to attach file"), path);
 	}
       }
       keep = 0;
@@ -203,37 +206,13 @@ void mutt_edit_headers (const char *editor,
 
 
 
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
+#ifdef _PGPPATH
+    else if (mutt_strncasecmp ("pgp:", cur->data, 4) == 0)
+    {
+      msg->pgp = mutt_parse_pgp_hdr (cur->data + 4, 0);
+      keep = 0;
+    }
+#endif
 
 
     if (keep)
