@@ -1220,8 +1220,6 @@ int imap_fetch_message (MESSAGE *msg, CONTEXT *ctx, int msgno)
   ctx->hdrs[msgno]->env = mutt_read_rfc822_header (msg->fp, ctx->hdrs[msgno],0);
 
   /* This needs to be done in case this is a multipart message */
-    if (mutt_needs_mailcap (ctx->hdrs[msgno]->content))
-      ctx->hdrs[msgno]->mailcap = 1;
 
   mutt_clear_error();
   rewind (msg->fp);
@@ -1242,6 +1240,7 @@ int imap_append_message (CONTEXT *ctx, MESSAGE *msg)
   FILE *fp;
   char seq[8];
   char buf[LONG_STRING];
+  char mbox[LONG_STRING];
   size_t len;
   int c, last;
   
@@ -1262,8 +1261,9 @@ int imap_append_message (CONTEXT *ctx, MESSAGE *msg)
   
   mutt_message _("Sending APPEND command ...");
   imap_make_sequence (seq, sizeof (seq));
+  imap_quote_string (mbox, sizeof (mbox), CTX_DATA->mailbox);
   snprintf (buf, sizeof (buf), "%s APPEND %s {%d}\r\n", seq, 
-      CTX_DATA->mailbox, len);
+	    mbox, len);
 
   mutt_socket_write (CTX_DATA->conn, buf);
 
