@@ -34,17 +34,20 @@ struct mapping_t Menus[] = {
  { "browser",	MENU_FOLDER },
  { "compose",	MENU_COMPOSE },
  { "editor",	MENU_EDITOR },
- { "generic",	MENU_GENERIC },
  { "index",	MENU_MAIN },
  { "pager",	MENU_PAGER },
  { "postpone",	MENU_POST },
 
   
 
+#ifdef _PGPPATH
+ { "pgp",	MENU_PGP },
+#endif  
   
   
 
  { "query",	MENU_QUERY },
+ { "generic",	MENU_GENERIC },
  { NULL,	0 }
 };
 
@@ -62,8 +65,13 @@ static struct mapping_t KeyNames[] = {
   { "<Insert>",	KEY_IC },
   { "<Home>",	KEY_HOME },
   { "<End>",	KEY_END },
+#ifdef KEY_ENTER
   { "<Enter>",	KEY_ENTER },
+#endif
   { "<Return>",	M_ENTER_C },
+  { "<Esc>",	'\033' },
+  { "<Tab>",	'\t' },
+  { "<Space>",	' ' },
   { NULL,	0 }
 };
 
@@ -83,28 +91,58 @@ static struct keymap_t *allocKeys (int len, keycode_t *keys)
   return (p);
 }
 
-static int parsekeys (char *s, keycode_t *d, int max)
+static int parse_fkey(char *s)
+{
+  char *t;
+  int n = 0;
+
+  if(s[0] != '<' || tolower(s[1]) != 'f')
+    return -1;
+
+  for(t = s + 2; *t && isdigit((unsigned char) *t); t++)
+  {
+    n *= 10;
+    n += *t - '0';
+  }
+
+  if(*t != '>')
+    return -1;
+  else
+    return n;
+}
+
+static int parsekeys (char *str, keycode_t *d, int max)
 {
   int n, len = max;
+  char buff[SHORT_STRING];
+  char c;
+  char *s, *t;
 
+  strfcpy(buff, str, sizeof(buff));
+  s = buff;
+  
   while (*s && len)
   {
-    if ((n = mutt_getvaluebyname (s, KeyNames)) != -1)
+    *d = '\0';
+    if(*s == '<' && (t = strchr(s, '>')))
     {
-      s += strlen (s);
-      *d = n;
-    }
-    else if (tolower (*s) == 'f' && isdigit ((unsigned char) s[1]))
-    {
-      n = 0;
-      for (s++; isdigit ((unsigned char) *s) ; s++)
+      t++; c = *t; *t = '\0';
+      
+      if ((n = mutt_getvaluebyname (s, KeyNames)) != -1)
       {
-	n *= 10;
-	n += *s - '0';
+	s = t;
+	*d = n;
       }
-      *d = KEY_F(n);
+      else if ((n = parse_fkey(s)) > 0)
+      {
+	s = t;
+	*d = KEY_F (n);
+      }
+      
+      *t = c;
     }
-    else
+
+    if(!*d)
     {
       *d = *s;
       s++;
@@ -119,7 +157,7 @@ static int parsekeys (char *s, keycode_t *d, int max)
 /* insert a key sequence into the specified map.  the map is sorted by ASCII
  * value (lowest to highest)
  */
-void km_bindkey (char *s, int menu, int op, char *macro)
+void km_bind (char *s, int menu, int op, char *macro, char *descr)
 {
   struct keymap_t *map, *tmp, *last = NULL, *next;
   keycode_t buf[MAX_SEQ];
@@ -130,6 +168,7 @@ void km_bindkey (char *s, int menu, int op, char *macro)
   map = allocKeys (len, buf);
   map->op = op;
   map->macro = safe_strdup (macro);
+  map->descr = safe_strdup (descr);
 
   tmp = Keymaps[menu];
 
@@ -142,10 +181,10 @@ void km_bindkey (char *s, int menu, int op, char *macro)
       {
 	len = tmp->eq;
 	next = tmp->next;
-	if (tmp->macro)
-	  free (tmp->macro);
-	free (tmp->keys);
-	free (tmp);
+	FREE (&tmp->macro);
+	FREE (&tmp->keys);
+	FREE (&tmp->descr);
+	FREE (&tmp);
 	tmp = next;
       }
       while (tmp && len >= pos);
@@ -180,11 +219,43 @@ void km_bindkey (char *s, int menu, int op, char *macro)
     Keymaps[menu] = map;
 }
 
+void km_bindkey (char *s, int menu, int op)
+{
+  km_bind (s, menu, op, NULL, NULL);
+}
+
+static int get_op (struct binding_t *bindings, const char *start, size_t len)
+{
+  int i;
+
+  for (i = 0; bindings[i].name; i++)
+  {
+    if (!mutt_strncasecmp (start, bindings[i].name, len) &&   
+	mutt_strlen (bindings[i].name) == len)
+      return bindings[i].op;
+  }
+
+  return OP_NULL;
+}
+
+static char *get_func (struct binding_t *bindings, int op)
+{
+  int i;
+
+  for (i = 0; bindings[i].name; i++)
+  {
+    if (bindings[i].op == op)
+      return bindings[i].name;
+  }
+
+  return NULL;
+}
+
 static void push_string (char *s)
 {
-  char *pp, *p = s + strlen (s) - 1;
+  char *pp, *p = s + mutt_strlen (s) - 1;
   size_t l;
-  int i;
+  int i, op = OP_NULL;
 
   while (p >= s)
   {
@@ -199,19 +270,39 @@ static void push_string (char *s)
 	l = p - pp + 1;
 	for (i = 0; KeyNames[i].name; i++)
 	{
-	  if (!strncasecmp (pp, KeyNames[i].name, l))
+	  if (!mutt_strncasecmp (pp, KeyNames[i].name, l))
 	    break;
 	}
 	if (KeyNames[i].name)
 	{
 	  /* found a match */
-	  mutt_ungetch (KeyNames[i].value);
+	  mutt_ungetch (KeyNames[i].value, 0);
+	  p = pp - 1;
+	  continue;
+	}
+
+	/* See if it is a valid command
+	 * skip the '<' and the '>' when comparing */
+	for (i = 0; Menus[i].name; i++)
+	{
+	  struct binding_t *binding = km_get_table (Menus[i].value);
+	  if (binding)
+	  {
+	    op = get_op (binding, pp + 1, l - 2);
+	    if (op != OP_NULL)
+	      break;
+	  }
+	}
+
+	if (op != OP_NULL)
+	{
+	  mutt_ungetch (0, op);
 	  p = pp - 1;
 	  continue;
 	}
       }
     }
-    mutt_ungetch (*p--);
+    mutt_ungetch (*p--, 0);
   }
 }
 
@@ -220,9 +311,9 @@ static int retry_generic (int menu, keycode_t *keys, int keyslen, int lastkey)
   if (menu != MENU_EDITOR && menu != MENU_GENERIC && menu != MENU_PAGER)
   {
     if (lastkey)
-      mutt_ungetch (lastkey);
+      mutt_ungetch (lastkey, 0);
     for (; keyslen; keyslen--)
-      mutt_ungetch (keys[keyslen - 1]);
+      mutt_ungetch (keys[keyslen - 1], 0);
     return (km_dokey (MENU_GENERIC));
   }
   if (menu != MENU_EDITOR)
@@ -240,18 +331,86 @@ static int retry_generic (int menu, keycode_t *keys, int keyslen, int lastkey)
  */
 int km_dokey (int menu)
 {
+  event_t tmp;
   struct keymap_t *map = Keymaps[menu];
   int pos = 0;
   int n = 0;
+  int i;
 
   if (!map)
     return (retry_generic (menu, NULL, 0, 0));
 
   FOREVER
   {
-    if ((LastKey = mutt_getch ()) == ERR)
-      return (-1);
+/* ncurses doesn't return on resized screen when timeout is set to zero */
+#if !defined (USE_SLANG_CURSES) && defined (HAVE_RESIZETERM)
+    if (menu == MENU_MAIN || menu == MENU_PAGER)
+      timeout (menu == MENU_MAIN && Timeout > 0 ? Timeout * 1000 : INT_MAX);
+#else
+    if (menu == MENU_MAIN && Timeout > 0)
+      timeout (Timeout * 1000); /* milliseconds */
+#endif
 
+    tmp = mutt_getch();
+
+#if !defined (USE_SLANG_CURSES) && defined (HAVE_RESIZETERM)
+    if (menu == MENU_MAIN || menu == MENU_PAGER)
+#else
+    if (menu == MENU_MAIN && Timeout > 0)
+#endif
+      timeout (-1); /* restore blocking operation */
+
+    LastKey = tmp.ch;
+    if (LastKey == -1)
+      return -1;
+
+    /* do we have an op already? */
+    if (tmp.op)
+    {
+      char *func = NULL;
+      struct binding_t *bindings;
+
+      /* is this a valid op for this menu? */
+      bindings = km_get_table (menu);
+      if ((func = get_func (bindings, tmp.op)))
+	return tmp.op;
+
+      if (menu == MENU_EDITOR && get_func (OpEditor, tmp.op))
+	return tmp.op;
+
+      if (menu != MENU_EDITOR && menu != MENU_PAGER)
+      {
+	/* check generic menu */
+	bindings = OpGeneric; 
+	if ((func = get_func (bindings, tmp.op)))
+	  return tmp.op;
+      }
+
+      /* Sigh. Valid function but not in this context.
+       * Find the literal string and push it back */
+      for (i = 0; Menus[i].name; i++)
+      {
+	bindings = km_get_table (Menus[i].value);
+	if (bindings)
+	{
+	  func = get_func (bindings, tmp.op);
+	  if (func)
+	  {
+	    /* careful not to feed the <..> as one token. otherwise 
+	    * push_string() will push the bogus op right back! */
+	    mutt_ungetch ('>', 0);
+	    push_string (func);
+	    mutt_ungetch ('<', 0);
+	    break;
+	  }
+	}
+      }
+      /* continue to chew */
+      if (func)
+	continue;
+    }
+
+    /* Nope. Business as usual */
     while (LastKey > map->keys[pos])
     {
       if (pos > map->eq || !map->next)
@@ -264,14 +423,15 @@ int km_dokey (int menu)
 
     if (++pos == map->len)
     {
+
       if (map->op != OP_MACRO)
-	return (map->op);
+	return map->op;
 
       if (n++ == 10)
       {
 	mutt_flushinp ();
-	mutt_error ("Macro loop detected.");
-	return (-1);
+	mutt_error _("Macro loop detected.");
+	return -1;
       }
 
       push_string (map->macro);
@@ -289,30 +449,17 @@ static void create_bindings (struct binding_t *map, int menu)
 
   for (i = 0 ; map[i].name ; i++)
     if (map[i].seq)
-      km_bindkey (map[i].seq, menu, map[i].op, NULL);
+      km_bindkey (map[i].seq, menu, map[i].op);
 }
 
 char *km_keyname (int c)
 {
-  static char buf[5];
+  static char buf[10];
   char *p;
 
   if ((p = mutt_getnamebyvalue (c, KeyNames)))
     return p;
 
-  switch (c)
-  {
-    case '\033':
-      return "ESC";
-    case ' ':
-      return "SPC";
-    case '\n':
-    case '\r':
-      return "RET";
-    case '\t':
-      return "TAB";
-  }
-
   if (c < 256 && c > -128 && iscntrl ((unsigned char) c))
   {
     if (c < 0)
@@ -328,7 +475,7 @@ char *km_keyname (int c)
       snprintf (buf, sizeof (buf), "\\%d%d%d", c >> 6, (c >> 3) & 7, c & 7);
   }
   else if (c >= KEY_F0 && c < KEY_F(256)) /* this maximum is just a guess */
-    sprintf (buf, "F%d", c - KEY_F0);
+    sprintf (buf, "<F%d>", c - KEY_F0);
   else if (IsPrint (c))
     snprintf (buf, sizeof (buf), "%c", (unsigned char) c);
   else
@@ -347,13 +494,12 @@ int km_expand_key (char *s, size_t len, struct keymap_t *map)
   FOREVER
   {
     strfcpy (s, km_keyname (map->keys[p]), len);
-    len -= (1 + (l = strlen (s)));
+    len -= (l = mutt_strlen (s));
 
     if (++p >= map->len || !len)
       return (1);
 
     s += l;
-    *(s++) = ' ';
   }
 
   /* not reached */
@@ -383,84 +529,107 @@ void km_init (void)
 
 
 
+#ifdef _PGPPATH
+  create_bindings (OpPgp, MENU_PGP);
+#endif
 
 
   
   /* bindings for the line editor */
   create_bindings (OpEditor, MENU_EDITOR);
   
-  km_bindkey ("<up>", MENU_EDITOR, OP_EDITOR_HISTORY_UP, NULL);
-  km_bindkey ("<down>", MENU_EDITOR, OP_EDITOR_HISTORY_DOWN, NULL);
-  km_bindkey ("<left>", MENU_EDITOR, OP_EDITOR_BACKWARD_CHAR, NULL);
-  km_bindkey ("<right>", MENU_EDITOR, OP_EDITOR_FORWARD_CHAR, NULL);
-  km_bindkey ("<home>", MENU_EDITOR, OP_EDITOR_BOL, NULL);
-  km_bindkey ("<end>", MENU_EDITOR, OP_EDITOR_EOL, NULL);
-  km_bindkey ("<backspace>", MENU_EDITOR, OP_EDITOR_BACKSPACE, NULL);
-  km_bindkey ("<delete>", MENU_EDITOR, OP_EDITOR_BACKSPACE, NULL);
-  km_bindkey ("\177", MENU_EDITOR, OP_EDITOR_BACKSPACE, NULL);
+  km_bindkey ("<up>", MENU_EDITOR, OP_EDITOR_HISTORY_UP);
+  km_bindkey ("<down>", MENU_EDITOR, OP_EDITOR_HISTORY_DOWN);
+  km_bindkey ("<left>", MENU_EDITOR, OP_EDITOR_BACKWARD_CHAR);
+  km_bindkey ("<right>", MENU_EDITOR, OP_EDITOR_FORWARD_CHAR);
+  km_bindkey ("<home>", MENU_EDITOR, OP_EDITOR_BOL);
+  km_bindkey ("<end>", MENU_EDITOR, OP_EDITOR_EOL);
+  km_bindkey ("<backspace>", MENU_EDITOR, OP_EDITOR_BACKSPACE);
+  km_bindkey ("<delete>", MENU_EDITOR, OP_EDITOR_BACKSPACE);
+  km_bindkey ("\177", MENU_EDITOR, OP_EDITOR_BACKSPACE);
   
   /* generic menu keymap */
   create_bindings (OpGeneric, MENU_GENERIC);
   
-  km_bindkey ("<home>", MENU_GENERIC, OP_FIRST_ENTRY, NULL);
-  km_bindkey ("<end>", MENU_GENERIC, OP_LAST_ENTRY, NULL);
-  km_bindkey ("<pagedown>", MENU_GENERIC, OP_NEXT_PAGE, NULL);
-  km_bindkey ("<pageup>", MENU_GENERIC, OP_PREV_PAGE, NULL);
-  km_bindkey ("<right>", MENU_GENERIC, OP_NEXT_PAGE, NULL);
-  km_bindkey ("<left>", MENU_GENERIC, OP_PREV_PAGE, NULL);
-  km_bindkey ("<up>", MENU_GENERIC, OP_PREV_ENTRY, NULL);
-  km_bindkey ("<down>", MENU_GENERIC, OP_NEXT_ENTRY, NULL);
-  km_bindkey ("1", MENU_GENERIC, OP_JUMP, NULL);
-  km_bindkey ("2", MENU_GENERIC, OP_JUMP, NULL);
-  km_bindkey ("3", MENU_GENERIC, OP_JUMP, NULL);
-  km_bindkey ("4", MENU_GENERIC, OP_JUMP, NULL);
-  km_bindkey ("5", MENU_GENERIC, OP_JUMP, NULL);
-  km_bindkey ("6", MENU_GENERIC, OP_JUMP, NULL);
-  km_bindkey ("7", MENU_GENERIC, OP_JUMP, NULL);
-  km_bindkey ("8", MENU_GENERIC, OP_JUMP, NULL);
-  km_bindkey ("9", MENU_GENERIC, OP_JUMP, NULL);
+  km_bindkey ("<home>", MENU_GENERIC, OP_FIRST_ENTRY);
+  km_bindkey ("<end>", MENU_GENERIC, OP_LAST_ENTRY);
+  km_bindkey ("<pagedown>", MENU_GENERIC, OP_NEXT_PAGE);
+  km_bindkey ("<pageup>", MENU_GENERIC, OP_PREV_PAGE);
+  km_bindkey ("<right>", MENU_GENERIC, OP_NEXT_PAGE);
+  km_bindkey ("<left>", MENU_GENERIC, OP_PREV_PAGE);
+  km_bindkey ("<up>", MENU_GENERIC, OP_PREV_ENTRY);
+  km_bindkey ("<down>", MENU_GENERIC, OP_NEXT_ENTRY);
+  km_bindkey ("1", MENU_GENERIC, OP_JUMP);
+  km_bindkey ("2", MENU_GENERIC, OP_JUMP);
+  km_bindkey ("3", MENU_GENERIC, OP_JUMP);
+  km_bindkey ("4", MENU_GENERIC, OP_JUMP);
+  km_bindkey ("5", MENU_GENERIC, OP_JUMP);
+  km_bindkey ("6", MENU_GENERIC, OP_JUMP);
+  km_bindkey ("7", MENU_GENERIC, OP_JUMP);
+  km_bindkey ("8", MENU_GENERIC, OP_JUMP);
+  km_bindkey ("9", MENU_GENERIC, OP_JUMP);
 
   /* Miscellaneous extra bindings */
   
-  km_bindkey (" ", MENU_MAIN, OP_DISPLAY_MESSAGE, NULL);
-  km_bindkey ("<up>", MENU_MAIN, OP_MAIN_PREV_UNDELETED, NULL);
-  km_bindkey ("<down>", MENU_MAIN, OP_MAIN_NEXT_UNDELETED, NULL);
-  km_bindkey ("J", MENU_MAIN, OP_NEXT_ENTRY, NULL);
-  km_bindkey ("K", MENU_MAIN, OP_PREV_ENTRY, NULL);
-  km_bindkey ("x", MENU_MAIN, OP_EXIT, NULL);
+  km_bindkey (" ", MENU_MAIN, OP_DISPLAY_MESSAGE);
+  km_bindkey ("<up>", MENU_MAIN, OP_MAIN_PREV_UNDELETED);
+  km_bindkey ("<down>", MENU_MAIN, OP_MAIN_NEXT_UNDELETED);
+  km_bindkey ("J", MENU_MAIN, OP_NEXT_ENTRY);
+  km_bindkey ("K", MENU_MAIN, OP_PREV_ENTRY);
+  km_bindkey ("x", MENU_MAIN, OP_EXIT);
 
-  km_bindkey ("x", MENU_PAGER, OP_PAGER_EXIT, NULL);
-  km_bindkey ("q", MENU_PAGER, OP_PAGER_EXIT, NULL);
-  km_bindkey ("<backspace>", MENU_PAGER, OP_PREV_LINE, NULL);
-  km_bindkey ("<pagedown>", MENU_PAGER, OP_NEXT_PAGE, NULL);
-  km_bindkey ("<pageup>", MENU_PAGER, OP_PREV_PAGE, NULL);
-  km_bindkey ("<up>", MENU_PAGER, OP_MAIN_PREV_UNDELETED, NULL);
-  km_bindkey ("<right>", MENU_PAGER, OP_MAIN_NEXT_UNDELETED, NULL);
-  km_bindkey ("<down>", MENU_PAGER, OP_MAIN_NEXT_UNDELETED, NULL);
-  km_bindkey ("<left>", MENU_PAGER, OP_MAIN_PREV_UNDELETED, NULL);
-  km_bindkey ("<home>", MENU_PAGER, OP_PAGER_TOP, NULL);
-  km_bindkey ("<end>", MENU_PAGER, OP_PAGER_BOTTOM, NULL);
-  km_bindkey ("1", MENU_PAGER, OP_JUMP, NULL);
-  km_bindkey ("2", MENU_PAGER, OP_JUMP, NULL);
-  km_bindkey ("3", MENU_PAGER, OP_JUMP, NULL);
-  km_bindkey ("4", MENU_PAGER, OP_JUMP, NULL);
-  km_bindkey ("5", MENU_PAGER, OP_JUMP, NULL);
-  km_bindkey ("6", MENU_PAGER, OP_JUMP, NULL);
-  km_bindkey ("7", MENU_PAGER, OP_JUMP, NULL);
-  km_bindkey ("8", MENU_PAGER, OP_JUMP, NULL);
-  km_bindkey ("9", MENU_PAGER, OP_JUMP, NULL);
+  km_bindkey ("x", MENU_PAGER, OP_PAGER_EXIT);
+  km_bindkey ("q", MENU_PAGER, OP_PAGER_EXIT);
+  km_bindkey ("<backspace>", MENU_PAGER, OP_PREV_LINE);
+  km_bindkey ("<pagedown>", MENU_PAGER, OP_NEXT_PAGE);
+  km_bindkey ("<pageup>", MENU_PAGER, OP_PREV_PAGE);
+  km_bindkey ("<up>", MENU_PAGER, OP_MAIN_PREV_UNDELETED);
+  km_bindkey ("<right>", MENU_PAGER, OP_MAIN_NEXT_UNDELETED);
+  km_bindkey ("<down>", MENU_PAGER, OP_MAIN_NEXT_UNDELETED);
+  km_bindkey ("<left>", MENU_PAGER, OP_MAIN_PREV_UNDELETED);
+  km_bindkey ("<home>", MENU_PAGER, OP_PAGER_TOP);
+  km_bindkey ("<end>", MENU_PAGER, OP_PAGER_BOTTOM);
+  km_bindkey ("1", MENU_PAGER, OP_JUMP);
+  km_bindkey ("2", MENU_PAGER, OP_JUMP);
+  km_bindkey ("3", MENU_PAGER, OP_JUMP);
+  km_bindkey ("4", MENU_PAGER, OP_JUMP);
+  km_bindkey ("5", MENU_PAGER, OP_JUMP);
+  km_bindkey ("6", MENU_PAGER, OP_JUMP);
+  km_bindkey ("7", MENU_PAGER, OP_JUMP);
+  km_bindkey ("8", MENU_PAGER, OP_JUMP);
+  km_bindkey ("9", MENU_PAGER, OP_JUMP);
 
-  km_bindkey ("<return>", MENU_ALIAS, OP_TAG, NULL);
+  km_bindkey ("<return>", MENU_ALIAS, OP_TAG);
+
+  /* edit-to (default "t") hides generic tag-entry in Compose menu
+     This will bind tag-entry to  "T" in the Compose menu */
+  km_bindkey ("T", MENU_COMPOSE, OP_TAG);
 }
 
 void km_error_key (int menu)
 {
   char buf[SHORT_STRING];
+  struct keymap_t *key;
 
-  if (km_expand_key (buf, sizeof (buf), km_find_func (menu, OP_HELP)))
-    mutt_error ("Key is not bound.  Press '%s' for help.", buf);
-  else
-    mutt_error ("Key is not bound.  See the manual.");
+  if(!(key = km_find_func(menu, OP_HELP)))
+    key = km_find_func(MENU_GENERIC, OP_HELP);
+  
+  if(!(km_expand_key(buf, sizeof(buf), key)))
+  {
+    mutt_error _("Key is not bound.");
+    return;
+  }
+
+  /* make sure the key is really the help key in this menu */
+  push_string (buf);
+  if (km_dokey (menu) != OP_HELP)
+  {
+    mutt_error _("Key is not bound.");
+    return;
+  }
+
+  mutt_error (_("Key is not bound.  Press '%s' for help."), buf);
+  return;
 }
 
 int mutt_parse_push (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
@@ -470,7 +639,7 @@ int mutt_parse_push (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
   mutt_extract_token (buf, s, M_TOKEN_CONDENSE);
   if (MoreArgs (s))
   {
-    strfcpy (err->data, "push: too many arguments", err->dsize);
+    strfcpy (err->data, _("push: too many arguments"), err->dsize);
     r = -1;
   }
   else
@@ -491,7 +660,7 @@ char *parse_keymap (int *menu, BUFFER *s, BUFFER *err)
   {
     if ((*menu = mutt_check_menu (buf.data)) == -1)
     {
-      snprintf (err->data, err->dsize, "%s: no such menu", buf.data);
+      snprintf (err->data, err->dsize, _("%s: no such menu"), buf.data);
     }
     else
     {
@@ -500,7 +669,7 @@ char *parse_keymap (int *menu, BUFFER *s, BUFFER *err)
 
       if (!*buf.data)
       {
-	strfcpy (err->data, "null key sequence", err->dsize);
+	strfcpy (err->data, _("null key sequence"), err->dsize);
       }
       else if (MoreArgs (s))
 	return (buf.data);
@@ -508,7 +677,7 @@ char *parse_keymap (int *menu, BUFFER *s, BUFFER *err)
   }
   else
   {
-    strfcpy (err->data, "too few arguments", err->dsize);
+    strfcpy (err->data, _("too few arguments"), err->dsize);
   }
   FREE (&buf.data);
   return (NULL);
@@ -520,9 +689,9 @@ try_bind (char *key, int menu, char *func, struct binding_t *bindings)
   int i;
   
   for (i = 0; bindings[i].name; i++)
-    if (strcmp (func, bindings[i].name) == 0)
+    if (mutt_strcmp (func, bindings[i].name) == 0)
     {
-      km_bindkey (key, menu, bindings[i].op, NULL);
+      km_bindkey (key, menu, bindings[i].op);
       return (0);
     }
   return (-1);
@@ -553,6 +722,10 @@ struct binding_t *km_get_table (int menu)
 
 
 
+#ifdef _PGPPATH
+    case MENU_PGP:
+      return OpPgp;
+#endif
 
 
 
@@ -574,11 +747,11 @@ int mutt_parse_bind (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
   mutt_extract_token (buf, s, 0);
   if (MoreArgs (s))
   {
-    strfcpy (err->data, "bind: too many arguments", err->dsize);
+    strfcpy (err->data, _("bind: too many arguments"), err->dsize);
     r = -1;
   }
-  else if (strcasecmp ("noop", buf->data) == 0)
-    km_bindkey (key, menu, OP_NULL, NULL); /* the `unbind' command */
+  else if (mutt_strcasecmp ("noop", buf->data) == 0)
+    km_bindkey (key, menu, OP_NULL); /* the `unbind' command */
   else
   {
     /* First check the "generic" list of commands */
@@ -589,7 +762,7 @@ int mutt_parse_bind (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
       bindings = km_get_table (menu);
       if (bindings && try_bind (key, menu, buf->data, bindings) != 0)
       {
-	snprintf (err->data, err->dsize, "%s: no such function in map", buf->data);
+	snprintf (err->data, err->dsize, _("%s: no such function in map"), buf->data);
 	r = -1;
       }
     }
@@ -598,10 +771,11 @@ int mutt_parse_bind (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
   return (r);
 }
 
-/* macro <menu> <key> <macro> */
+/* macro <menu> <key> <macro> <description> */
 int mutt_parse_macro (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
 {
   int menu, r = -1;
+  char *seq = NULL;
   char *key;
 
   if ((key = parse_keymap (&menu, s, err)) == NULL)
@@ -611,17 +785,76 @@ int mutt_parse_macro (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
   /* make sure the macro sequence is not an empty string */
   if (!*buf->data)
   {
-    strfcpy (err->data, "macro: empty key sequence", err->dsize);
-  }
-  else if (MoreArgs (s))
-  {
-    strfcpy (err->data, "macro: too many arguments", err->dsize);
+    strfcpy (err->data, _("macro: empty key sequence"), err->dsize);
   }
   else
   {
-    km_bindkey (key, menu, OP_MACRO, buf->data);
-    r = 0;
+    if (MoreArgs (s))
+    {
+      seq = strdup (buf->data);
+      mutt_extract_token (buf, s, M_TOKEN_CONDENSE);
+
+      if (MoreArgs (s))
+      {
+	strfcpy (err->data, _("macro: too many arguments"), err->dsize);
+      }
+      else
+      {
+	km_bind (key, menu, OP_MACRO, seq, buf->data);
+	r = 0;
+      }
+
+      FREE (&seq);
+    }
+    else
+    {
+      km_bind (key, menu, OP_MACRO, buf->data, NULL);
+      r = 0;
+    }
   }
   FREE (&key);
   return (r);
 }
+
+/* exec command-name */
+int mutt_parse_exec (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
+{
+  int ops[128]; 
+  int nops = 0;
+  struct binding_t *bindings = NULL;
+  char *command;
+  
+  if (!MoreArgs (s))
+  {
+    strfcpy (err->data, _("exec: too few arguments"), err->dsize);
+    return (-1);
+  }
+
+  do
+  {
+    mutt_extract_token (buf, s, 0);
+    command = buf->data;
+
+    if ((bindings = km_get_table (CurrentMenu)) == NULL 
+	&& CurrentMenu != MENU_PAGER)
+      bindings = OpGeneric;
+    
+    ops[nops] = get_op (bindings, command, mutt_strlen(command));
+    if (ops[nops] == OP_NULL && CurrentMenu != MENU_PAGER)
+      ops[nops] = get_op (OpGeneric, command, mutt_strlen(command));
+    
+    if (ops[nops] == OP_NULL)
+    {
+      mutt_flushinp ();
+      mutt_error (_("%s: no such command"), command);
+      return (-1);
+    }
+    nops++;
+  }
+  while(MoreArgs(s) && nops < sizeof(ops)/sizeof(ops[0]));
+  
+  while(nops)
+    mutt_ungetch(0, ops[--nops]);
+  
+  return 0;
+}
