@@ -25,13 +25,15 @@
 /* type for key storage, the rest of mutt works fine with int type */
 typedef short keycode_t;
 
-void km_bindkey (char *, int, int, char *);
+void km_bind (char *, int, int, char *, char *);
+void km_bindkey (char *, int, int);
 int km_dokey (int);
 
 /* entry in the keymap tree */
 struct keymap_t
 {
   char *macro;           /* macro expansion (op == OP_MACRO) */
+  char *descr;           /* description of a macro for the help menu */
   struct keymap_t *next; /* next key in map */
   short op;              /* operation to perform */
   short eq;              /* number of leading keys equal to next entry */
@@ -57,6 +59,9 @@ enum
   MENU_PAGER,
   MENU_POST,
 
+#ifdef _PGPPATH
+  MENU_PGP,
+#endif
 
 
 
@@ -93,6 +98,9 @@ extern struct binding_t OpBrowser[];
 extern struct binding_t OpEditor[];
 extern struct binding_t OpQuery[];
 
+#ifdef _PGPPATH
+extern struct binding_t OpPgp[];
+#endif /* _PGPPATH */
 
 #include "keymap_defs.h"
 
