@@ -19,7 +19,11 @@
 #include "config.h"
 
 #include <stdio.h>
+#ifdef HAVE_UNISTD_H
+# include <unistd.h> /* needed for SEEK_SET */
+#endif
 #include <sys/types.h>
+#include <sys/stat.h>
 #include <time.h>
 #include <limits.h>
 #include <stdarg.h>
@@ -31,6 +35,25 @@
 #include "rfc822.h"
 #include "hash.h"
 
+#ifdef ENABLE_NLS
+# include <libintl.h>
+# define _(a) (gettext (a))
+# ifdef gettext_noop
+#  define N_(a) gettext_noop (a)
+# else
+#  define N_(a) (a)
+# endif
+#else
+# define _(a) (a)
+# define N_(a) a
+#endif
+
+#ifdef SUBVERSION
+# define MUTT_VERSION (VERSION SUBVERSION)
+#else  
+# define MUTT_VERSION (VERSION)
+#endif
+
 /* nifty trick I stole from ELM 2.5alpha. */
 #ifdef MAIN_C
 #define WHERE 
@@ -61,6 +84,7 @@
 #define CH_MIME		(1<<9) /* ignore MIME fields */
 #define CH_UPDATE_LEN	(1<<10) /* update Lines: and Content-Length: */
 #define CH_TXTPLAIN	(1<<11) /* generate text/plain MIME headers */
+#define CH_NOLEN	(1<<12) /* don't write Content-Length: and Lines: */
 
 /* flags for mutt_enter_string() */
 #define  M_ALIAS   1      /* do alias "completion" by calling up the alias-menu */
@@ -70,6 +94,7 @@
 #define  M_PASS    (1<<4) /* password mode (no echo) */
 #define  M_CLEAR   (1<<5) /* clear input if printable character is pressed */
 #define  M_COMMAND (1<<6) /* do command completion */
+#define  M_PATTERN (1<<7) /* pattern mode - only used for history classes */
 
 /* flags for mutt_get_token() */
 #define M_TOKEN_EQUAL		1	/* treat '=' as a special */
@@ -88,6 +113,12 @@ typedef struct
   int destroy;	/* destroy `data' when done? */
 } BUFFER;
 
+typedef struct
+{
+  int ch; /* raw key pressed */
+  int op; /* function op */
+} event_t;
+
 /* flags for _mutt_system() */
 #define M_DETACH_PROCESS	1	/* detach subprocess from group */
 
@@ -97,7 +128,10 @@ typedef enum
   M_FORMAT_FORCESUBJ	= (1<<0), /* print the subject even if unchanged */
   M_FORMAT_TREE		= (1<<1), /* draw the thread tree */
   M_FORMAT_MAKEPRINT	= (1<<2), /* make sure that all chars are printable */
-  M_FORMAT_OPTIONAL	= (1<<3)
+  M_FORMAT_OPTIONAL	= (1<<3),
+  M_FORMAT_STAT_FILE	= (1<<4), /* used by mutt_attach_fmt */
+  M_FORMAT_ARROWCURSOR	= (1<<5), /* reserve space for arrow_cursor */
+  M_FORMAT_INDEX	= (1<<6)  /* this is a main index entry */
 } format_flag;
 
 /* types for mutt_add_hook() */
@@ -106,6 +140,9 @@ typedef enum
 #define M_SENDHOOK	(1<<2)
 #define M_FCCHOOK	(1<<3)
 #define M_SAVEHOOK	(1<<4)
+#ifdef _PGPPATH
+#define M_PGPHOOK	(1<<5)
+#endif
 
 /* tree characters for linearize_tree and print_enriched_string */
 #define M_TREE_LLCORNER		1
@@ -119,6 +156,12 @@ typedef enum
 #define M_TREE_HIDDEN		9
 #define M_TREE_MAX		10
 
+#define M_THREAD_COLLAPSE	(1<<0)
+#define M_THREAD_UNCOLLAPSE	(1<<1)
+#define M_THREAD_GET_HIDDEN	(1<<2)
+#define M_THREAD_UNREAD		(1<<3)
+#define M_THREAD_NEXT_UNREAD	(1<<4)
+
 enum
 {
   /* modes for mutt_view_attachment() */
@@ -149,6 +192,7 @@ enum
   M_OR,
   M_TO,
   M_CC,
+  M_COLLAPSED,
   M_SUBJECT,
   M_FROM,
   M_DATE,
@@ -167,13 +211,21 @@ enum
   M_PERSONAL_RECIP,
   M_PERSONAL_FROM,
   M_ADDRESS,
-
+#ifdef _PGPPATH
+  M_PGP_SIGN,
+  M_PGP_ENCRYPT,
+  M_PGP_KEY,
+#endif
+  
   /* Options for Mailcap lookup */
   M_EDIT,
   M_COMPOSE,
   M_PRINT,
   M_AUTOVIEW,
 
+  /* options for socket code */
+  M_NEW_SOCKET,
+
   /* Options for mutt_save_attachment */
   M_SAVE_APPEND
 };
@@ -191,8 +243,10 @@ enum
 enum
 {
 
+#ifdef _PGPPATH
+  OPT_VERIFYSIG, /* verify PGP signatures */
+#endif
 
-  OPT_USEMAILCAP,
   OPT_PRINT,
   OPT_INCLUDE,
   OPT_DELETE,
@@ -216,65 +270,93 @@ enum
 #define SENDBATCH	(1<<5)
 #define SENDMAILX	(1<<6)
 #define SENDKEY		(1<<7)
+#define SENDEDITMSG	(1<<8)
 
 /* boolean vars */
 enum
 {
-  OPTPROMPTAFTER,
-  OPTSTATUSONTOP,
   OPTALLOW8BIT,
-  OPTASCIICHARS,
-  OPTMETOO,
-  OPTEDITHDRS,
   OPTARROWCURSOR,
-  OPTASKCC,
-  OPTHEADER,
-  OPTREVALIAS,
-  OPTREVNAME,
-  OPTFORCENAME,
-  OPTSAVEEMPTY,
-  OPTPAGERSTOP,
-  OPTSIGDASHES,
+  OPTASCIICHARS,
   OPTASKBCC,
+  OPTASKCC,
+  OPTATTACHSPLIT,
   OPTAUTOEDIT,
-  OPTMARKOLD,
-  OPTCONFIRMCREATE,
-  OPTCONFIRMAPPEND,
-  OPTPOPDELETE,
-  OPTSAVENAME,
-  OPTTHOROUGHSRC,
-  OPTTILDE,
-  OPTMARKERS,
-  OPTFCCATTACH,
-  OPTPIPESPLIT,
-  OPTPIPEDECODE,
-  OPTREADONLY,
-  OPTRESOLVE,
-  OPTSTRICTTHREADS,
   OPTAUTOTAG,
   OPTBEEP,
-  OPTHELP,
-  OPTHDRS,
-  OPTWEED,
-  OPTWRAP,
-  OPTCHECKNEW,
-  OPTFASTREPLY,
-  OPTWAITKEY,
-  OPTWRAPSEARCH,
-  OPTIGNORELISTREPLYTO,
-  OPTSAVEADDRESS,
-  OPTSUSPEND,
-  OPTSORTRE,
-  OPTUSEDOMAIN,
-  OPTUSEFROM,
-  OPTUSE8BITMIME,
-  OPTFORWDECODE,
-  OPTMIMEFORWDECODE,
-  OPTFORWQUOTE,
   OPTBEEPNEW,
+  OPTCHECKNEW,
+  OPTCOLLAPSEUNREAD,
+  OPTCONFIRMAPPEND,
+  OPTCONFIRMCREATE,
+  OPTEDITHDRS,
+  OPTFASTREPLY,
+  OPTFCCATTACH,
   OPTFOLLOWUPTO,
+  OPTFORCENAME,
+  OPTFORWDECODE,
+  OPTFORWWEEDHEADER,
+  OPTFORWQUOTE,
+  OPTHDRS,
+  OPTHEADER,
+  OPTHELP,
+  OPTHIDDENHOST,
+  OPTIGNORELISTREPLYTO,
+  OPTIMPLICITAUTOVIEW,
+  OPTMAILCAPSANITIZE,
+  OPTMARKERS,
+  OPTMARKOLD,
   OPTMENUSCROLL,	/* scroll menu instead of implicit next-page */
   OPTMETAKEY,		/* interpret ALT-x as ESC-x */
+  OPTMETOO,
+  OPTMHPURGE,
+  OPTMIMEFORWDECODE,
+  OPTPAGERSTOP,
+  OPTPIPEDECODE,
+  OPTPIPESPLIT,
+  OPTPOPDELETE,
+  OPTPOPLAST,
+  OPTPROMPTAFTER,
+  OPTREADONLY,
+  OPTREPLYSELF,
+  OPTRESOLVE,
+  OPTREVALIAS,
+  OPTREVNAME,
+  OPTSAVEADDRESS,
+  OPTSAVEEMPTY,
+  OPTSAVENAME,
+  OPTSIGDASHES,
+  OPTSORTRE,
+  OPTSTATUSONTOP,
+  OPTSTRICTTHREADS,
+  OPTSUSPEND,
+  OPTTHOROUGHSRC,
+  OPTTILDE,
+  OPTUNCOLLAPSEJUMP,
+  OPTUSE8BITMIME,
+  OPTUSEDOMAIN,
+  OPTUSEFROM,
+  OPTWAITKEY,
+  OPTWEED,
+  OPTWRAP,
+  OPTWRAPSEARCH,
+  OPTWRITEBCC,		/* write out a bcc header? */
+
+  /* PGP options */
+  
+#ifdef _PGPPATH
+  OPTPGPAUTOSIGN,
+  OPTPGPAUTOENCRYPT,
+  OPTPGPLONGIDS,
+  OPTPGPREPLYENCRYPT,
+  OPTPGPREPLYSIGN,
+  OPTPGPENCRYPTSELF,
+  OPTPGPSTRICTENC,
+  OPTFORWDECRYPT,
+#endif
+
+  /* pseudo options */
+
   OPTAUXSORT,		/* (pseudo) using auxillary sort function */
   OPTFORCEREFRESH,	/* (pseudo) refresh even during macros */
   OPTLOCALES,		/* (pseudo) set if user has valid locale definition */
@@ -290,8 +372,13 @@ enum
   OPTFORCEREDRAWPAGER,	/* (pseudo) used to force a redraw in the pager */
   OPTSORTSUBTHREADS,	/* (pseudo) used when $sort_aux changes */
   OPTNEEDRESCORE,	/* (pseudo) set when the `score' command is used */
-
-
+  OPTUSEHEADERDATE,	/* (pseudo) used by edit-message */
+  OPTATTACHMSG,		/* (pseudo) used by attach-message */
+  
+#ifdef _PGPPATH
+  OPTPGPCHECKTRUST,	/* (pseudo) used by pgp_select_key () */
+  OPTDONTHANDLEPGPKEYS,	/* (pseudo) used to extract PGP keys */
+#endif
 
 
 
@@ -354,6 +441,7 @@ typedef struct envelope
   char *real_subj;		/* offset of the real subject */
   char *message_id;
   char *supersedes;
+  char *date;
   LIST *references;		/* message references (in reverse order) */
   LIST *userhdrs;		/* user defined headers */
 } ENVELOPE;
@@ -376,10 +464,12 @@ typedef struct content
   unsigned int binary : 1; /* long lines, or CR not in CRLF pair */
   unsigned int from : 1;   /* has a line beginning with "From "? */
   unsigned int dot : 1;    /* has a line consisting of a single dot? */
+  unsigned int nonasc : 1; /* has unicode characters out of ASCII range */
 } CONTENT;
 
 typedef struct body
 {
+  char *xtype;			/* content-type if x-unknown */
   char *subtype;                /* content-type subtype */
   PARAMETER *parameter;         /* parameters of the content-type */
   char *description;            /* content-description */
@@ -408,6 +498,10 @@ typedef struct body
   struct body *parts;           /* parts of a multipart or message/rfc822 */
   struct header *hdr;		/* header information for message/rfc822 */
 
+  time_t stamp;			/* time stamp of last
+				 * encoding update.
+				 */
+  
   unsigned int type : 3;        /* content-type primary type */
   unsigned int encoding : 3;    /* content-transfer-encoding */
   unsigned int disposition : 2; /* content-disposition */
@@ -423,8 +517,11 @@ typedef struct body
 
 typedef struct header
 {
+#ifdef _PGPPATH
+  unsigned int pgp : 3;
+#endif
+
   unsigned int mime : 1;    /* has a Mime-Version header? */
-  unsigned int mailcap : 1; /* requires mailcap to display? */
   unsigned int flagged : 1; /* marked important? */
   unsigned int tagged : 1;
   unsigned int deleted : 1;
@@ -434,25 +531,14 @@ typedef struct header
   unsigned int read : 1;
   unsigned int expired : 1; /* already expired? */
   unsigned int superseded : 1; /* got superseded? */
-
-
-
-
-
-
-
-
-
-
-
-
-
   unsigned int replied : 1;
   unsigned int subject_changed : 1; /* used for threading */
   unsigned int display_subject : 1; /* used for threading */
   unsigned int fake_thread : 1;     /* no ref matched, but subject did */
   unsigned int threaded : 1;        /* message has been threaded */
-
+  unsigned int recip_valid : 1;  /* is_recipient is valid */
+  unsigned int active : 1;	    /* message is not to be removed */
+  
   /* timezone of the sender of this message */
   unsigned int zhours : 5;
   unsigned int zminutes : 6;
@@ -462,6 +548,13 @@ typedef struct header
   unsigned int searched : 1;
   unsigned int matched : 1;
 
+  /* the following are used to support collapsing threads  */
+  unsigned int collapsed : 1; /* is this message part of a collapsed thread? */
+  unsigned int limited : 1;   /* is this message in a limited view?  */
+  size_t num_hidden;          /* number of hidden messages in this view */
+
+  short recipient;	/* user_is_recipient()'s return value, cached */
+  
   int pair; /* color-pair to use when displaying in the index */
 
   time_t date_sent;     /* time when the message was sent (UTC) */
@@ -499,6 +592,7 @@ typedef struct pattern_t
 {
   short op;
   short not;
+  short alladdr;
   int min;
   int max;
   struct pattern_t *next;
@@ -511,6 +605,7 @@ typedef struct
   char *path;
   FILE *fp;
   time_t mtime;
+  time_t mtime_cur;		/* used with maildir folders */
   off_t size;
   off_t vsize;
   char *pattern;                /* limit pattern string */
@@ -531,7 +626,6 @@ typedef struct
   int msgnotreadyet;		/* which msg "new" in pager, -1 if none */
 #ifdef USE_IMAP
   void *data;			/* driver specific data */
-  int fd;
 #endif /* USE_IMAP */
 
   short magic;			/* mailbox type */
@@ -544,13 +638,17 @@ typedef struct
   unsigned int setgid : 1;
   unsigned int quiet : 1;	/* inhibit status messages? */
   unsigned int revsort : 1;	/* mailbox sorted in reverse? */
+  unsigned int collapsed : 1;   /* are all threads collapsed? */
+  unsigned int closing : 1;	/* mailbox is being closed */
 } CONTEXT;
 
 typedef struct attachptr
 {
   BODY *content;
+  int parent_type;
   char *tree;
   int level;
+  int num;
 } ATTACHPTR;
 
 typedef struct
@@ -566,11 +664,22 @@ typedef struct
 
 
 
+#ifdef _PGPPATH
+#define M_VERIFY	(1<<1) /* perform signature verification */
+#endif
 
 
 
+#define M_PENDINGPREFIX (1<<2) /* prefix to write, but character must follow */
+#define M_WEED          (1<<3) /* weed headers even when not in display mode */
+
+#define state_set_prefix(s) ((s)->flags |= M_PENDINGPREFIX)
+#define state_reset_prefix(s) ((s)->flags &= ~M_PENDINGPREFIX)
 #define state_puts(x,y) fputs(x,(y)->fpout)
 #define state_putc(x,y) fputc(x,(y)->fpout)
 
+void state_prefix_putc(char, STATE *);
+int  state_printf(STATE *, const char *, ...);
+
 #include "protos.h"
 #include "globals.h"
