@@ -29,15 +29,18 @@
 #include <sys/stat.h>
 
 static struct mapping_t PostponeHelp[] = {
-  { "Exit",  OP_EXIT },
-  { "Del",   OP_DELETE },
-  { "Undel", OP_UNDELETE },
-  { "Help",  OP_HELP },
+  { N_("Exit"),  OP_EXIT },
+  { N_("Del"),   OP_DELETE },
+  { N_("Undel"), OP_UNDELETE },
+  { N_("Help"),  OP_HELP },
   { NULL }
 };
 
 
 
+#ifdef _PGPPATH
+#include "pgp.h"
+#endif /* _PGPPATH */
 
 
 
@@ -91,7 +94,8 @@ static void post_entry (char *s, size_t slen, MUTTMENU *menu, int entry)
 {
   CONTEXT *ctx = (CONTEXT *) menu->data;
 
-  mutt_make_string (s, slen, NONULL (HdrFmt), ctx, ctx->hdrs[entry]);
+  _mutt_make_string (s, slen, NONULL (HdrFmt), ctx, ctx->hdrs[entry],
+		     M_FORMAT_ARROWCURSOR);
 }
 
 static HEADER *select_msg (void)
@@ -104,7 +108,7 @@ static HEADER *select_msg (void)
   menu->make_entry = post_entry;
   menu->menu = MENU_POST;
   menu->max = PostContext->msgcount;
-  menu->title = "Postponed Messages";
+  menu->title = _("Postponed Messages");
   menu->data = PostContext;
   menu->help = mutt_compile_help (helpstr, sizeof (helpstr), MENU_POST, PostponeHelp);
 
@@ -153,21 +157,21 @@ static HEADER *select_msg (void)
  *	hdr	envelope/attachment info for recalled message
  *	cur	if message was a reply, `cur' is set to the message which
  *		`hdr' is in reply to
+ *	fcc	fcc for the recalled message
+ *	fcclen	max length of fcc
  *
  * return vals:
  *	-1		error/no messages
  *	0		normal exit
  *	SENDREPLY	recalled message is a reply
  */
-int mutt_get_postponed (CONTEXT *ctx, HEADER *hdr, HEADER **cur)
+int mutt_get_postponed (CONTEXT *ctx, HEADER *hdr, HEADER **cur, char *fcc, size_t fcclen)
 {
   HEADER *h;
-  MESSAGE *msg;
   int code = SENDPOSTPONED;
   LIST *tmp;
   LIST *last = NULL;
   LIST *next;
-  char file[_POSIX_PATH_MAX];
   char *p;
   int opt_delete;
 
@@ -177,7 +181,7 @@ int mutt_get_postponed (CONTEXT *ctx, HEADER *hdr, HEADER **cur)
   if ((PostContext = mx_open_mailbox (Postponed, M_NOSORT, NULL)) == NULL)
   {
     PostCount = 0;
-    mutt_error ("No postponed messages.");
+    mutt_error _("No postponed messages.");
     return (-1);
   }
   
@@ -186,7 +190,7 @@ int mutt_get_postponed (CONTEXT *ctx, HEADER *hdr, HEADER **cur)
     PostCount = 0;
     mx_close_mailbox (PostContext);
     safe_free ((void **) &PostContext);
-    mutt_error ("No postponed messages.");
+    mutt_error _("No postponed messages.");
     return (-1);
   }
 
@@ -202,78 +206,13 @@ int mutt_get_postponed (CONTEXT *ctx, HEADER *hdr, HEADER **cur)
     return (-1);
   }
 
-  if ((msg = mx_open_message (PostContext, h->msgno)) == NULL)
+  if (mutt_prepare_edit_message (PostContext, hdr, h) < 0)
   {
-    mx_close_mailbox (PostContext);
-    safe_free ((void **) &PostContext);
-    return (-1);
-  }
-
-  fseek (msg->fp, h->offset, 0);
-  hdr->env = mutt_read_rfc822_header (msg->fp, NULL);
-
-  if (h->content->type == TYPEMESSAGE || h->content->type == TYPEMULTIPART)
-  {
-    BODY *b;
-
-    fseek (msg->fp, h->content->offset, 0);
-
-    if (h->content->type == TYPEMULTIPART)
-    {
-      h->content->parts = mutt_parse_multipart (msg->fp, 
-	       mutt_get_parameter ("boundary", h->content->parameter),
-	       h->content->offset + h->content->length,
-	       strcasecmp ("digest", h->content->subtype) == 0);
-    }
-    else
-      h->content->parts = mutt_parse_messageRFC822 (msg->fp, h->content);
-
-    /* Now that we know what was in the other message, convert to the new
-     * message.
-     */
-    hdr->content = h->content->parts;
-    b = h->content->parts;
-    while (b != NULL)
-    {
-      file[0] = '\0';
-      if (b->filename)
-	strfcpy (file, b->filename, sizeof (file));
-      mutt_adv_mktemp (file);
-      if (mutt_save_attachment (msg->fp, b, file, 0) == -1)
-      {
-	mutt_free_envelope (&hdr->env);
-	mutt_free_body (&hdr->content);
-	mx_close_message (&msg);
-	mx_fastclose_mailbox (PostContext);
-	safe_free ((void **) &PostContext);
-	return (-1);
-      }
-      safe_free ((void *) &b->filename);
-      b->filename = safe_strdup (file);
-      b->unlink = 1;
-      mutt_free_body (&b->parts);
-      b = b->next;
-    }
-    h->content->parts = NULL;
-  }
-  else
-  {
-    mutt_mktemp (file);
-    if (mutt_save_attachment (msg->fp, h->content, file, 0) == -1)
-    {
-      mutt_free_envelope (&hdr->env);
-      mx_close_message (&msg);
       mx_fastclose_mailbox (PostContext);
       safe_free ((void **) &PostContext);
       return (-1);
-    }
-    hdr->content = mutt_make_attach (file);
-    hdr->content->use_disp = 0;	/* no content-disposition */
-    hdr->content->unlink = 1;	/* delete when we are done */
   }
 
-  mx_close_message (&msg);
-
   /* finished with this message, so delete it. */
   mutt_set_flag (PostContext, h, M_DELETE, 1);
 
@@ -290,7 +229,7 @@ int mutt_get_postponed (CONTEXT *ctx, HEADER *hdr, HEADER **cur)
 
   for (tmp = hdr->env->userhdrs; tmp; )
   {
-    if (strncasecmp ("X-Mutt-References:", tmp->data, 18) == 0)
+    if (mutt_strncasecmp ("X-Mutt-References:", tmp->data, 18) == 0)
     {
       if (ctx)
       {
@@ -313,9 +252,42 @@ int mutt_get_postponed (CONTEXT *ctx, HEADER *hdr, HEADER **cur)
       if (*cur)
 	code |= SENDREPLY;
     }
+    else if (mutt_strncasecmp ("X-Mutt-Fcc:", tmp->data, 11) == 0)
+    {
+      p = tmp->data + 11;
+      SKIPWS (p);
+      strfcpy (fcc, p, fcclen);
+      mutt_pretty_mailbox (fcc);
+
+      /* remove the X-Mutt-Fcc: header field */
+      next = tmp->next;
+      if (last)
+	last->next = tmp->next;
+      else
+	hdr->env->userhdrs = tmp->next;
+      tmp->next = NULL;
+      mutt_free_list (&tmp);
+      tmp = next;
+    }
 
 
 
+#ifdef _PGPPATH
+    else if (mutt_strncmp ("Pgp:", tmp->data, 4) == 0)
+    {
+      hdr->pgp = mutt_parse_pgp_hdr (tmp->data+4, 1);
+       
+      /* remove the pgp field */
+      next = tmp->next;
+      if (last)
+	last->next = tmp->next;
+      else
+	hdr->env->userhdrs = tmp->next;
+      tmp->next = NULL;
+      mutt_free_list (&tmp);
+      tmp = next;
+    }
+#endif /* _PGPPATH */
 
 
 
@@ -330,3 +302,162 @@ int mutt_get_postponed (CONTEXT *ctx, HEADER *hdr, HEADER **cur)
 
 
 
+#ifdef _PGPPATH
+
+int mutt_parse_pgp_hdr (char *p, int set_signas)
+{
+  int pgp = 0;
+  char pgp_sign_as[LONG_STRING] = "\0", *q;
+  char pgp_sign_micalg[LONG_STRING] = "\0";
+   
+  SKIPWS (p);
+  for (; *p; p++)
+  {    
+     
+    switch (*p)
+    {
+      case 'e':
+      case 'E':
+        pgp |= PGPENCRYPT;
+        break;
+
+      case 's':    
+      case 'S':
+        pgp |= PGPSIGN;
+        q = pgp_sign_as;
+      
+        if (*(p+1) == '<')
+        {
+          for (p += 2; 
+	       *p && *p != '>' && q < pgp_sign_as + sizeof (pgp_sign_as) - 1;
+               *q++ = *p++)
+	    ;
+
+          if (*p!='>')
+          {
+            mutt_error _("Illegal PGP header");
+            return 0;
+          }
+        }
+       
+        *q = '\0';
+        break;
+
+      case 'm':
+      case 'M':
+   	q = pgp_sign_micalg;
+	
+        if(*(p+1) == '<')
+	{
+	  for(p += 2; *p && *p != '>' && q < pgp_sign_micalg + sizeof(pgp_sign_micalg) - 1;
+	      *q++ = *p++)
+	    ;
+	  
+	  if(*p != '>')
+	  {
+	    mutt_error _("Illegal PGP header");
+	    return 0;
+	  }
+	}
+
+	*q = '\0';
+	break;
+	  
+      default:
+        mutt_error _("Illegal PGP header");
+        return 0;
+    }
+     
+  }
+ 
+  if (set_signas || *pgp_sign_as)
+  {
+    safe_free((void **) &PgpSignAs);
+    PgpSignAs = safe_strdup(pgp_sign_as);
+  }
+
+  /* the micalg field must not be empty */
+  if (set_signas && *pgp_sign_micalg)
+  {
+    safe_free((void **) &PgpSignMicalg);
+    PgpSignMicalg = safe_strdup(pgp_sign_micalg);
+  }
+
+  return pgp;
+}
+#endif /* _PGPPATH */
+
+
+
+int mutt_prepare_edit_message (CONTEXT *ctx, HEADER *newhdr, HEADER *hdr)
+{
+  MESSAGE *msg = mx_open_message (ctx, hdr->msgno);
+  char file[_POSIX_PATH_MAX];
+
+  if (msg == NULL)
+    return (-1);
+
+  fseek (msg->fp, hdr->offset, 0);
+  newhdr->env = mutt_read_rfc822_header (msg->fp, newhdr, 1);
+
+  if (hdr->content->type == TYPEMESSAGE || hdr->content->type == TYPEMULTIPART)
+  {
+    BODY *b;
+
+    mutt_parse_part (msg->fp, hdr->content);
+
+    /* Now that we know what was in the other message, convert to the new
+     * message.
+     */
+    newhdr->content = hdr->content->parts;
+    b = hdr->content->parts;
+    while (b != NULL)
+    {
+      file[0] = '\0';
+      if (b->filename)
+	strfcpy (file, b->filename, sizeof (file));
+      else
+	/* avoid Content-Disposition: header with temporary filename */
+	b->use_disp = 0;
+      mutt_adv_mktemp (file, sizeof(file));
+      if (mutt_save_attachment (msg->fp, b, file, 0, NULL) == -1)
+      {
+	mutt_free_envelope (&newhdr->env);
+	mutt_free_body (&newhdr->content);
+	mx_close_message (&msg);
+	return (-1);
+      }
+      safe_free ((void *) &b->filename);
+      b->filename = safe_strdup (file);
+      b->unlink = 1;
+      mutt_stamp_attachment (b);
+      mutt_free_body (&b->parts);
+      b = b->next;
+    }
+    hdr->content->parts = NULL;
+  }
+  else
+  {
+    mutt_mktemp (file);
+    if (mutt_save_attachment (msg->fp, hdr->content, file, 0, NULL) == -1)
+    {
+      mutt_free_envelope (&newhdr->env);
+      mx_close_message (&msg);
+      return (-1);
+    }
+    newhdr->content = mutt_make_file_attach (file);
+    
+    FREE (&newhdr->content->subtype);
+    FREE (&newhdr->content->xtype);
+    
+    newhdr->content->type = hdr->content->type;
+    newhdr->content->xtype = safe_strdup (hdr->content->xtype);
+    newhdr->content->subtype = safe_strdup (hdr->content->subtype);
+    
+    newhdr->content->use_disp = 0;	/* no content-disposition */
+    newhdr->content->unlink = 1;	/* delete when we are done */
+  }
+
+  mx_close_message (&msg);
+  return 0;
+}
