@@ -38,58 +38,81 @@
 /* The command semantics include the following:
  * %s is the filename that contains the mail body data
  * %t is the content type, like text/plain
+ * %{parameter} is replaced by the parameter value from the content-type field
  * \% is %
  * Unsupported rfc1524 parameters: these would probably require some doing
  * by mutt, and can probably just be done by piping the message to metamail
- * %{parameter} is replaced by the parameter value from the content-type field
  * %n is the integer number of sub-parts in the multipart
  * %F is "content-type filename" repeated for each sub-part
  *
  * In addition, this function returns a 0 if the command works on a file,
  * and 1 if the command works on a pipe.
  */
-int rfc1524_expand_command (BODY *a, char *filename, char *type, 
+int rfc1524_expand_command (BODY *a, char *filename, char *_type,
     char *command, int clen)
 {
   int x=0,y=0;
   int needspipe = TRUE;
   char buf[LONG_STRING];
+  char type[LONG_STRING];
+  
+  strfcpy (type, _type, sizeof (type));
+  
+  if (option (OPTMAILCAPSANITIZE))
+    mutt_sanitize_filename (type, 0);
 
-  while (command[x] && x<clen && y<sizeof(buf)) {
+  while (command[x] && x<clen && y<sizeof(buf)) 
+  {
     if (command[x] == '\\') {
       x++;
       buf[y++] = command[x++];
     }
-    else if (command[x] == '%') {
+    else if (command[x] == '%') 
+    {
       x++;
-      if (command[x] == '{') {
+      if (command[x] == '{') 
+      {
 	char param[STRING];
+	char pvalue[LONG_STRING];
 	int z = 0;
 	char *ret = NULL;
+	char *pv;
 
 	x++;
 	while (command[x] && command[x] != '}' && z<sizeof(param))
 	  param[z++] = command[x++];
 	param[z] = '\0';
 	dprint(2,(debugfile,"Parameter: %s  Returns: %s\n",param,ret));
-	ret = mutt_get_parameter(param,a->parameter);
+	pv = mutt_get_parameter (param, a->parameter);
+	strfcpy (pvalue, NONULL(pv), sizeof (pvalue));
+	if (option (OPTMAILCAPSANITIZE)) 
+	  mutt_sanitize_filename (pvalue, 0);
+	ret = mutt_quote_filename (pvalue);
 	dprint(2,(debugfile,"Parameter: %s  Returns: %s\n",param,ret));
 	z = 0;
 	while (ret && ret[z] && y<sizeof(buf))
 	  buf[y++] = ret[z++];
+	FREE(&ret);
       }
       else if (command[x] == 's' && filename != NULL)
       {
-	char *fn = filename;
-
-	while (*fn && y < sizeof (buf))
-	  buf[y++] = *fn++;
+	char *fn = mutt_quote_filename(filename);
+	int i;
+	
+	for(i = 0; fn[i] && y < sizeof(buf); i++)
+	  buf[y++] = fn[i];
+	
+	FREE(&fn);
 	needspipe = FALSE;
       }
       else if (command[x] == 't')
       {
-	while (*type && y < sizeof (buf))
-	  buf[y++] = *type++;
+	char *t = mutt_quote_filename (type);
+	char *s;
+	
+	for (s = t; *s && y < sizeof (buf);)
+	  buf[y++] = *s++;
+	FREE (&t);
       }
       x++;
     }
@@ -138,7 +161,8 @@ static int get_field_text (char *field, char **entry,
   {
     if (entry)
     {
-      field = mutt_skip_whitespace (++field);
+      field++;
+      field = mutt_skip_whitespace (field);
       safe_free ((void **) entry);
       *entry = safe_strdup (field);
     }
@@ -146,7 +170,7 @@ static int get_field_text (char *field, char **entry,
   }
   else 
   {
-    mutt_error ("Improperly formated entry for type %s in \"%s\" line %d",
+    mutt_error (_("Improperly formated entry for type %s in \"%s\" line %d"),
 		type, filename, line);
     return 0;
   }
@@ -198,10 +222,10 @@ static int rfc1524_mailcap_parse (BODY *a,
 
       /* check type */
       ch = get_field (buf);
-      if (strcasecmp (buf, type) &&
-	  (strncasecmp (buf, type, btlen) ||
+      if (mutt_strcasecmp (buf, type) &&
+	  (mutt_strncasecmp (buf, type, btlen) ||
 	   (buf[btlen] != 0 &&			/* implicit wild */
-	    strcmp (buf + btlen, "/*"))))	/* wildsubtype */
+	    mutt_strcmp (buf + btlen, "/*"))))	/* wildsubtype */
 	continue;
 
       /* next field is the viewcommand */
@@ -223,51 +247,53 @@ static int rfc1524_mailcap_parse (BODY *a,
 	ch = get_field (ch);
 	dprint (2, (debugfile, "field: %s\n", field));
 
-	if (!strcasecmp (field, "needsterminal"))
+	if (!mutt_strcasecmp (field, "needsterminal"))
 	{
 	  if (entry)
 	    entry->needsterminal = TRUE;
 	}
-	else if (!strcasecmp (field, "copiousoutput"))
+	else if (!mutt_strcasecmp (field, "copiousoutput"))
 	{
 	  copiousoutput = TRUE;
 	  if (entry)
 	    entry->copiousoutput = TRUE;
 	}
-	else if (!strncasecmp (field, "composetyped", 12))
+	else if (!mutt_strncasecmp (field, "composetyped", 12))
 	{
 	  /* this compare most occur before compose to match correctly */
-	  /* this isn't particularly well specified in the rfc, so skip it */
+	  if (get_field_text (field + 12, entry ? &entry->composetypecommand : NULL,
+			      type, filename, line))
+	    composecommand = TRUE;
 	}
-	else if (!strncasecmp (field, "compose", 7))
+	else if (!mutt_strncasecmp (field, "compose", 7))
 	{
 	  if (get_field_text (field + 7, entry ? &entry->composecommand : NULL,
 			      type, filename, line))
 	    composecommand = TRUE;
 	}
-	else if (!strncasecmp (field, "print", 5))
+	else if (!mutt_strncasecmp (field, "print", 5))
 	{
 	  if (get_field_text (field + 5, entry ? &entry->printcommand : NULL,
 			      type, filename, line))
 	    printcommand = TRUE;
 	}
-	else if (!strncasecmp (field, "edit", 4))
+	else if (!mutt_strncasecmp (field, "edit", 4))
 	{
 	  if (get_field_text (field + 4, entry ? &entry->editcommand : NULL,
 			      type, filename, line))
 	    editcommand = TRUE;
 	}
-	else if (!strncasecmp (field, "nametemplate", 12))
+	else if (!mutt_strncasecmp (field, "nametemplate", 12))
 	{
 	  get_field_text (field + 12, entry ? &entry->nametemplate : NULL,
 			  type, filename, line);
 	}
-	else if (!strncasecmp (field, "x-convert", 9))
+	else if (!mutt_strncasecmp (field, "x-convert", 9))
 	{
 	  get_field_text (field + 9, entry ? &entry->convert : NULL,
 			  type, filename, line);
 	}
-	else if (!strncasecmp (field, "test", 4))
+	else if (!mutt_strncasecmp (field, "test", 4))
 	{
 	  /* 
 	   * This routine executes the given test command to determine
@@ -279,7 +305,7 @@ static int rfc1524_mailcap_parse (BODY *a,
 	  if (get_field_text (field + 4, &test_command, type, filename, line)
 	      && test_command)
 	  {
-	    len = strlen (test_command) + STRING;
+	    len = mutt_strlen (test_command) + STRING;
 	    safe_realloc ((void **) &test_command, len);
 	    rfc1524_expand_command (a, NULL, type, test_command, len);
 	    if (mutt_system (test_command))
@@ -287,7 +313,7 @@ static int rfc1524_mailcap_parse (BODY *a,
 	      /* a non-zero exit code means test failed */
 	      found = FALSE;
 	    }
-	    free (test_command);
+	    FREE (&test_command);
 	  }
 	}
       } /* while (ch) */
@@ -320,6 +346,7 @@ static int rfc1524_mailcap_parse (BODY *a,
 	{
 	  safe_free ((void **) &entry->command);
 	  safe_free ((void **) &entry->composecommand);
+	  safe_free ((void **) &entry->composetypecommand);
 	  safe_free ((void **) &entry->editcommand);
 	  safe_free ((void **) &entry->printcommand);
 	  safe_free ((void **) &entry->nametemplate);
@@ -335,14 +362,9 @@ static int rfc1524_mailcap_parse (BODY *a,
   return found;
 }
 
-rfc1524_entry *rfc1524_new_entry()
+rfc1524_entry *rfc1524_new_entry(void)
 {
-  rfc1524_entry *tmp;
-
-  tmp = (rfc1524_entry *)safe_malloc(sizeof(rfc1524_entry));
-  memset(tmp,0,sizeof(rfc1524_entry));
-
-  return tmp;
+  return (rfc1524_entry *)safe_calloc(1, sizeof(rfc1524_entry));
 }
 
 void rfc1524_free_entry(rfc1524_entry **entry)
@@ -352,6 +374,7 @@ void rfc1524_free_entry(rfc1524_entry **entry)
   safe_free((void **)&p->command);
   safe_free((void **)&p->testcommand);
   safe_free((void **)&p->composecommand);
+  safe_free((void **)&p->composetypecommand);
   safe_free((void **)&p->editcommand);
   safe_free((void **)&p->printcommand);
   safe_free((void **)&p->nametemplate);
@@ -377,9 +400,9 @@ int rfc1524_mailcap_lookup (BODY *a, char *type, rfc1524_entry *entry, int opt)
    * and overriden by the MAILCAPS environment variable, and, just to be nice, 
    * we'll make it specifiable in .muttrc
    */
-  if (!*curr)
+  if (!curr || !*curr)
   {
-    mutt_error ("No mailcap path specified");
+    mutt_error _("No mailcap path specified");
     return 0;
   }
 
@@ -405,7 +428,7 @@ int rfc1524_mailcap_lookup (BODY *a, char *type, rfc1524_entry *entry, int opt)
   }
 
   if (entry && !found)
-    mutt_error ("mailcap entry for type %s not found", type);
+    mutt_error (_("mailcap entry for type %s not found"), type);
 
   return found;
 }
@@ -416,156 +439,171 @@ int rfc1524_mailcap_lookup (BODY *a, char *type, rfc1524_entry *entry, int opt)
  * Renamed to mutt_adv_mktemp so I only have to change where it's
  * called, and not all possible cases.
  */
-void mutt_adv_mktemp (char *s)
+void mutt_adv_mktemp (char *s, size_t l)
 {
   char buf[_POSIX_PATH_MAX];
   char tmp[_POSIX_PATH_MAX];
   char *period;
-
-  strfcpy (buf, Tempdir, sizeof (buf));
+  size_t sl;
+  
+  strfcpy (buf, NONULL (Tempdir), sizeof (buf));
   mutt_expand_path (buf, sizeof (buf));
   if (s[0] == '\0')
   {
-    sprintf (s, "%s/muttXXXXXX", buf);
+    snprintf (s, l, "%s/muttXXXXXX", buf);
     mktemp (s);
   }
   else
   {
     strfcpy (tmp, s, sizeof (tmp));
-    sprintf (s, "%s/%s", buf, tmp);
+    snprintf (s, l, "%s/%s", buf, tmp);
     if (access (s, F_OK) != 0)
       return;
     if ((period = strrchr (tmp, '.')) != NULL)
       *period = 0;
-    sprintf (s, "%s/%s.XXXXXX", buf, tmp);
+    snprintf (s, l, "%s/%s.XXXXXX", buf, tmp);
     mktemp (s);
     if (period != NULL)
     {
       *period = '.';
-      strcat (s, period);
+      sl = mutt_strlen(s);
+      strfcpy(s + sl, period, l - sl);
     }
   }
 }
 
-/* This routine expands the filename given to match the format of the
- * nametemplate given.  It returns various values based on what operations
- * it performs.
- *
+/* This routine will create a _temporary_ filename matching the
+ * name template given if this needs to be done.
+ * 
+ * Please note that only the last path element of the
+ * template and/or the old file name will be used for the
+ * comparison and the temporary file name.
+ * 
  * Returns 0 if oldfile is fine as is.
  * Returns 1 if newfile specified
  */
 
+static void strnfcpy(char *d, char *s, size_t siz, size_t len)
+{
+  if(len > siz)
+    len = siz - 1;
+  strfcpy(d, s, len);
+}
+
 int rfc1524_expand_filename (char *nametemplate,
 			     char *oldfile, 
 			     char *newfile,
 			     size_t nflen)
 {
-  int z = 0;
-  int i = 0, j = 0;
-  int lmatch = TRUE;
-  int match = TRUE;
-  size_t len = 0;
+  int i, j, k, ps, r;
   char *s;
-
+  short lmatch = 0, rmatch = 0; 
+  char left[_POSIX_PATH_MAX];
+  char right[_POSIX_PATH_MAX];
+  
   newfile[0] = 0;
 
+  /* first, ignore leading path components.
+   */
+  
   if (nametemplate && (s = strrchr (nametemplate, '/')))
     nametemplate = s + 1;
 
   if (oldfile && (s = strrchr (oldfile, '/')))
-  {
-    len = s - oldfile + 1;
-    if (len > nflen)
-      len = nflen;
-    strfcpy (newfile, oldfile, len + 1);
-    oldfile += len;
-  }
-
-  /* If nametemplate is NULL, create a newfile from oldfile and return 0 */
+    oldfile = s + 1;
+    
   if (!nametemplate)
   {
     if (oldfile)
       strfcpy (newfile, oldfile, nflen);
-    mutt_adv_mktemp (newfile);
-    return 0;
   }
-
-  /* If oldfile is NULL, just return a newfile name */
-  if (!oldfile)
+  else if (!oldfile)
   {
     snprintf (newfile, nflen, nametemplate, "mutt");
-    mutt_adv_mktemp (newfile);
+  }
+  else /* oldfile && nametemplate */
+  {
+
+    /* first, compare everything left from the "%s" 
+     * (if there is one).
+     */
+    
+    lmatch = 1; ps = 0;
+    for(i = 0; nametemplate[i]; i++)
+    {
+      if(nametemplate[i] == '%' && nametemplate[i+1] == 's')
+      { 
+	ps = 1;
+	break;
+      }
+
+      /* note that the following will _not_ read beyond oldfile's end. */
+
+      if(lmatch && nametemplate[i] != oldfile[i])
+	lmatch = 0;
+    }
+
+    if(ps)
+    {
+      
+      /* If we had a "%s", check the rest. */
+      
+      /* now, for the right part: compare everything right from 
+       * the "%s" to the final part of oldfile.
+       * 
+       * The logic here is as follows:
+       * 
+       * - We start reading from the end.
+       * - There must be a match _right_ from the "%s",
+       *   thus the i + 2.  
+       * - If there was a left hand match, this stuff
+       *   must not be counted again.  That's done by the
+       *   condition (j >= (lmatch ? i : 0)).
+       */
+      
+      rmatch = 1;
+
+      for(r = 0, j = mutt_strlen(oldfile) - 1, k = mutt_strlen(nametemplate) - 1 ;
+	  j >= (lmatch ? i : 0) && k >= i + 2;
+	  j--, k--)
+      {
+	if(nametemplate[k] != oldfile[j])
+	{
+	  rmatch = 0;
+	  break;
+	}
+      }
+      
+      /* Now, check if we had a full match. */
+      
+      if(k >= i + 2)
+	rmatch = 0;
+      
+      if(lmatch) *left = 0;
+      else strnfcpy(left, nametemplate, sizeof(left), i);
+      
+      if(rmatch) *right = 0;
+      else strfcpy(right, nametemplate + i + 2, sizeof(right));
+      
+      snprintf(newfile, nflen, "%s%s%s", left, oldfile, right);
+    }
+    else
+    {
+      /* no "%s" in the name template. */
+      strfcpy(newfile, nametemplate, nflen);
+    }
+  }
+  
+  mutt_adv_mktemp(newfile, nflen);
+
+  if(rmatch && lmatch)
     return 0;
-  }
-
-  /* Next, attempt to determine if the oldfile already matches nametemplate */
-  /* Nametemplate is of the form pre%spost, only replace pre or post if
-   * they don't already match the oldfilename */
-  /* Test pre */
-
-  if ((s = strrchr (nametemplate, '%')) != NULL)
-  {
-    newfile[len] = '\0';
-
-    z = s - nametemplate;
-
-    for (i = 0; i < z && i < nflen; i++)
-    {
-      if (oldfile[i] != nametemplate[i])
-      {
-	lmatch=FALSE;
-	break;
-      }
-    }
-
-    if (!lmatch)
-    {
-      match = FALSE;
-      i = nflen - len;
-      if (i > z)
-	i = z;
-      strfcpy (newfile + len, nametemplate, i);
-    }
-
-    strfcpy (newfile + strlen (newfile), 
-	    oldfile, nflen - strlen (newfile));
-
-    dprint (1, (debugfile,"template: %s, oldfile: %s, newfile: %s\n",
-	      nametemplate, oldfile, newfile));
-
-    /* test post */
-    lmatch = TRUE;
-
-    for (z += 2, i = strlen (oldfile) - 1, j = strlen (nametemplate) - 1; 
-	i && j > z; i--, j--)
-      if (oldfile[i] != nametemplate[j])
-      {
-	lmatch = FALSE;
-	break;
-      }
-
-    if (!lmatch)
-    {
-      match = FALSE;
-      strfcpy (newfile + strlen (newfile),
-	      nametemplate + z, nflen - strlen (newfile));
-    }
-
-    if (match) 
-      return 0;
-
+  else 
     return 1;
-  }
-  else
-  {
-    /* no %s in nametemplate, graft unto path of oldfile */
-    strfcpy (newfile, nametemplate, nflen);
-    return 1;
-  }
+  
 }
 
-/* For nametemplate support, we may need to rename a file.
- * If rfc1524_expand_command() is used on a recv'd message, then
+/* If rfc1524_expand_command() is used on a recv'd message, then
  * the filename doesn't exist yet, but if its used while sending a message,
  * then we need to rename the existing file.
  *
