@@ -16,30 +16,29 @@
  *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */ 
 
-#include <ctype.h>
 #include <string.h>
+#include <ctype.h>
 #include <stdlib.h>
-#include <stdio.h>
 
+#ifndef TESTING
+#include "mutt.h"
+#else
+#define safe_strdup strdup
+#define safe_malloc malloc
+#define SKIPWS(x) while(isspace(*x))x++
+#define FREE(x) safe_free(x)
+#define ISSPACE isspace
+#define strfcpy(a,b,c) {if (c) {strncpy(a,b,c);a[c-1]=0;}}
+#define STRING 128
 #include "rfc822.h"
-
-#define strfcpy(a,b,c) if (c) {strncpy(a,b,c);a[c-1]=0;}
-
-#define STRDUP(a) (a && a[0]) ? strdup (a) : NULL
-
-#ifndef SKIPWS
-#define SKIPWS(c) while(*c && isspace((unsigned char)*c))c++
 #endif
 
-#ifndef NULL
-#define NULL 0
-#endif
-
-static const char rfc822_specials[] = "@.,:;<>[]\\\"()";
-#define is_special(x) strchr(rfc822_specials,x)
+const char RFC822Specials[] = "@.,:;<>[]\\\"()";
+#define is_special(x) strchr(RFC822Specials,x)
 
 int RFC822Error = 0;
 
+/* these must defined in the same order as the numerated errors given in rfc822.h */
 const char *RFC822Errors[] = {
   "out of memory",
   "mismatched parenthesis",
@@ -80,14 +79,11 @@ void rfc822_free_address (ADDRESS **p)
     t = *p;
     *p = (*p)->next;
 #ifdef EXACT_ADDRESS
-    if (t->val)
-      free (t->val);
+    FREE (&t->val);
 #endif
-    if (t->personal)
-      free (t->personal);
-    if (t->mailbox)
-      free (t->mailbox);
-    free (t);
+    FREE (&t->personal);
+    FREE (&t->mailbox);
+    FREE (&t);
   }
 }
 
@@ -163,7 +159,7 @@ next_token (const char *s, char *token, size_t *tokenlen, size_t tokenmax)
   }
   while (*s)
   {
-    if (isspace ((unsigned char)*s) || is_special (*s))
+    if (ISSPACE (*s) || is_special (*s))
       break;
     if (*tokenlen < tokenmax)
       token[(*tokenlen)++] = *s;
@@ -203,36 +199,34 @@ parse_mailboxdomain (const char *s, const char *nonspecial,
 
 static const char *
 parse_address (const char *s,
+               char *token, size_t *tokenlen, size_t tokenmax,
 	       char *comment, size_t *commentlen, size_t commentmax,
 	       ADDRESS *addr)
 {
-  char token[128];
-  size_t tokenlen = 0;
-
   s = parse_mailboxdomain (s, ".\"(\\",
-			   token, &tokenlen, sizeof (token) - 1,
+			   token, tokenlen, tokenmax,
 			   comment, commentlen, commentmax);
   if (!s)
     return NULL;
 
   if (*s == '@')
   {
-    if (tokenlen < sizeof (token) - 1)
-      token[tokenlen++] = '@';
+    if (*tokenlen < tokenmax)
+      token[(*tokenlen)++] = '@';
     s = parse_mailboxdomain (s + 1, ".([]\\",
-			     token, &tokenlen, sizeof (token) - 1,
+			     token, tokenlen, tokenmax,
 			     comment, commentlen, commentmax);
     if (!s)
       return NULL;
   }
 
-  token[tokenlen] = 0;
-  addr->mailbox = STRDUP (token);
+  token[*tokenlen] = 0;
+  addr->mailbox = safe_strdup (token);
 
   if (*commentlen && !addr->personal)
   {
     comment[*commentlen] = 0;
-    addr->personal = STRDUP (comment);
+    addr->personal = safe_strdup (comment);
   }
 
   return s;
@@ -243,30 +237,37 @@ parse_route_addr (const char *s,
 		  char *comment, size_t *commentlen, size_t commentmax,
 		  ADDRESS *addr)
 {
+  char token[STRING];
+  size_t tokenlen = 0;
+
   SKIPWS (s);
 
   /* find the end of the route */
   if (*s == '@')
   {
-    char token[128];
-    size_t tokenlen = 0;
-
     while (s && *s == '@')
-      s = parse_mailboxdomain (s + 1, ".[](\\", token,
+    {
+      if (tokenlen < sizeof (token) - 1)
+	token[tokenlen++] = '@';
+      s = parse_mailboxdomain (s + 1, ".\\[](", token,
 			       &tokenlen, sizeof (token) - 1,
 			       comment, commentlen, commentmax);
+    }
     if (!s || *s != ':')
     {
       RFC822Error = ERR_BAD_ROUTE;
       return NULL; /* invalid route */
     }
+
+    if (tokenlen < sizeof (token) - 1)
+      token[tokenlen++] = ':';
     s++;
   }
 
-  if ((s = parse_address (s, comment, commentlen, commentmax, addr)) == NULL)
+  if ((s = parse_address (s, token, &tokenlen, sizeof (token) - 1, comment, commentlen, commentmax, addr)) == NULL)
     return NULL;
 
-  if (*s != '>')
+  if (*s != '>' || !addr->mailbox)
   {
     RFC822Error = ERR_BAD_ROUTE_ADDR;
     return NULL;
@@ -281,7 +282,10 @@ parse_addr_spec (const char *s,
 		 char *comment, size_t *commentlen, size_t commentmax,
 		 ADDRESS *addr)
 {
-  s = parse_address (s, comment, commentlen, commentmax, addr);
+  char token[STRING];
+  size_t tokenlen = 0;
+
+  s = parse_address (s, token, &tokenlen, sizeof (token) - 1, comment, commentlen, commentmax, addr);
   if (s && *s && *s != ',' && *s != ';')
   {
     RFC822Error = ERR_BAD_ADDR_SPEC;
@@ -304,37 +308,12 @@ add_addrspec (ADDRESS **top, ADDRESS **last, const char *phrase,
   else
     *top = cur;
   *last = cur;
-  
-  if (*commentlen)
-  {
-    comment[*commentlen] = 0;
-    cur->personal = STRDUP (comment);
-  }
 }
 
-#ifdef EXACT_ADDRESS
-static char *substrdup (const char *begin, const char *end)
-{
-  size_t len;
-  char *p;
-
-  /* remove trailing whitespace */
-  while (end > begin && isspace ((unsigned char) *(end - 1)))
-    end--;
-  len = end - begin;
-  p = malloc (len + 1);
-  if (!p)
-    return NULL;
-  memcpy (p, begin, len);
-  p[len] = 0;
-  return p;
-}
-#endif
-
 ADDRESS *rfc822_parse_adrlist (ADDRESS *top, const char *s)
 {
   const char *begin, *ps;
-  char comment[128], phrase[128];
+  char comment[STRING], phrase[STRING];
   size_t phraselen = 0, commentlen = 0;
   ADDRESS *cur, *last = NULL;
   
@@ -358,12 +337,12 @@ ADDRESS *rfc822_parse_adrlist (ADDRESS *top, const char *s)
       else if (commentlen && last && !last->personal)
       {
 	comment[commentlen] = 0;
-	last->personal = STRDUP (comment);
+	last->personal = safe_strdup (comment);
       }
 
 #ifdef EXACT_ADDRESS
-      if (last)
-	last->val = substrdup (begin, s);
+      if (last && !last->val)
+	last->val = mutt_substrdup (begin, s);
 #endif
       commentlen = 0;
       phraselen = 0;
@@ -386,7 +365,7 @@ ADDRESS *rfc822_parse_adrlist (ADDRESS *top, const char *s)
     {
       cur = rfc822_new_address ();
       phrase[phraselen] = 0;
-      cur->mailbox = STRDUP (phrase);
+      cur->mailbox = safe_strdup (phrase);
       cur->group = 1;
 
       if (last)
@@ -396,7 +375,7 @@ ADDRESS *rfc822_parse_adrlist (ADDRESS *top, const char *s)
       last = cur;
 
 #ifdef EXACT_ADDRESS
-      last->val = substrdup (begin, s);
+      last->val = mutt_substrdup (begin, s);
 #endif
 
       phraselen = 0;
@@ -412,14 +391,14 @@ ADDRESS *rfc822_parse_adrlist (ADDRESS *top, const char *s)
 	phrase[phraselen] = 0;
 	add_addrspec (&top, &last, phrase, comment, &commentlen, sizeof (comment) - 1);
       }
-      else if (commentlen && !last->personal)
+      else if (commentlen && last && !last->personal)
       {
 	comment[commentlen] = 0;
-	last->personal = STRDUP (comment);
+	last->personal = safe_strdup (comment);
       }
 #ifdef EXACT_ADDRESS
-      if (last)
-	last->val = substrdup (begin, s);
+      if (last && !last->val)
+	last->val = mutt_substrdup (begin, s);
 #endif
 
       /* add group terminator */
@@ -443,14 +422,15 @@ ADDRESS *rfc822_parse_adrlist (ADDRESS *top, const char *s)
       if (phraselen)
       {
 	if (cur->personal)
-	  free (cur->personal);
+	  FREE (&cur->personal);
 	/* if we get something like "Michael R. Elkins" remove the quotes */
 	rfc822_dequote_comment (phrase);
-	cur->personal = STRDUP (phrase);
+	cur->personal = safe_strdup (phrase);
       }
       if ((ps = parse_route_addr (s + 1, comment, &commentlen, sizeof (comment) - 1, cur)) == NULL)
       {
 	rfc822_free_address (&top);
+	rfc822_free_address (&cur);
 	return NULL;
       }
 
@@ -487,11 +467,11 @@ ADDRESS *rfc822_parse_adrlist (ADDRESS *top, const char *s)
   else if (commentlen && last && !last->personal)
   {
     comment[commentlen] = 0;
-    last->personal = STRDUP (comment);
+    last->personal = safe_strdup (comment);
   }
 #ifdef EXACT_ADDRESS
   if (last)
-    last->val = substrdup (begin, s);
+    last->val = mutt_substrdup (begin, s);
 #endif
 
   return top;
@@ -504,10 +484,9 @@ void rfc822_qualify (ADDRESS *addr, const char *host)
   for (; addr; addr = addr->next)
     if (!addr->group && addr->mailbox && strchr (addr->mailbox, '@') == NULL)
     {
-      if (!(p = malloc (strlen (addr->mailbox) + strlen (host) + 2)))
-	return;
+      p = safe_malloc (mutt_strlen (addr->mailbox) + mutt_strlen (host) + 2);
       sprintf (p, "%s@%s", addr->mailbox, host);
-      free (addr->mailbox);
+      safe_free ((void **) &addr->mailbox);
       addr->mailbox = p;
     }
 }
@@ -521,7 +500,7 @@ rfc822_cat (char *buf, size_t buflen, const char *value, const char *specials)
     size_t tmplen = sizeof (tmp) - 3;
 
     *pc++ = '"';
-    for (; *value && tmplen; value++)
+    for (; *value && tmplen > 1; value++)
     {
       if (*value == '\\' || *value == '"')
       {
@@ -550,171 +529,172 @@ void rfc822_write_address_single (char *buf, size_t buflen, ADDRESS *addr)
 
   buflen--; /* save room for the terminal nul */
 
-  len = strlen (buf);
-  pbuf += len;
-  buflen -= len;
-
 #ifdef EXACT_ADDRESS
   if (addr->val)
   {
+    if (!buflen)
+      goto done;
     strfcpy (pbuf, addr->val, buflen);
-    len = strlen (pbuf);
+    len = mutt_strlen (pbuf);
     pbuf += len;
     buflen -= len;
     if (addr->group)
     {
-      if (buflen)
-      {
-	*pbuf++ = ':';
-	buflen--;
-	*pbuf = 0;
-      }
+      if (!buflen)
+	goto done;
+      *pbuf++ = ':';
+      buflen--;
+      *pbuf = 0;
     }
     return;
   }
 #endif
-  pbuf = buf;
+
   if (addr->personal)
   {
-    if (strpbrk (addr->personal, rfc822_specials))
+    if (strpbrk (addr->personal, RFC822Specials))
     {
-      if (buflen)
+      if (!buflen)
+	goto done;
+      *pbuf++ = '"';
+      buflen--;
+      for (pc = addr->personal; *pc && buflen > 0; pc++)
       {
-	*pbuf++ = '"';
-	buflen--;
-	for (pc = addr->personal; *pc && buflen > 0; pc++)
+	if (*pc == '"' || *pc == '\\')
 	{
-	  if (*pc == '"' || *pc == '\\')
-	  {
-	    if (buflen)
-	    {
-	      *pbuf++ = '\\';
-	      buflen--;
-	    }
-	  }
-	  if (buflen)
-	  {
-	    *pbuf++ = *pc;
-	    buflen--;
-	  }
-	}
-
-	if (buflen)
-	{
-	  *pbuf++ = '"';
+	  if (!buflen)
+	    goto done;
+	  *pbuf++ = '\\';
 	  buflen--;
 	}
+	if (!buflen)
+	  goto done;
+	*pbuf++ = *pc;
+	buflen--;
       }
+      if (!buflen)
+	goto done;
+      *pbuf++ = '"';
+      buflen--;
     }
     else
     {
+      if (!buflen)
+	goto done;
       strfcpy (pbuf, addr->personal, buflen);
-      len = strlen (pbuf);
+      len = mutt_strlen (pbuf);
       pbuf += len;
       buflen -= len;
     }
 
-    if (buflen)
-    {
-      *pbuf++ = ' ';
-      buflen--;
+    if (!buflen)
+      goto done;
+    *pbuf++ = ' ';
+    buflen--;
+  }
 
-      if (buflen)
-      {
-	*pbuf++ = '<';
-	buflen--;
-      }
-    }
+  if (addr->personal || (addr->mailbox && *addr->mailbox == '@'))
+  {
+    if (!buflen)
+      goto done;
+    *pbuf++ = '<';
+    buflen--;
   }
 
   if (addr->mailbox)
   {
+    if (!buflen)
+      goto done;
     strfcpy (pbuf, addr->mailbox, buflen);
-    len = strlen (pbuf);
+    len = mutt_strlen (pbuf);
     pbuf += len;
     buflen -= len;
 
-    if (addr->personal)
+    if (addr->personal || (addr->mailbox && *addr->mailbox == '@'))
     {
-      if (buflen)
-      {
-	*pbuf++ = '>';
-	buflen--;
-      }
+      if (!buflen)
+	goto done;
+      *pbuf++ = '>';
+      buflen--;
     }
-    
+
     if (addr->group)
     {
-      if (buflen)
-      {
-	*pbuf++ = ':';
-	buflen--;
-	if (buflen)
-	{
-	  *pbuf++ = ' ';
-	  buflen--;
-	}
-      }
+      if (!buflen)
+	goto done;
+      *pbuf++ = ':';
+      buflen--;
+      if (!buflen)
+	goto done;
+      *pbuf++ = ' ';
+      buflen--;
     }
   }
   else
   {
-    if (buflen)
-    {
-      *pbuf++ = ';';
-      buflen--;
-    }
+    if (!buflen)
+      goto done;
+    *pbuf++ = ';';
+    buflen--;
   }
-
+done:
+  /* no need to check for length here since we already save space at the
+     beginning of this routine */
   *pbuf = 0;
 }
 
+/* note: it is assumed that `buf' is nul terminated! */
 void rfc822_write_address (char *buf, size_t buflen, ADDRESS *addr)
 {
   char *pbuf = buf;
-  size_t len = strlen (buf);
+  size_t len = mutt_strlen (buf);
   
   buflen--; /* save room for the terminal nul */
 
-  if (len)
+  if (len > 0)
   {
+    if (len > buflen)
+      return; /* safety check for bogus arguments */
+
     pbuf += len;
     buflen -= len;
-    if (buflen)
-    {
-      *pbuf++ = ',';
-      buflen--;
-      if (buflen)
-      {
-	*pbuf++ = ' ';
-	buflen--;
-      }
-    }
+    if (!buflen)
+      goto done;
+    *pbuf++ = ',';
+    buflen--;
+    if (!buflen)
+      goto done;
+    *pbuf++ = ' ';
+    buflen--;
   }
 
-  for (; addr && buflen; addr = addr->next)
+  for (; addr && buflen > 0; addr = addr->next)
   {
-    *pbuf = 0;
     /* use buflen+1 here because we already saved space for the trailing
        nul char, and the subroutine can make use of it */
     rfc822_write_address_single (pbuf, buflen + 1, addr);
-    len = strlen (pbuf);
+
+    /* this should be safe since we always have at least 1 char passed into
+       the above call, which means `pbuf' should always be nul terminated */
+    len = mutt_strlen (pbuf);
     pbuf += len;
     buflen -= len;
-    if (addr->next && !addr->group)
+
+    /* if there is another address, and its not a group mailbox name or
+       group terminator, add a comma to separate the addresses */
+    if (addr->next && addr->next->mailbox && !addr->group)
     {
-      if (buflen)
-      {
-	*pbuf++ = ',';
-	buflen--;
-	if (buflen)
-	{
-	  *pbuf++ = ' ';
-	  buflen--;
-	}
-      }
+      if (!buflen)
+	goto done;
+      *pbuf++ = ',';
+      buflen--;
+      if (!buflen)
+	goto done;
+      *pbuf++ = ' ';
+      buflen--;
     }
   }
+done:
   *pbuf = 0;
 }
 
@@ -724,10 +704,10 @@ ADDRESS *rfc822_cpy_adr_real (ADDRESS *addr)
   ADDRESS *p = rfc822_new_address ();
 
 #ifdef EXACT_ADDRESS
-  p->val = STRDUP (addr->val);
+  p->val = safe_strdup (addr->val);
 #endif
-  p->personal = STRDUP (addr->personal);
-  p->mailbox = STRDUP (addr->mailbox);
+  p->personal = safe_strdup (addr->personal);
+  p->mailbox = safe_strdup (addr->mailbox);
   p->group = addr->group;
   return p;
 }
@@ -750,46 +730,46 @@ ADDRESS *rfc822_cpy_adr (ADDRESS *addr)
   return top;
 }
 
-#if 0
-void print_addr_list (ADDRESS *list)
+/* append list 'b' to list 'a' and return the last element in the new list */
+ADDRESS *rfc822_append (ADDRESS **a, ADDRESS *b)
 {
-  while (list)
-  {
-    puts ("=====");
-#ifdef EXACT_ADDRESS
-    if (list->val)
-      printf ("val = [%s]\n", list->val);
-#endif
-    if (list->personal)
-      printf ("personal = [%s]\n", list->personal);
-    if (list->mailbox)
-      printf ("mailbox = [%s]\n", list->mailbox);
-    if (list->group)
-      printf ("group = [%d]\n", list->group);
-    list = list->next;
-  }
+  ADDRESS *tmp = *a;
+
+  while (tmp && tmp->next)
+    tmp = tmp->next;
+  if (!b)
+    return tmp;
+  if (tmp)
+    tmp->next = rfc822_cpy_adr (b);
+  else
+    tmp = *a = rfc822_cpy_adr (b);
+  while (tmp && tmp->next)
+    tmp = tmp->next;
+  return tmp;
+}
+
+#ifdef TESTING
+int safe_free (void **p)
+{
+  free(*p);
+  *p = 0;
 }
 
 int main (int argc, char **argv)
 {
   ADDRESS *list;
   char buf[256];
+# if 0
+  char *str = "michael, Michael Elkins <me@cs.hmc.edu>, testing a really complex address: this example <@contains.a.source.route@with.multiple.hosts:address@example.com>;, lothar@of.the.hillpeople (lothar)";
+# else
+  char *str = "a b c ";
+# endif
   
-  list = rfc822_parse_adrlist (NULL, " Michael . Elkins @ tis.com, \"Michael Elkins\" <michael@tis.com>, the  group : michael \"R.\" elkins <michael@tis.com>, me@cs.hmc.edu (michael elkins), \":sysmail\" @ (i am) vegas . (the greatest) com, elkins (michael) @ (the)aero(space corp).org , <michael> (michael elkins);, (michael) \"quoted string\" . legal @ [ 127.0.0. (localhost)1 ], foo <one(last)type(of)@address(testing)>");
-  if (list)
-    print_addr_list (list);
-  else
-    printf ("RFC822Error = %d.\n", RFC822Error);
-  rfc822_free_address (&list);
-  
-  list = rfc822_new_address ();
-  list->personal = STRDUP ("michael r. elkins");
-  list->mailbox = STRDUP ("michael");
+  list = rfc822_parse_adrlist (NULL, str);
   buf[0] = 0;
-  rfc822_write_list (buf, sizeof (buf), list);
+  rfc822_write_address (buf, sizeof (buf), list);
   rfc822_free_address (&list);
   puts (buf);
-
   exit (0);
 }
 #endif
