@@ -18,6 +18,7 @@
 
 #include "mutt.h"
 #include "mutt_curses.h"
+#include "rfc2047.h"
 #include "keymap.h"
 #include "mime.h"
 #include "mailbox.h"
@@ -33,15 +34,13 @@
 #include <sys/wait.h>
 #include <dirent.h>
 
-extern char RFC822Specials[];
+#ifdef _PGPPATH
+#include "pgp.h"
+#endif
 
 
 
-
-
-
-
-static void append_signature (ENVELOPE *env, FILE *f)
+static void append_signature (FILE *f)
 {
   FILE *tmpfp;
   pid_t thepid;
@@ -62,7 +61,7 @@ static int mutt_addrcmp (ADDRESS *a, ADDRESS *b)
 {
   if (!a->mailbox || !b->mailbox)
     return 0;
-  if (strcasecmp (a->mailbox, b->mailbox))
+  if (mutt_strcasecmp (a->mailbox, b->mailbox))
     return 0;
   return 1;
 }
@@ -217,7 +216,7 @@ static int edit_envelope (ENVELOPE *en)
     buf[0] = 0;
     for (; uh; uh = uh->next)
     {
-      if (strncasecmp ("subject:", uh->data, 8) == 0)
+      if (mutt_strncasecmp ("subject:", uh->data, 8) == 0)
       {
 	p = uh->data + 8;
 	SKIPWS (p);
@@ -227,9 +226,9 @@ static int edit_envelope (ENVELOPE *en)
   }
   
   if (mutt_get_field ("Subject: ", buf, sizeof (buf), 0) != 0 ||
-      (!buf[0] && query_quadoption (OPT_SUBJECT, "No subject, abort?") != 0))
+      (!buf[0] && query_quadoption (OPT_SUBJECT, _("No subject, abort?")) != 0))
   {
-    mutt_message ("No subject, aborting.");
+    mutt_message _("No subject, aborting.");
     return (-1);
   }
   safe_free ((void **) &en->subject);
@@ -244,11 +243,11 @@ static void process_user_recips (ENVELOPE *env)
 
   for (; uh; uh = uh->next)
   {
-    if (strncasecmp ("to:", uh->data, 3) == 0)
+    if (mutt_strncasecmp ("to:", uh->data, 3) == 0)
       env->to = rfc822_parse_adrlist (env->to, uh->data + 3);
-    else if (strncasecmp ("cc:", uh->data, 3) == 0)
+    else if (mutt_strncasecmp ("cc:", uh->data, 3) == 0)
       env->cc = rfc822_parse_adrlist (env->cc, uh->data + 3);
-    else if (strncasecmp ("bcc:", uh->data, 4) == 0)
+    else if (mutt_strncasecmp ("bcc:", uh->data, 4) == 0)
       env->bcc = rfc822_parse_adrlist (env->bcc, uh->data + 4);
   }
 }
@@ -264,21 +263,21 @@ static void process_user_header (ENVELOPE *env)
 
   for (; uh; uh = uh->next)
   {
-    if (strncasecmp ("from:", uh->data, 5) == 0)
+    if (mutt_strncasecmp ("from:", uh->data, 5) == 0)
     {
       /* User has specified a default From: address.  Remove default address */
       rfc822_free_address (&env->from);
       env->from = rfc822_parse_adrlist (env->from, uh->data + 5);
     }
-    else if (strncasecmp ("reply-to:", uh->data, 9) == 0)
+    else if (mutt_strncasecmp ("reply-to:", uh->data, 9) == 0)
     {
       rfc822_free_address (&env->reply_to);
       env->reply_to = rfc822_parse_adrlist (env->reply_to, uh->data + 9);
     }
-    else if (strncasecmp ("to:", uh->data, 3) != 0 &&
-	     strncasecmp ("cc:", uh->data, 3) != 0 &&
-	     strncasecmp ("bcc:", uh->data, 4) != 0 &&
-	     strncasecmp ("subject:", uh->data, 8) != 0)
+    else if (mutt_strncasecmp ("to:", uh->data, 3) != 0 &&
+	     mutt_strncasecmp ("cc:", uh->data, 3) != 0 &&
+	     mutt_strncasecmp ("bcc:", uh->data, 4) != 0 &&
+	     mutt_strncasecmp ("subject:", uh->data, 8) != 0)
     {
       if (last)
       {
@@ -316,9 +315,15 @@ static int include_forward (CONTEXT *ctx, HEADER *cur, FILE *out)
 {
   char buffer[STRING];
   int chflags = CH_DECODE, cmflags = 0;
+  
 
-
-
+#ifdef _PGPPATH
+  if ((cur->pgp & PGPENCRYPT) && option (OPTFORWDECODE))
+  {
+    /* make sure we have the user's passphrase before proceeding... */
+    pgp_valid_passphrase ();
+  }
+#endif /* _PGPPATH */
 
 
 
@@ -330,7 +335,11 @@ static int include_forward (CONTEXT *ctx, HEADER *cur, FILE *out)
   if (option (OPTFORWDECODE))
   {
     cmflags |= M_CM_DECODE;
-    chflags |= CH_WEED;
+    if (option (OPTFORWWEEDHEADER))
+    {
+      chflags |= CH_WEED;
+      cmflags |= M_CM_WEED;
+    }
   }
   if (option (OPTFORWQUOTE))
     cmflags |= M_CM_PREFIX;
@@ -347,6 +356,16 @@ static int include_reply (CONTEXT *ctx, HEADER *cur, FILE *out)
 
 
 
+#ifdef _PGPPATH
+  if (cur->pgp)
+  {
+    if (cur->pgp & PGPENCRYPT)
+    {
+      /* make sure we have the user's passphrase before proceeding... */
+      pgp_valid_passphrase ();
+    }
+  }
+#endif /* _PGPPATH */
 
 
 
@@ -369,72 +388,42 @@ static int include_reply (CONTEXT *ctx, HEADER *cur, FILE *out)
   return 0;
 }
 
-static BODY *make_forward (CONTEXT *ctx, HEADER *hdr)
-{
-  char buffer[LONG_STRING];
-  BODY *body;
-  FILE *fpout;
-
-  mutt_mktemp (buffer);
-  if ((fpout = safe_fopen (buffer, "w")) == NULL)
-    return NULL;
-
-  body = mutt_new_body ();
-  body->type = TYPEMESSAGE;
-  body->subtype = safe_strdup ("rfc822");
-  body->filename = safe_strdup (buffer);
-  body->unlink = 1;
-  body->use_disp = 0;
-
-  /* this MUST come after setting ->filename because we reuse buffer[] */
-  strfcpy (buffer, "Forwarded message from ", sizeof (buffer));
-  rfc822_write_address (buffer + 23, sizeof (buffer) - 23, hdr->env->from);
-  body->description = safe_strdup (buffer);
-
-  mutt_parse_mime_message (ctx, hdr);
-  mutt_copy_message (fpout, ctx, hdr,
-		     option (OPTMIMEFORWDECODE) ? M_CM_DECODE : 0,
-		     CH_XMIT | (option (OPTMIMEFORWDECODE) ? (CH_MIME | CH_TXTPLAIN ) : 0));
-  
-  fclose (fpout);
-  mutt_update_encoding (body);
-  return (body);
-}
-
 static int default_to (ADDRESS **to, ENVELOPE *env, int group)
 {
   char prompt[STRING];
   int i = 0;
 
-  if (group)
+  if (group && env->mail_followup_to)
   {
-    if (env->mail_followup_to)
-    {
-      rfc822_append (to, env->mail_followup_to);
-      return 0;
-    }
-    if (mutt_addr_is_user (env->from))
-      return 0;
+    rfc822_append (to, env->mail_followup_to);
+    return 0;
   }
 
-  if (!group && mutt_addr_is_user (env->from))
+  if (!option(OPTREPLYSELF) && mutt_addr_is_user (env->from))
   {
     /* mail is from the user, assume replying to recipients */
     rfc822_append (to, env->to);
   }
   else if (env->reply_to)
   {
-    if (option (OPTIGNORELISTREPLYTO) &&
+    if ((mutt_addrcmp (env->from, env->reply_to) && !env->reply_to->next) || 
+	(option (OPTIGNORELISTREPLYTO) &&
 	mutt_is_mail_list (env->reply_to) &&
 	(mutt_addrsrc (env->reply_to, env->to) ||
-	mutt_addrsrc (env->reply_to, env->cc)))
+	mutt_addrsrc (env->reply_to, env->cc))))
     {
       /* If the Reply-To: address is a mailing list, assume that it was
        * put there by the mailing list, and use the From: address
+       * 
+       * We also take the from header if our correspondant has a reply-to
+       * header which is identical to the electronic mail address given
+       * in his From header.
+       * 
        */
       rfc822_append (to, env->from);
     }
-    else if (!mutt_addrcmp (env->from, env->reply_to) &&
+    else if (!(mutt_addrcmp (env->from, env->reply_to) && 
+	       !env->reply_to->next) &&
 	     quadoption (OPT_REPLYTO) != M_YES)
     {
       /* There are quite a few mailing lists which set the Reply-To:
@@ -442,7 +431,9 @@ static int default_to (ADDRESS **to, ENVELOPE *env, int group)
        * to send a message to only the sender of the message.  This
        * provides a way to do that.
        */
-      snprintf (prompt, sizeof (prompt), "Reply to %s?", env->reply_to->mailbox);
+      snprintf (prompt, sizeof (prompt), _("Reply to %s%s?"),
+		env->reply_to->mailbox, 
+		env->reply_to->next?",...":"");
       if ((i = query_quadoption (OPT_REPLYTO, prompt)) == M_YES)
 	rfc822_append (to, env->reply_to);
       else if (i == M_NO)
@@ -475,13 +466,31 @@ static int fetch_recips (ENVELOPE *out, ENVELOPE *in, int flags)
 
     if ((flags & SENDGROUPREPLY) && !in->mail_followup_to)
     {
-      rfc822_append (&out->to, in->to);
+      /* if(!mutt_addr_is_user(in->to)) */
+      rfc822_append (&out->cc, in->to);
       rfc822_append (&out->cc, in->cc);
     }
   }
   return 0;
 }
 
+static LIST *make_references(ENVELOPE *e)
+{
+  LIST *t, *l;
+  
+  l = mutt_copy_list(e->references);
+  
+  if(e->message_id)
+  {
+    t = mutt_new_list();
+    t->data = safe_strdup(e->message_id);
+    t->next = l;
+    l = t;
+  }
+  
+  return l;
+}
+
 static int
 envelope_defaults (ENVELOPE *env, CONTEXT *ctx, HEADER *cur, int flags)
 {
@@ -506,7 +515,7 @@ envelope_defaults (ENVELOPE *env, CONTEXT *ctx, HEADER *cur, int flags)
       /* This could happen if the user tagged some messages and then did
        * a limit such that none of the tagged message are visible.
        */
-      mutt_error ("No tagged messages are visible!");
+      mutt_error _("No tagged messages are visible!");
       return (-1);
     }
   }
@@ -531,7 +540,7 @@ envelope_defaults (ENVELOPE *env, CONTEXT *ctx, HEADER *cur, int flags)
 
     if ((flags & SENDLISTREPLY) && !env->to)
     {
-      mutt_error ("No mailing lists found!");
+      mutt_error _("No mailing lists found!");
       return (-1);
     }
 
@@ -551,7 +560,7 @@ envelope_defaults (ENVELOPE *env, CONTEXT *ctx, HEADER *cur, int flags)
 
     if (curenv->real_subj)
     {
-      env->subject = safe_malloc (strlen (curenv->real_subj) + 5);
+      env->subject = safe_malloc (mutt_strlen (curenv->real_subj) + 5);
       sprintf (env->subject, "Re: %s", curenv->real_subj);
     }
     else
@@ -575,22 +584,30 @@ envelope_defaults (ENVELOPE *env, CONTEXT *ctx, HEADER *cur, int flags)
       tmp->data = safe_strdup (buffer);
     }
 
-    env->references = mutt_copy_list (curenv->references);
-
-    if (curenv->message_id)
+    if(tag)
     {
-      LIST *t;
+      HEADER *h;
+      LIST **p;
 
-      t = mutt_new_list ();
-      t->data = safe_strdup (curenv->message_id);
-      t->next = env->references;
-      env->references = t;
+      env->references = NULL;
+      p = &env->references;
+      
+      for(i = 0; i < ctx->vcount; i++)
+      {
+	while(*p) p = &(*p)->next;
+	h = ctx->hdrs[ctx->v2r[i]];
+	if(h->tagged)
+	  *p = make_references(h->env);
+      }
     }
+    else
+      env->references = make_references(curenv);
+
   }
   else if (flags & SENDFORWARD)
   {
     /* set the default subject for the message. */
-    mutt_make_string (buffer, sizeof (buffer), ForwFmt, ctx, cur);
+    mutt_make_string (buffer, sizeof (buffer), NONULL(ForwFmt), ctx, cur);
     env->subject = safe_strdup (buffer);
   }
 
@@ -610,7 +627,7 @@ generate_body (FILE *tempfp,	/* stream for outgoing message */
 
   if (flags & SENDREPLY)
   {
-    if ((i = query_quadoption (OPT_INCLUDE, "Include message in reply?")) == -1)
+    if ((i = query_quadoption (OPT_INCLUDE, _("Include message in reply?"))) == -1)
       return (-1);
 
     if (i == M_YES)
@@ -624,7 +641,7 @@ generate_body (FILE *tempfp,	/* stream for outgoing message */
 	  {
 	    if (include_reply (ctx, h, tempfp) == -1)
 	    {
-	      mutt_error ("Could not include all requested messages!");
+	      mutt_error _("Could not include all requested messages!");
 	      return (-1);
 	    }
 	    fputc ('\n', tempfp);
@@ -637,7 +654,7 @@ generate_body (FILE *tempfp,	/* stream for outgoing message */
   }
   else if (flags & SENDFORWARD)
   {
-    if (query_quadoption (OPT_MIMEFWD, "Forward MIME encapsulated?"))
+    if (query_quadoption (OPT_MIMEFWD, _("Forward MIME encapsulated?")))
     {
       BODY *last = msg->content;
 
@@ -646,7 +663,7 @@ generate_body (FILE *tempfp,	/* stream for outgoing message */
 
       if (cur)
       {
-	tmp = make_forward (ctx, cur);
+	tmp = mutt_make_message_attach (ctx, cur, 0);
 	if (last)
 	  last->next = tmp;
 	else
@@ -658,7 +675,7 @@ generate_body (FILE *tempfp,	/* stream for outgoing message */
 	{
 	  if (ctx->hdrs[ctx->v2r[i]]->tagged)
 	  {
-	    tmp = make_forward (ctx, ctx->hdrs[ctx->v2r[i]]);
+	    tmp = mutt_make_message_attach (ctx, ctx->hdrs[ctx->v2r[i]], 0);
 	    if (last)
 	    {
 	      last->next = tmp;
@@ -683,13 +700,24 @@ generate_body (FILE *tempfp,	/* stream for outgoing message */
 
 
 
+#ifdef _PGPPATH
+  else if (flags & SENDKEY) 
+  {
+    BODY *tmp;
+    if ((tmp = pgp_make_key_attachment (NULL)) == NULL)
+      return -1;
+
+    tmp->next = msg->content;
+    msg->content = tmp;
+  }
+#endif
 
 
 
   return (0);
 }
 
-static void mutt_set_followup_to (ENVELOPE *e)
+void mutt_set_followup_to (ENVELOPE *e)
 {
   ADDRESS *t = NULL;
 
@@ -697,7 +725,7 @@ static void mutt_set_followup_to (ENVELOPE *e)
      it hasn't already been set */
   if (option (OPTFOLLOWUPTO) && !e->mail_followup_to)
   {
-    if (mutt_is_list_recipient (e->to) || mutt_is_list_recipient (e->cc))
+    if (mutt_is_list_recipient (0, e->to, e->cc))
     {
       /* i am a list recipient, so set the Mail-Followup-To: field so that
        * i don't end up getting multiple copies of responses to my mail
@@ -711,20 +739,6 @@ static void mutt_set_followup_to (ENVELOPE *e)
   }
 }
 
-/* remove the multipart body if it exists */
-static BODY *remove_multipart (BODY *b)
-{
-  BODY *t;
-
-  if (b->parts)
-  {
-    t = b;
-    b = b->parts;
-    t->parts = NULL;
-    mutt_free_body (&t);
-  }
-  return b;
-}
 
 /* look through the recipients of the message we are replying to, and if
    we find an address that matches $alternates, we use that as the default
@@ -760,19 +774,72 @@ static ADDRESS *set_reverse_name (ENVELOPE *env)
 static ADDRESS *mutt_default_from (void)
 {
   ADDRESS *adr = rfc822_new_address ();
-
+  const char *fqdn = mutt_fqdn(1);
+  
   /* don't set realname here, it will be set later */
 
   if (option (OPTUSEDOMAIN))
   {
-    adr->mailbox = safe_malloc (strlen (Username) + strlen (Fqdn) + 2);
-    sprintf (adr->mailbox, "%s@%s", Username, Fqdn);
+    adr->mailbox = safe_malloc (mutt_strlen (Username) + mutt_strlen (fqdn) + 2);
+    sprintf (adr->mailbox, "%s@%s", NONULL(Username), NONULL(fqdn));
   }
   else
-    adr->mailbox = safe_strdup (Username);
+    adr->mailbox = safe_strdup (NONULL(Username));
   return (adr);
 }
 
+static int send_message (HEADER *msg)
+{  
+  char tempfile[_POSIX_PATH_MAX];
+  FILE *tempfp;
+  int i;
+
+  /* Write out the message in MIME form. */
+  mutt_mktemp (tempfile);
+  if ((tempfp = safe_fopen (tempfile, "w")) == NULL)
+    return (-1);
+
+  mutt_write_rfc822_header (tempfp, msg->env, msg->content, 0);
+  fputc ('\n', tempfp); /* tie off the header. */
+
+  if ((mutt_write_mime_body (msg->content, tempfp) == -1))
+  {
+    fclose(tempfp);
+    unlink (tempfile);
+    return (-1);
+  }
+  
+  if (fclose (tempfp) != 0)
+  {
+    mutt_perror (tempfile);
+    unlink (tempfile);
+    return (-1);
+  }
+
+  i = mutt_invoke_sendmail (msg->env->to, msg->env->cc, msg->env->bcc,
+		       tempfile, (msg->content->encoding == ENC8BIT));
+  return (i ? -1 : 0);
+}
+
+/* rfc2047 encode the content-descriptions */
+static void encode_descriptions (BODY *b)
+{
+  BODY *t;
+  char tmp[LONG_STRING];
+
+  for (t = b; t; t = t->next)
+  {
+    if (t->description)
+    {
+      rfc2047_encode_string (tmp, sizeof (tmp), (unsigned char *) t->description);
+      safe_free ((void **) &t->description);
+      t->description = safe_strdup (tmp);
+    }
+    if (t->parts)
+      encode_descriptions (t->parts);
+  }
+}
+
 void
 ci_send_message (int flags,		/* send mode */
 		 HEADER *msg,		/* template to use for new message */
@@ -785,11 +852,34 @@ ci_send_message (int flags,		/* send mode */
   FILE *tempfp = NULL;
   BODY *pbody;
   int i, killfrom = 0;
+#ifdef _PGPPATH
+  BODY *save_content = NULL;
+  char *pgpkeylist = NULL;
+  /* save current value of "pgp_sign_as" */
+  char *signas = NULL;
+  char *signmic = NULL;
+#endif
 
+  if (!flags && !msg && quadoption (OPT_RECALL) != M_NO && mutt_num_postponed ())
+  {
+    /* If the user is composing a new message, check to see if there
+     * are any postponed messages first.
+     */
+    if ((i = query_quadoption (OPT_RECALL, _("Recall postponed message?"))) == -1)
+      return;
 
-
-   
-
+    if(i == M_YES)
+      flags |= SENDPOSTPONED;
+  }
+  
+  
+#ifdef _PGPPATH
+  if (flags & SENDPOSTPONED)
+  {
+    signas = safe_strdup(PgpSignAs);
+    signmic = safe_strdup(PgpSignMicalg);
+  }
+#endif /* _PGPPATH */
 
   if (msg)
   {
@@ -801,27 +891,18 @@ ci_send_message (int flags,		/* send mode */
   {
     msg = mutt_new_header ();
 
-    if (flags == SENDPOSTPONED)
+    if (flags == SENDEDITMSG)
     {
-      if ((flags = mutt_get_postponed (ctx, msg, &cur)) < 0)
+      if (mutt_prepare_edit_message(ctx, msg, cur) < 0)
 	goto cleanup;
     }
-    else if (!flags && quadoption (OPT_RECALL) != M_NO && mutt_num_postponed ())
+    else if (flags == SENDPOSTPONED)
     {
-      /* If the user is composing a new message, check to see if there
-       * are any postponed messages first.
-       */
-      if ((i = query_quadoption (OPT_RECALL, "Recall postponed message?")) == -1)
+      if ((flags = mutt_get_postponed (ctx, msg, &cur, fcc, sizeof (fcc))) < 0)
 	goto cleanup;
-
-      if (i == M_YES)
-      {
-	if ((flags = mutt_get_postponed (ctx, msg, &cur)) < 0)
-	  flags = 0;
-      }
     }
 
-    if (flags & SENDPOSTPONED)
+    if (flags & (SENDPOSTPONED | SENDEDITMSG))
     {
       if ((tempfp = safe_fopen (msg->content->filename, "a+")) == NULL)
       {
@@ -834,7 +915,7 @@ ci_send_message (int flags,		/* send mode */
       msg->env = mutt_new_envelope ();
   }
 
-  if (! (flags & (SENDKEY | SENDPOSTPONED)))
+  if (! (flags & (SENDKEY | SENDPOSTPONED | SENDEDITMSG)))
   {
     pbody = mutt_new_body ();
     pbody->next = msg->content; /* don't kill command-line attachments */
@@ -866,16 +947,16 @@ ci_send_message (int flags,		/* send mode */
   }
 
   /* this is handled here so that the user can match ~f in send-hook */
-  if (cur && option (OPTREVNAME))
+  if (cur && option (OPTREVNAME) && !(flags & (SENDPOSTPONED | SENDEDITMSG)))
   {
     /* we shouldn't have to worry about freeing `msg->env->from' before
-       setting it here since this code will only execute when doing some
-       sort of reply.  the pointer will only be set when using the -H command
-       line option */
+     * setting it here since this code will only execute when doing some
+     * sort of reply.  the pointer will only be set when using the -H command
+     * line option */
     msg->env->from = set_reverse_name (cur->env);
   }
 
-  if (!msg->env->from && option (OPTUSEFROM))
+  if (!msg->env->from && option (OPTUSEFROM) && !(flags & (SENDEDITMSG|SENDPOSTPONED)))
     msg->env->from = mutt_default_from ();
 
   if (flags & SENDBATCH) 
@@ -887,7 +968,7 @@ ci_send_message (int flags,		/* send mode */
       process_user_header (msg->env);
     }
   }
-  else if (! (flags & SENDPOSTPONED))
+  else if (! (flags & (SENDPOSTPONED | SENDEDITMSG)))
   {
     if ((flags & (SENDREPLY | SENDFORWARD)) &&
 	envelope_defaults (msg->env, ctx, cur, flags) == -1)
@@ -905,9 +986,10 @@ ci_send_message (int flags,		/* send mode */
     }
 
     /* the from address must be set here regardless of whether or not
-       $use_from is set so that the `~P' (from you) operator in send-hook
-       patterns will work.  if $use_from is unset, the from address is killed
-       after send-hooks are evaulated */
+     * $use_from is set so that the `~P' (from you) operator in send-hook
+     * patterns will work.  if $use_from is unset, the from address is killed
+     * after send-hooks are evaulated */
+
     if (!msg->env->from)
     {
       msg->env->from = mutt_default_from ();
@@ -915,6 +997,7 @@ ci_send_message (int flags,		/* send mode */
     }
 
     /* change settings based upon recipients */
+    
     mutt_send_hook (msg);
 
     if (killfrom)
@@ -928,6 +1011,19 @@ ci_send_message (int flags,		/* send mode */
 
 
 
+#ifdef _PGPPATH
+    if (! (flags & SENDMAILX))
+    {
+      if (option (OPTPGPAUTOSIGN))
+	msg->pgp |= PGPSIGN;
+      if (option (OPTPGPAUTOENCRYPT))
+	msg->pgp |= PGPENCRYPT;
+      if (option (OPTPGPREPLYENCRYPT) && cur && cur->pgp & PGPENCRYPT)
+	msg->pgp |= PGPENCRYPT;
+      if (option (OPTPGPREPLYSIGN) && cur && cur->pgp & PGPSIGN)
+	msg->pgp |= PGPSIGN;
+    }
+#endif /* _PGPPATH */
 
 
 
@@ -935,16 +1031,22 @@ ci_send_message (int flags,		/* send mode */
     if (generate_body (tempfp, msg, flags, ctx, cur) == -1)
       goto cleanup;
 
-    if (! (flags & (SENDMAILX | SENDKEY)) && strcmp (Editor, "builtin") != 0)
-      append_signature (msg->env, tempfp);
+    if (! (flags & (SENDMAILX | SENDKEY)) && Editor && mutt_strcmp (Editor, "builtin") != 0)
+      append_signature (tempfp);
   }
   /* wait until now to set the real name portion of our return address so
      that $realname can be set in a send-hook */
-  if (msg->env->from && !msg->env->from->personal)
+  if (msg->env->from && !msg->env->from->personal && !(flags & (SENDEDITMSG | SENDPOSTPONED)))
     msg->env->from->personal = safe_strdup (Realname);
 
 
 
+#ifdef _PGPPATH
+  
+  if (! (flags & SENDKEY))
+  {
+
+#endif
     
 
 
@@ -953,6 +1055,11 @@ ci_send_message (int flags,		/* send mode */
     
 
 
+#ifdef _PGPPATH
+    
+  }
+
+#endif
 
 
 
@@ -975,8 +1082,8 @@ ci_send_message (int flags,		/* send mode */
     if(! (flags & SENDKEY))
     {
       if (mutt_needs_mailcap (msg->content))
-	mutt_edit_attachment (msg->content, 0);
-      else if (strcmp ("builtin", Editor) == 0)
+	mutt_edit_attachment (msg->content);
+      else if (!Editor || mutt_strcmp ("builtin", Editor) == 0)
 	mutt_builtin_editor (msg->content->filename, msg, cur);
       else if (option (OPTEDITHDRS))
 	mutt_edit_headers (Editor, msg->content->filename, msg, fcc, sizeof (fcc));
@@ -984,15 +1091,15 @@ ci_send_message (int flags,		/* send mode */
 	mutt_edit_file (Editor, msg->content->filename);
     }
 
-    if (! (flags & (SENDPOSTPONED | SENDFORWARD | SENDKEY)))
+    if (! (flags & (SENDPOSTPONED | SENDEDITMSG | SENDFORWARD | SENDKEY)))
     {
       if (stat (msg->content->filename, &st) == 0)
       {
 	/* if the file was not modified, bail out now */
 	if (mtime == st.st_mtime &&
-	    query_quadoption (OPT_ABORT, "Abort unmodified message?") == M_YES)
+	    query_quadoption (OPT_ABORT, _("Abort unmodified message?")) == M_YES)
 	{
-	  mutt_message ("Aborted unmodified message.");
+	  mutt_message _("Aborted unmodified message.");
 	  goto cleanup;
 	}
       }
@@ -1002,7 +1109,8 @@ ci_send_message (int flags,		/* send mode */
   }
 
   /* specify a default fcc.  if we are in batchmode, only save a copy of
-     the message if the value of $copy is yes or ask-yes */
+   * the message if the value of $copy is yes or ask-yes */
+
   if (!fcc[0] && (!(flags & SENDBATCH) || (quadoption (OPT_COPY) & 0x1)))
   {
     /* set the default FCC */
@@ -1026,10 +1134,11 @@ ci_send_message (int flags,		/* send mode */
   {
 main_loop:
 
-    if ((i = mutt_send_menu (msg, fcc, sizeof (fcc), cur)) == -1)
+    i = mutt_compose_menu (msg, fcc, sizeof (fcc), cur);
+    if (i == -1)
     {
       /* abort */
-      mutt_message ("Mail not sent.");
+      mutt_message _("Mail not sent.");
       goto cleanup;
     }
     else if (i == 1)
@@ -1037,12 +1146,12 @@ main_loop:
       /* postpone the message until later. */
       if (msg->content->next)
 	msg->content = mutt_make_multipart (msg->content);
-      if (mutt_write_fcc (NONULL (Postponed), msg, (cur && (flags & SENDREPLY)) ? cur->env->message_id : NULL, 1) < 0)
+      if (!Postponed || mutt_write_fcc (NONULL (Postponed), msg, (cur && (flags & SENDREPLY)) ? cur->env->message_id : NULL, 1, fcc) < 0)
       {
-	msg->content = remove_multipart (msg->content);
+	msg->content = mutt_remove_multipart (msg->content);
 	goto main_loop;
       }
-      mutt_message ("Message postponed.");
+      mutt_message _("Message postponed.");
       goto cleanup;
     }
   }
@@ -1051,67 +1160,175 @@ main_loop:
   {
     if (! (flags & SENDBATCH))
     {
-      mutt_error ("No recipients are specified!");
+      mutt_error _("No recipients are specified!");
       goto main_loop;
     }
     else
     {
-      puts ("No recipients were specified.");
+      puts _("No recipients were specified.");
       goto cleanup;
     }
   }
 
   if (!msg->env->subject && ! (flags & SENDBATCH) &&
-      (i = query_quadoption (OPT_SUBJECT, "No subject, abort sending?")) != M_NO)
+      (i = query_quadoption (OPT_SUBJECT, _("No subject, abort sending?"))) != M_NO)
   {
     /* if the abort is automatic, print an error message */
     if (quadoption (OPT_SUBJECT) == M_YES)
-      mutt_error ("No subject specified.");
+      mutt_error _("No subject specified.");
     goto main_loop;
   }
 
   if (msg->content->next)
     msg->content = mutt_make_multipart (msg->content);
 
+  /* Ok, we need to do it this way instead of handling all fcc stuff in
+   * one place in order to avoid going to main_loop with encoded "env"
+   * in case of error.  Ugh.
+   */
+#ifdef _PGPPATH
+  if (msg->pgp)
+  {
+    if (pgp_get_keys (msg, &pgpkeylist) == -1)
+      goto main_loop;
 
+    /* save the decrypted attachments */
+    save_content = msg->content;
 
+    if (pgp_protect (msg, pgpkeylist) == -1)
+    {
+      if (msg->content->parts)
+      {
+	/* remove the toplevel multipart structure */
+	pbody = msg->content;
+	msg->content = msg->content->parts;
+	pbody->parts = NULL;
+	mutt_free_body (&pbody);
+      }
+      if (pgpkeylist)
+	FREE (&pgpkeylist);
+      goto main_loop;
+    }
+  }
+#endif /* _PGPPATH */
 
+  if (flags & SENDEDITMSG)
+  {
+   int really_send = mutt_yesorno (_("Message edited. Really send?"), 1);
+   if (really_send != M_YES)
+     goto main_loop;
+  }
 
+  if (!option (OPTNOCURSES) && !(flags & SENDMAILX))
+    mutt_message _("Sending message...");
 
+  mutt_prepare_envelope (msg->env);
+  encode_descriptions (msg->content);
+
+  /* save a copy of the message, if necessary. */
   mutt_expand_path (fcc, sizeof (fcc));
-
-  if (!option (OPTNOCURSES) && ! (flags & SENDMAILX))
-    mutt_message ("Sending message...");
-
-  if (msg->env->bcc && ! (msg->env->to || msg->env->cc))
+  if (*fcc && mutt_strcmp ("/dev/null", fcc) != 0)
   {
-    /* some MTA's will put an Apparently-To: header field showing the Bcc:
-     * recipients if there is no To: or Cc: field, so attempt to suppress
-     * it by using an empty To: field.
-     */
-    msg->env->to = rfc822_new_address ();
-    msg->env->to->mailbox = safe_strdup ("undisclosed-recipients:;");
-    msg->env->to->group = 1;
-    msg->env->to->next = rfc822_new_address ();
- 
-    buffer[0] = 0;
-    rfc822_cat (buffer, sizeof (buffer), msg->env->to->mailbox, RFC822Specials);
-    msg->env->to->mailbox = safe_strdup (buffer);
+    BODY *tmpbody = msg->content;
+#ifdef _PGPPATH
+    BODY *save_sig = NULL;
+    BODY *save_parts = NULL;
+#endif /* _PGPPATH */
+
+    /* check to see if the user wants copies of all attachments */
+    if (!option (OPTFCCATTACH) && msg->content->type == TYPEMULTIPART)
+    {
+#ifdef _PGPPATH
+      if (mutt_strcmp (msg->content->subtype, "encrypted") == 0 ||
+	  mutt_strcmp (msg->content->subtype, "signed") == 0)
+      {
+	if (save_content->type == TYPEMULTIPART)
+	{
+	  if (!(msg->pgp & PGPENCRYPT) && (msg->pgp & PGPSIGN))
+	  {
+	    /* save initial signature and attachments */
+	    save_sig = msg->content->parts->next;
+	    save_parts = msg->content->parts->parts->next;
+	  }
+
+	  /* this means writing only the main part */
+	  msg->content = save_content->parts;
+
+	  if (pgp_protect (msg, pgpkeylist) == -1)
+	  {
+	    /* we can't do much about it at this point, so
+	     * fallback to saving the whole thing to fcc
+	     */
+	    msg->content = tmpbody;
+	    save_sig = NULL;
+	    goto full_fcc;
+	  }
+
+	  if (msg->pgp & PGPENCRYPT)
+	  {
+	    /* not released in pgp_encrypt_message() */
+	    mutt_free_body (&save_content->parts);
+	    /* make sure we release the right thing later */
+	    save_content->parts = msg->content;
+
+	    encode_descriptions (msg->content);
+	  }
+	  else
+	    save_content = msg->content;
+	}
+      }
+      else
+#endif /* _PGPPATH */
+	msg->content = msg->content->parts;
+    }
+
+#ifdef _PGPPATH
+full_fcc:
+#endif /* _PGPPATH */
+    if (msg->content)
+      mutt_write_fcc (fcc, msg, NULL, 0, NULL);
+    msg->content = tmpbody;
+
+#ifdef _PGPPATH
+    if (save_sig)
+    {
+      /* cleanup the second signature structures */
+      mutt_free_body (&save_content->parts->next);
+      save_content->parts = NULL;
+      mutt_free_body (&save_content);
+
+      /* restore old signature and attachments */
+      msg->content->parts->next = save_sig;
+      msg->content->parts->parts->next = save_parts;
+    }
+#endif /* _PGPPATH */
   }
 
-  mutt_set_followup_to (msg->env);
-
-  if (mutt_send_message (msg, fcc) == -1)
+#ifdef _PGPPATH
+  if (msg->pgp & PGPENCRYPT)
   {
-    msg->content = remove_multipart (msg->content);
-    goto main_loop;
+    /* cleanup structures from the first encryption */
+    mutt_free_body (&save_content);
+    FREE (&pgpkeylist);
+  }
+#endif /* _PGPPATH */
+
+  if (send_message (msg) == -1)
+  {
+    if (!(flags & SENDBATCH))
+    {
+      msg->content = mutt_remove_multipart (msg->content);
+      goto main_loop;
+    }
+    else
+    {
+      puts _("Could not send the message.");
+      goto cleanup;
+    }
   }
 
   if (!option (OPTNOCURSES) && ! (flags & SENDMAILX))
-    mutt_message ("Mail sent.");
-
-  /* now free up the memory used to generate this message. */
-  mutt_free_header (&msg);
+    mutt_message _("Mail sent.");
 
   if (flags & SENDREPLY)
   {
@@ -1125,21 +1342,30 @@ main_loop:
     }
   }
 
-
-
-   
-
-
-  return; /* all done */
-
 cleanup:
 
 
 
+#ifdef _PGPPATH
+  if (flags & SENDPOSTPONED)
+  {
+    
+    if(signas)
+    {
+      safe_free((void **) &PgpSignAs);
+      PgpSignAs = signas;
+    }
+    
+    if(signmic)
+    {
+      safe_free((void **) &PgpSignMicalg);
+      PgpSignMicalg = signmic;
+    }
+  }
+#endif /* _PGPPATH */
    
-
-
   if (tempfp)
     fclose (tempfp);
   mutt_free_header (&msg);
+
 }
