@@ -23,6 +23,7 @@
 #include "mime.h"
 #include "mailbox.h"
 #include "copy.h"
+#include "charset.h"
 
 #include <string.h>
 #include <stdlib.h>
@@ -33,10 +34,69 @@
 #include <signal.h>
 #include <sys/wait.h>
 #include <fcntl.h>
+
+#ifdef HAVE_SYSEXITS_H
 #include <sysexits.h>
+#endif
 
+extern char RFC822Specials[];
 
-
+static struct sysexits
+{
+  int v;
+  const char *str;
+} 
+sysexits_h[] = 
+{
+#ifdef EX_USAGE
+  { EX_USAGE, "The command was used incorrectly." },
+#endif
+#ifdef EX_DATAERR
+  { EX_DATAERR, "The input data was incorrect." },
+#endif
+#ifdef EX_NOINPUT
+  { EX_NOINPUT, "No input." },
+#endif
+#ifdef EX_NOUSER
+  { EX_NOUSER, "No such user." },
+#endif
+#ifdef EX_NOHOST
+  { EX_NOHOST, "Host not found." },
+#endif
+#ifdef EX_UNAVAILABLE
+  { EX_UNAVAILABLE, "Service unavailable." },
+#endif
+#ifdef EX_SOFTWARE
+  { EX_SOFTWARE, "Software error." },
+#endif
+#ifdef EX_OSERR
+  { EX_OSERR, "Operating system error." },
+#endif
+#ifdef EX_OSFILE
+  { EX_OSFILE, "System file is missing." },
+#endif
+#ifdef EX_CANTCREAT
+  { EX_CANTCREAT, "Can't create output." },
+#endif
+#ifdef EX_IOERR
+  { EX_IOERR, "I/O error." },
+#endif
+#ifdef EX_TEMPFAIL
+  { EX_TEMPFAIL, "Temporary failure." },
+#endif
+#ifdef EX_PROTOCOL
+  { EX_PROTOCOL, "Protocol error." },
+#endif
+#ifdef EX_NOPERM
+  { EX_NOPERM, "Permission denied." },
+#endif
+  { -1, NULL}
+};
+      
+    
+#ifdef _PGPPATH
+#include "pgp.h"
+#endif /* _PGPPATH */
 
 
 
@@ -64,7 +124,7 @@ static void encode_quoted (FILE * fin, FILE *fout, int istext)
   while ((c = fgetc (fin)) != EOF)
   {
     /* Escape lines that begin with "the message separator". */
-    if (linelen == 5 && !strncmp ("From ", line, 5))
+    if (linelen == 5 && !mutt_strncmp ("From ", line, 5))
     {
       strfcpy (line, "=46rom ", sizeof (line));
       linelen = 7;
@@ -108,11 +168,11 @@ static void encode_quoted (FILE * fin, FILE *fout, int istext)
     if (c == '\n' && istext)
     {
       /* Check to make sure there is no trailing space on this line. */
-      if (line[linelen-1] == ' ' || line[linelen-1] == '\t')
+      if (linelen > 0 && (line[linelen-1] == ' ' || line[linelen-1] == '\t'))
       {
         if (linelen < 74)
 	{
-          sprintf (line+linelen-1, "=%2.2X", line[linelen-1]);
+          sprintf (line+linelen-1, "=%2.2X", (unsigned char) line[linelen-1]);
           fputs (line, fout);
         }
         else
@@ -122,7 +182,7 @@ static void encode_quoted (FILE * fin, FILE *fout, int istext)
           line[linelen-1] = '=';
           line[linelen] = 0;
           fputs (line, fout);
-          fprintf (fout, "\n=%2.2X", savechar);
+          fprintf (fout, "\n=%2.2X", (unsigned char) savechar);
         }
       }
       else
@@ -146,7 +206,7 @@ static void encode_quoted (FILE * fin, FILE *fout, int istext)
         fputc ('\n', fout);
         linelen = 0;
       }
-      sprintf (line+linelen,"=%2.2X", c);
+      sprintf (line+linelen,"=%2.2X", (unsigned char) c);
       linelen += 3;
     }
     else
@@ -165,7 +225,7 @@ static void encode_quoted (FILE * fin, FILE *fout, int istext)
     {
       /* take care of trailing whitespace */
       if (linelen < 74)
-        sprintf (line+linelen-1, "=%2.2X", line[linelen-1]);
+        sprintf (line+linelen-1, "=%2.2X", (unsigned char) line[linelen-1]);
       else
       {
         savechar = line[linelen-1];
@@ -173,7 +233,7 @@ static void encode_quoted (FILE * fin, FILE *fout, int istext)
         line[linelen] = 0;
         fputs (line, fout);
         fputc ('\n', fout);
-        sprintf (line, "=%2.2X", savechar);
+        sprintf (line, "=%2.2X", (unsigned char) savechar);
       }
     }
     else
@@ -182,123 +242,141 @@ static void encode_quoted (FILE * fin, FILE *fout, int istext)
   }
 }
 
+static char b64_buffer[3];
+static short b64_num;
+static short b64_linelen;
+
+static void b64_flush(FILE *fout)
+{
+  short i;
+
+  if(!b64_num)
+    return;
+
+  if(b64_linelen >= 72)
+  {
+    fputc('\n', fout);
+    b64_linelen = 0;
+  }
+  
+  for(i = b64_num; i < 3; i++)
+    b64_buffer[i] = '\0';
+  
+  fputc(B64Chars[(b64_buffer[0] >> 2) & 0x3f], fout); 
+  b64_linelen++;
+  fputc(B64Chars[((b64_buffer[0] & 0x3) << 4) | ((b64_buffer[1] >> 4) & 0xf) ], fout);
+  b64_linelen++;
+  
+  if(b64_num > 1)
+  {
+    fputc(B64Chars[((b64_buffer[1] & 0xf) << 2) | ((b64_buffer[2] >> 6) & 0x3) ], fout);
+    b64_linelen++;
+    if(b64_num > 2)
+    {
+      fputc(B64Chars[b64_buffer[2] & 0x3f], fout);
+      b64_linelen++;
+    }
+  }
+  
+  while(b64_linelen % 4)
+  {
+    fputc('=', fout);
+    b64_linelen++;
+  }
+  
+  b64_num = 0;
+}
+    
+  
+static void b64_putc(char c, FILE *fout)
+{
+  if(b64_num == 3)
+    b64_flush(fout);
+  
+  b64_buffer[b64_num++] = c;
+}
+  
+  
 static void encode_base64 (FILE * fin, FILE *fout, int istext)
 {
-  int c1, c2, c3, ch;
-  int insert_newline = 0;
-  int linelen = 0;
-
-  FOREVER
-  {
-    if (istext)
-    {
-      if (insert_newline)
-      {
-        c1 = '\n';
-        insert_newline = 0;
+  int ch, ch1 = EOF;
   
-        c2 = fgetc(fin);
-        if (c2 == '\n')
-	{
-          c2 = '\r';
-          c3 = '\n';
-        }
-        else
-	{
-          c3 = fgetc(fin);
-          if (c3 == '\n')
-	  {
-            c3 = '\r';
-            insert_newline = 1;
-          }
-        }
+  b64_num = b64_linelen = 0;
+  
+  while((ch = fgetc(fin)) != EOF)
+  {
+    if(istext && ch == '\n' && ch1 != '\r')
+      b64_putc('\r', fout);
+    b64_putc(ch, fout);
+    ch1 = ch;
+  }
+  b64_flush(fout);
+  fputc('\n', fout);
+}
+
+#ifdef PERMIT_DEPRECATED_UUENCODED_MESSAGES
+
+#define UUENC(c) ((c) ? ((c) & 077) + ' ' : '`')
+
+static void encode_uuenc (FILE * fin, FILE * fout)
+{
+  register int ch, linelen;
+  register unsigned char *p;
+  unsigned char line[80];
+
+  while ((linelen = fread(line, 1, 45, fin)))
+  {
+    ch = UUENC(linelen);
+    fputc (ch, fout);
+
+    for (p = line; linelen>0; linelen -= 3, p += 3)
+    {
+      ch = *p >> 2;
+      ch = UUENC(ch);
+      fputc(ch, fout);
+
+      if (linelen>1)
+      {
+	ch = ((*p & 0x3) << 4) | (p[1] >> 4);
+	ch = UUENC(ch);
+	fputc(ch, fout);
       }
       else
       {
-        c1 = fgetc(fin);
-        if (c1 == '\n')
-	{
-          c1 = '\r';
-          c2 = '\n';
-          c3 = fgetc(fin);
-          if (c3 == '\n')
-	  {
-            c3 = '\r';
-            insert_newline = 1;
-          }
-        }
-        else
-	{
-          c2 = fgetc(fin);
-          if (c2 == '\n')
-	  {
-            c2 = '\r';
-            c3 = '\n';
-          }
-          else
-	  {
-            c3 = fgetc(fin);
-            if (c3 == '\n')
-	    {
-              c3 = '\r';
-              insert_newline = 1;
-            }
-          }
-        }
+	ch = (*p & 0x3) << 4;
+	ch = UUENC(ch);
+	fputc(ch, fout);
+	break;
       }
-    }
-    else /* !istext */
-    {
-      if ((c1 = fgetc(fin)) == EOF)
-        break;
-      c2 = fgetc(fin);
-      c3 = fgetc(fin);
+
+      if (linelen>2)
+      {
+	ch = ((p[1] & 0xf) << 2) | (p[2] >> 6);
+	ch = UUENC(ch);
+	fputc(ch, fout);
+      }
+      else
+      {
+	ch = (p[1] & 0xf) << 2;
+	ch = UUENC(ch);
+	fputc(ch, fout);
+	break;
+      }
+
+      ch = p[2] & 0x3f;
+      ch = UUENC(ch);
+      fputc(ch, fout);
     }
 
-    if (linelen + 4 >= 76)
-    {
-      fputc('\n', fout);
-      linelen = 0;
-    }
-
-    ch = c1 >> 2;
-    fputc (B64Chars[ch], fout);
-
-    if (c2 != EOF)
-    {
-      ch = ((c1 & 0x3) << 4) | (c2 >> 4);
-      fputc (B64Chars[ch], fout);
-    }
-    else
-    {
-      ch = (c1 & 0x3) << 4;
-      fputc (B64Chars[ch], fout);
-      fputs("==", fout);
-      break;
-    }
-
-    if (c3 != EOF)
-    {
-      ch = ((c2 & 0xf) << 2) | (c3 >> 6);
-      fputc (B64Chars[ch], fout);
-    }
-    else
-    {
-      ch = (c2 & 0xf) << 2;
-      fputc(B64Chars[ch], fout);
-      fputc('=', fout);
-      break;
-    }
-
-    ch = c3 & 0x3f;
-    fputc(B64Chars[ch], fout);
-
-    linelen += 4;
+    fputc('\n', fout);
   }
-
+  ch = UUENC('\0');
+  fputc(ch, fout);
   fputc('\n', fout);
 }
 
+#endif
+
 int mutt_write_mime_header (BODY *a, FILE *f)
 {
   PARAMETER *p;
@@ -308,21 +386,24 @@ int mutt_write_mime_header (BODY *a, FILE *f)
   int len;
   int tmplen;
   
-  fprintf (f, "Content-Type: %s/%s", TYPE (a->type), a->subtype);
+  fprintf (f, "Content-Type: %s/%s", TYPE (a), a->subtype);
 
   if (a->parameter)
   {
-    len = 25 + strlen (a->subtype); /* approximate len. of content-type */
+    len = 25 + mutt_strlen (a->subtype); /* approximate len. of content-type */
 
-    p = a->parameter;
-    while (p)
+    for(p = a->parameter; p; p = p->next)
     {
+      
+      if(!p->value)
+	continue;
+      
       fputc (';', f);
 
       buffer[0] = 0;
       rfc822_cat (buffer, sizeof (buffer), p->value, MimeSpecials);
 
-      tmplen = strlen (buffer) + strlen (p->attribute) + 1;
+      tmplen = mutt_strlen (buffer) + mutt_strlen (p->attribute) + 1;
 
       if (len + tmplen + 2 > 76)
       {
@@ -337,15 +418,11 @@ int mutt_write_mime_header (BODY *a, FILE *f)
 
       fprintf (f, "%s=%s", p->attribute, buffer);
 
-      p = p->next;
     }
   }
 
   fputc ('\n', f);
 
-  if (a->encoding != ENC7BIT)
-    fprintf(f, "Content-Transfer-Encoding: %s\n", ENCODING (a->encoding));
-
   if (a->description)
     fprintf(f, "Content-Description: %s\n", a->description);
 
@@ -372,6 +449,9 @@ int mutt_write_mime_header (BODY *a, FILE *f)
     fputc ('\n', f);
   }
 
+  if (a->encoding != ENC7BIT)
+    fprintf(f, "Content-Transfer-Encoding: %s\n", ENCODING (a->encoding));
+
   /* Do NOT add the terminator here!!! */
   return (ferror (f) ? -1 : 0);
 }
@@ -381,6 +461,9 @@ int mutt_write_mime_body (BODY *a, FILE *f)
   char *p, boundary[SHORT_STRING];
   FILE *fpin;
   BODY *t;
+#ifdef PERMIT_DEPRECATED_UUENCODED_MESSAGES
+  char *r;
+#endif
 
   if (a->type == TYPEMULTIPART)
   {
@@ -388,6 +471,7 @@ int mutt_write_mime_body (BODY *a, FILE *f)
     if (!(p = mutt_get_parameter ("boundary", a->parameter)))
     {
       dprint (1, (debugfile, "mutt_write_mime_body(): no boundary parameter found!\n"));
+      mutt_error _("No boundary parameter found! [report this error]");
       return (-1);
     }
     strfcpy (boundary, p, sizeof (boundary));
@@ -407,12 +491,21 @@ int mutt_write_mime_body (BODY *a, FILE *f)
 
 
 
+#ifdef _PGPPATH
+  /* This is pretty gross, but it's the best solution for now... */
+  if (a->type == TYPEAPPLICATION && mutt_strcmp (a->subtype, "pgp-encrypted") == 0)
+  {
+    fputs ("Version: 1\n", f);
+    return 0;
+  }
+#endif /* _PGPPATH */
 
 
 
   if ((fpin = fopen (a->filename, "r")) == NULL)
   {
     dprint(1,(debugfile, "write_mime_body: %s no longer exists!\n",a->filename));
+    mutt_error (_("%s no longer exists!"), a->filename);
     return -1;
   }
 
@@ -420,6 +513,19 @@ int mutt_write_mime_body (BODY *a, FILE *f)
     encode_quoted (fpin, f, mutt_is_text_type (a->type, a->subtype));
   else if (a->encoding == ENCBASE64)
     encode_base64 (fpin, f, mutt_is_text_type (a->type, a->subtype));
+#ifdef PERMIT_DEPRECATED_UUENCODED_MESSAGES
+  else if (a->encoding == ENCUUENCODED)
+  {
+    /* Strip off the leading path... */
+    if ((r = strrchr (a->filename, '/')))
+      r++;
+    else
+      r = a->filename;
+    fprintf (f, "begin 600 %s\n", r);
+    encode_uuenc (fpin, f);
+    fprintf (f, "end\n");
+  }
+#endif
   else
     mutt_copy_stream (fpin, f);
   fclose (fpin);
@@ -428,16 +534,18 @@ int mutt_write_mime_body (BODY *a, FILE *f)
 }
 
 #define BOUNDARYLEN 16
-char *mutt_generate_boundary (void)
+void mutt_generate_boundary (PARAMETER **parm)
 {
-  char *rs = (char *)safe_malloc (BOUNDARYLEN + 1);
+  char rs[BOUNDARYLEN + 1];
   char *p = rs;
   int i;
 
   rs[BOUNDARYLEN] = 0;
-  for (i=0;i<BOUNDARYLEN;i++) *p++ = B64Chars[LRAND() % sizeof (B64Chars)];
+  for (i=0;i<BOUNDARYLEN;i++) 
+    *p++ = B64Chars[LRAND() % sizeof (B64Chars)];
   *p = 0;
-  return (rs);
+  
+  mutt_set_parameter ("boundary", rs, parm);
 }
 
 /* analyze the contents of a file to determine which MIME encoding to use */
@@ -445,6 +553,7 @@ static CONTENT *mutt_get_content_info (const char *fname)
 {
   CONTENT *info;
   FILE *fp;
+  CHARSET_MAP *cm;
   int ch, from=0, whitespace=0, dot=0, linelen=0;
 
   if ((fp = fopen (fname, "r")) == NULL)
@@ -454,6 +563,12 @@ static CONTENT *mutt_get_content_info (const char *fname)
     return (NULL);
   }
 
+  {
+    CHARSET *cs;
+
+    cm = (cs = mutt_get_charset(Charset)) ? cs->map : 0;
+  }
+
   info = safe_calloc (1, sizeof (CONTENT));
   while ((ch = fgetc (fp)) != EOF)
   {
@@ -526,6 +641,8 @@ static CONTENT *mutt_get_content_info (const char *fname)
       if (ch == ' ') whitespace++;
       info->ascii++;
     }
+    if (cm && mutt_unicode_char (cm, ch) & -128)
+      info->nonasc = 1;
     if (linelen > 1) dot = 0;
     if (ch != ' ' && ch != '\t') whitespace = 0;
   }
@@ -551,9 +668,9 @@ static int lookup_mime_type (char *d, const char *s)
   *d = 0;
   cur_n = TYPEOTHER;
   cur_sze = 0;
-  szf = strlen (s);
+  szf = mutt_strlen (s);
 
-  for (count = 0 ; count < 2 ; count++)
+  for (count = 0 ; count < 3 ; count++)
   {
     /*
      * can't use strtok() because we use it in an inner loop below, so use
@@ -562,9 +679,12 @@ static int lookup_mime_type (char *d, const char *s)
     switch (count)
     {
       case 0:
-	snprintf (buf, sizeof (buf), "%s/.mime.types", Homedir);
+	snprintf (buf, sizeof (buf), "%s/.mime.types", NONULL(Homedir));
 	break;
       case 1:
+        strfcpy (buf, SYSCONFDIR"/mime.types", sizeof(buf));
+        break;
+      case 2:
 	strfcpy (buf, SHAREDIR"/mime.types", sizeof (buf));
 	break;
       default:
@@ -592,9 +712,9 @@ static int lookup_mime_type (char *d, const char *s)
 	/* cycle through the file extensions */
 	while ((p = strtok (p, " \t\n")))
 	{
-	  sze = strlen (p);
+	  sze = mutt_strlen (p);
 	  if ((sze > cur_sze) && (szf >= sze) &&
-	      strcasecmp (s + szf - sze, p) == 0 &&
+	      mutt_strcasecmp (s + szf - sze, p) == 0 &&
 	      (szf == sze || s[szf - sze - 1] == '.'))
 	  {
 	    char *dc;
@@ -627,22 +747,23 @@ static int lookup_mime_type (char *d, const char *s)
 
 static char *set_text_charset (CONTENT *info)
 {
-  if (strcasecmp (NONULL (Charset), "us-ascii") == 0)
-  {
-    if (info->hibin != 0)
-      return "unknown-8bit";
-  }
-  else if (info->hibin == 0)
-    return "us-ascii";
+  CHARSET *cs;
 
-  /* if no charset is given, provide a reasonable default */
-  return (Charset ? Charset : "us-ascii");
+  /* if charset is unknown assume low bytes are ascii compatible */
+
+  if ((Charset == NULL || mutt_strcasecmp (Charset, "us-ascii") == 0)
+      && info->hibin)
+    return ("unknown-8bit");
+
+  if (((cs = mutt_get_charset (Charset)) && cs->map) ? info->nonasc : info->hibin)
+    return (Charset);
+
+  return ("us-ascii");
 }
 
 void mutt_message_to_7bit (BODY *a, FILE *fp)
 {
   char temp[_POSIX_PATH_MAX];
-  size_t linelen = 0;
   char *line = NULL;
   FILE *fpin = NULL;
   FILE *fpout = NULL;
@@ -652,7 +773,7 @@ void mutt_message_to_7bit (BODY *a, FILE *fp)
     fpin = fp;
   else if (!a->filename || !(fpin = fopen (a->filename, "r")))
   {
-    mutt_error ("Could not open %s", a->filename ? a->filename : "(null)");
+    mutt_error (_("Could not open %s"), a->filename ? a->filename : "(null)");
     return;
   } 
   else
@@ -688,7 +809,6 @@ void mutt_message_to_7bit (BODY *a, FILE *fp)
   
  cleanup:
   safe_free ((void **) &line);
-  linelen = 0;
 
   if (fpin && !fp)
     fclose (fpin);
@@ -710,6 +830,7 @@ void mutt_message_to_7bit (BODY *a, FILE *fp)
   }
   a->length = sb.st_size;
   mutt_free_body (&a->parts);
+  a->hdr->content = NULL;
 }
 
 static void transform_to_7bit (BODY *a, FILE *fpin)
@@ -728,7 +849,7 @@ static void transform_to_7bit (BODY *a, FILE *fpin)
 
       transform_to_7bit (a->parts, fpin);
     } 
-    else if (a->type == TYPEMESSAGE && strcasecmp (a->subtype, "delivery-status"))
+    else if (mutt_is_message_type(a->type, a->subtype))
     {
       mutt_message_to_7bit (a, fpin);
     }
@@ -798,6 +919,11 @@ static void mutt_set_encoding (BODY *b, CONTENT *info)
     b->encoding = ENC7BIT;
 }
 
+void mutt_stamp_attachment(BODY *a)
+{
+  a->stamp = time(NULL);
+}
+
 /* Assumes called from send mode where BODY->filename points to actual file */
 void mutt_update_encoding (BODY *a)
 {
@@ -807,29 +933,108 @@ void mutt_update_encoding (BODY *a)
     return;
 
   mutt_set_encoding (a, info);
-
+  mutt_stamp_attachment(a);
+  
   if (a->type == TYPETEXT)
-  {
-    /* make sure the charset is valid */
-    if (a->parameter)
-      safe_free ((void **) &a->parameter->value);
-    else
-    {
-      a->parameter = mutt_new_parameter ();
-      a->parameter->attribute = safe_strdup ("charset");
-    }
-    a->parameter->value = safe_strdup (set_text_charset (info));
-  }
-
-
+    mutt_set_parameter ("charset", set_text_charset (info), &a->parameter);
 
+#ifdef _PGPPATH
+  /* save the info in case this message is signed.  we will want to do Q-P
+   * encoding if any lines begin with "From " so the signature won't be munged,
+   * for example.
+   */ 
+  safe_free ((void **) &a->content);
+  a->content = info;
+  info = NULL;
+#endif
 
 
 
   safe_free ((void **) &info);
 }
 
-BODY *mutt_make_attach (const char *path)
+BODY *mutt_make_message_attach (CONTEXT *ctx, HEADER *hdr, int attach_msg)
+{
+  char buffer[LONG_STRING];
+  BODY *body;
+  FILE *fp;
+  int cmflags, chflags;
+#ifdef _PGPPATH
+  int pgp = hdr->pgp;
+#endif
+
+#ifdef _PGPPATH
+  if ((option(OPTMIMEFORWDECODE) || option(OPTFORWDECRYPT)) &&
+      (hdr->pgp & PGPENCRYPT) && !pgp_valid_passphrase())
+    return (NULL);
+#endif /* _PGPPATH */
+
+  mutt_mktemp (buffer);
+  if ((fp = safe_fopen (buffer, "w+")) == NULL)
+    return NULL;
+
+  body = mutt_new_body ();
+  body->type = TYPEMESSAGE;
+  body->subtype = safe_strdup ("rfc822");
+  body->filename = safe_strdup (buffer);
+  body->unlink = 1;
+  body->use_disp = 0;
+
+  mutt_parse_mime_message (ctx, hdr);
+
+  chflags = CH_XMIT;
+  cmflags = 0;
+
+  /* If we are attaching a message, ignore OPTMIMEFORWDECODE */
+  if (!attach_msg && option (OPTMIMEFORWDECODE))
+  {
+    chflags |= CH_MIME | CH_TXTPLAIN;
+    cmflags = M_CM_DECODE;
+#ifdef _PGPPATH
+    pgp &= ~PGPENCRYPT;
+#endif
+  }
+#ifdef _PGPPATH
+  else
+    if(option(OPTFORWDECRYPT)
+       && (hdr->pgp & PGPENCRYPT))
+  {
+    if(mutt_is_multipart_encrypted(hdr->content))
+    {
+      chflags |= CH_MIME | CH_NONEWLINE;
+      cmflags = M_CM_DECODE_PGP;
+      pgp &= ~PGPENCRYPT;
+    }
+    else if(mutt_is_application_pgp(hdr->content) & PGPENCRYPT)
+    {
+      chflags |= CH_MIME | CH_TXTPLAIN;
+      cmflags = M_CM_DECODE;
+      pgp &= ~PGPENCRYPT;
+    }
+  }
+#endif
+
+  mutt_copy_message (fp, ctx, hdr, cmflags, chflags);
+  
+  fflush(fp);
+  rewind(fp);
+
+  body->hdr = mutt_new_header();
+  body->hdr->offset = 0;
+  /* we don't need the user headers here */
+  body->hdr->env = mutt_read_rfc822_header(fp, body->hdr, 0);
+#ifdef _PGPPATH
+  body->hdr->pgp = pgp;
+#endif /* _PGPPATH */
+  mutt_update_encoding (body);
+  body->parts = body->hdr->content;
+
+  fclose(fp);
+  
+  return (body);
+}
+
+BODY *mutt_make_file_attach (const char *path)
 {
   BODY *att;
   CONTENT *info;
@@ -861,9 +1066,8 @@ BODY *mutt_make_attach (const char *path)
        */
       att->type = TYPETEXT;
       att->subtype = safe_strdup ("plain");
-      att->parameter = mutt_new_parameter ();
-      att->parameter->attribute = safe_strdup ("charset");
-      att->parameter->value = safe_strdup (set_text_charset (info));
+      
+      mutt_set_parameter("charset", set_text_charset(info), &att->parameter);
     }
     else
     {
@@ -873,9 +1077,17 @@ BODY *mutt_make_attach (const char *path)
   } 
 
   mutt_set_encoding (att, info);
+  mutt_stamp_attachment(att);
 
-
-
+#ifdef _PGPPATH
+  /*
+   * save the info in case this message is signed.  we will want to do Q-P
+   * encoding if any lines begin with "From " so the signature won't be munged,
+   * for example.
+   */
+  att->content = info;
+  info = NULL;
+#endif
 
 
 
@@ -907,32 +1119,38 @@ BODY *mutt_make_multipart (BODY *b)
   new->type = TYPEMULTIPART;
   new->subtype = safe_strdup ("mixed");
   new->encoding = get_toplevel_encoding (b);
-  new->parameter = mutt_new_parameter ();
-  new->parameter->attribute = safe_strdup ("boundary");
-  new->parameter->value = mutt_generate_boundary ();
+  mutt_generate_boundary(&new->parameter);
   new->use_disp = 0;  
   new->parts = b;
 
   return new;
 }
 
-static char *mutt_make_date (char *s)
+/* remove the multipart body if it exists */
+BODY *mutt_remove_multipart (BODY *b)
+{
+  BODY *t;
+
+  if (b->parts)
+  {
+    t = b;
+    b = b->parts;
+    t->parts = NULL;
+    mutt_free_body (&t);
+  }
+  return b;
+}
+
+char *mutt_make_date (char *s, size_t len)
 {
   time_t t = time (NULL);
-  struct tm *l = gmtime(&t);
-  int yday = l->tm_yday;
-  int tz = l->tm_hour * 60 + l->tm_min;
+  struct tm *l = localtime (&t);
+  time_t tz = mutt_local_tz (t);
 
-  l = localtime(&t);
-  tz = l->tm_hour * 60 + l->tm_min - tz;
-  yday = l->tm_yday - yday;
-
-  if (yday != 0)
-    tz += yday * 24 * 60; /* GMT is next or previous day! */
-
-  sprintf (s, "Date: %s, %d %s %d %02d:%02d:%02d %+03d%02d\n",
-	   Weekdays[l->tm_wday], l->tm_mday, Months[l->tm_mon], l->tm_year+1900,
-	   l->tm_hour, l->tm_min, l->tm_sec, tz/60, abs(tz) % 60);
+  snprintf (s, len,  "Date: %s, %d %s %d %02d:%02d:%02d %+03d%02d\n",
+	    Weekdays[l->tm_wday], l->tm_mday, Months[l->tm_mon],
+	    l->tm_year + 1900, l->tm_hour, l->tm_min, l->tm_sec,
+	    tz / 3600, abs (tz) % 3600);
   return (s);
 }
 
@@ -951,7 +1169,7 @@ void mutt_write_address_list (ADDRESS *adr, FILE *fp, int linelen)
     adr->next = NULL;
     buf[0] = 0;
     rfc822_write_address (buf, sizeof (buf), adr);
-    len = strlen (buf);
+    len = mutt_strlen (buf);
     if (count && linelen + len > 74)
     {
       if (count)
@@ -1025,12 +1243,17 @@ int mutt_write_rfc822_header (FILE *fp, ENVELOPE *env, BODY *attach, int mode)
   char buffer[LONG_STRING];
   LIST *tmp = env->userhdrs;
 
-  if (mode == 0)
+  if (option(OPTUSEHEADERDATE))
   {
-    if (env->message_id)
-      fprintf (fp, "Message-ID: %s\n", env->message_id);
-    fputs (mutt_make_date (buffer), fp);
+    if(env->date)
+      fprintf(fp, "Date: %s\n", env->date);
+    else
+      fputs (mutt_make_date(buffer, sizeof(buffer)), fp);
   }
+  else if (mode == 0)
+    fputs (mutt_make_date (buffer, sizeof(buffer)), fp);
+
+
 
   /* OPTUSEFROM is not consulted here so that we can still write a From:
    * field if the user sets it with the `my_hdr' command
@@ -1060,8 +1283,11 @@ int mutt_write_rfc822_header (FILE *fp, ENVELOPE *env, BODY *attach, int mode)
 
   if (env->bcc)
   {
-    fputs ("Bcc: ", fp);
-    mutt_write_address_list (env->bcc, fp, 5);
+    if(mode != 0 || option(OPTWRITEBCC))
+    {
+      fputs ("Bcc: ", fp);
+      mutt_write_address_list (env->bcc, fp, 5);
+    }
   }
   else if (mode > 0)
     fputs ("Bcc: \n", fp);
@@ -1071,6 +1297,10 @@ int mutt_write_rfc822_header (FILE *fp, ENVELOPE *env, BODY *attach, int mode)
   else if (mode == 1)
     fputs ("Subject: \n", fp);
 
+  /* save message id if the user has set it */
+  if (env->message_id)
+    fprintf (fp, "Message-ID: %s\n", env->message_id);
+
   if (env->reply_to)
   {
     fputs ("Reply-To: ", fp);
@@ -1103,7 +1333,7 @@ int mutt_write_rfc822_header (FILE *fp, ENVELOPE *env, BODY *attach, int mode)
   if (mode == 0)
   {
     /* Add a vanity header */
-    fprintf (fp, "X-Mailer: Mutt %s\n", VERSION);
+    fprintf (fp, "X-Mailer: Mutt %s\n", MUTT_VERSION);
   }
 #endif
 
@@ -1130,7 +1360,7 @@ static void encode_headers (LIST *h)
       *p++ = 0;
       SKIPWS (p);
       snprintf (tmp, sizeof (tmp), "%s: ", h->data);
-      len = strlen (tmp);
+      len = mutt_strlen (tmp);
       rfc2047_encode_string (tmp + len, sizeof (tmp) - len, (unsigned char *) p);
       safe_free ((void **) &h->data);
       h->data = safe_strdup (tmp);
@@ -1138,23 +1368,29 @@ static void encode_headers (LIST *h)
   }
 }
 
-/* rfc2047 encode the content-descriptions */
-static void encode_descriptions (BODY *b)
+const char *mutt_fqdn(short may_hide_host)
 {
-  BODY *t;
-  char tmp[LONG_STRING];
-
-  for (t = b; t; t = t->next)
+  char *p = NULL, *q;
+  
+  if(Fqdn && Fqdn[0] != '@')
   {
-    if (t->description)
+    p = Fqdn;
+    
+    if(may_hide_host && option(OPTHIDDENHOST))
     {
-      rfc2047_encode_string (tmp, sizeof (tmp), (unsigned char *) t->description);
-      safe_free ((void **) &t->description);
-      t->description = safe_strdup (tmp);
+      if((p = strchr(Fqdn, '.')))
+	p++;
+
+      /* sanity check: don't hide the host if
+       * the fqdn is something like detebe.org.
+       */
+      
+      if(!p || !(q = strchr(p, '.')))
+	p = Fqdn;
     }
-    if (t->parts)
-      encode_descriptions (t->parts);
   }
+
+  return p;
 }
 
 char *mutt_gen_msgid (void)
@@ -1162,13 +1398,16 @@ char *mutt_gen_msgid (void)
   char buf[SHORT_STRING];
   time_t now;
   struct tm *tm;
+  const char *fqdn;
 
   now = time (NULL);
   tm = localtime (&now);
+  if(!(fqdn = mutt_fqdn(0)))
+    fqdn = NONULL(Hostname);
+
   snprintf (buf, sizeof (buf), "<%d%02d%02d%02d%02d%02d.%c%d@%s>",
 	    tm->tm_year + 1900, tm->tm_mon + 1, tm->tm_mday, tm->tm_hour,
-	    tm->tm_min, tm->tm_sec, MsgIdPfx, getpid (),
-	    Fqdn[0] != '@' ? Fqdn : Hostname);
+	    tm->tm_min, tm->tm_sec, MsgIdPfx, getpid (), fqdn);
   MsgIdPfx = (MsgIdPfx == 'Z') ? 'A' : MsgIdPfx + 1;
   return (safe_strdup (buf));
 }
@@ -1365,8 +1604,23 @@ add_option (char **args, size_t *argslen, size_t *argsmax, char *s)
   return (args);
 }
 
-static int
-invoke_sendmail (ADDRESS *to, ADDRESS *cc, ADDRESS *bcc, /* recips */
+static const char *
+strsysexit(int e)
+{
+  int i;
+  
+  for(i = 0; sysexits_h[i].str; i++)
+  {
+    if(e == sysexits_h[i].v)
+      break;
+  }
+  
+  return sysexits_h[i].str;
+}
+
+
+int
+mutt_invoke_sendmail (ADDRESS *to, ADDRESS *cc, ADDRESS *bcc, /* recips */
 		 const char *msg, /* file containing message */
 		 int eightbit) /* message contains 8bit chars */
 {
@@ -1409,6 +1663,7 @@ invoke_sendmail (ADDRESS *to, ADDRESS *cc, ADDRESS *bcc, /* recips */
     args = add_option (args, &argslen, &argsmax, "-R");
     args = add_option (args, &argslen, &argsmax, DsnReturn);
   }
+  args = add_option (args, &argslen, &argsmax, "--");
   args = add_args (args, &argslen, &argsmax, to);
   args = add_args (args, &argslen, &argsmax, cc);
   args = add_args (args, &argslen, &argsmax, bcc);
@@ -1422,15 +1677,15 @@ invoke_sendmail (ADDRESS *to, ADDRESS *cc, ADDRESS *bcc, /* recips */
     endwin ();
   if ((i = send_msg (path, args, msg, &childout)) != (EX_OK & 0xff))
   {
-    char *e = strerror (errno);
+    const char *e = strsysexit(i);
 
-    fprintf (stderr, "Error sending message, child exited %d (%s).\n", i, NONULL (e));
+    fprintf (stderr, _("Error sending message, child exited %d (%s).\n"), i, NONULL (e));
     if (childout)
-      fprintf (stderr, "Saved output of child process to %s.\n", childout);
+      fprintf (stderr, _("Saved output of child process to %s.\n"), childout);
     if (!option (OPTNOCURSES))
     {
       mutt_any_key_to_continue (NULL);
-      mutt_error ("Error sending message.");
+      mutt_error _("Error sending message.");
     }
   }
   FREE (&childout);
@@ -1445,8 +1700,8 @@ invoke_sendmail (ADDRESS *to, ADDRESS *cc, ADDRESS *bcc, /* recips */
    string. */
 char *mutt_append_string (char *a, const char *b)
 {
-  size_t la = strlen (a);
-  safe_realloc ((void **) &a, la + strlen (b) + 1);
+  size_t la = mutt_strlen (a);
+  safe_realloc ((void **) &a, la + mutt_strlen (b) + 1);
   strcpy (a + la, b);
   return (a);
 }
@@ -1475,8 +1730,8 @@ char *mutt_quote_string (const char *s)
   char *r, *pr;
   size_t rlen;
 
-  rlen = strlen (s) + 3;
-  pr = r = malloc (rlen);
+  rlen = mutt_strlen (s) + 3;
+  pr = r = (char *) safe_malloc (rlen);
   *pr++ = '"';
   while (*s)
   {
@@ -1494,68 +1749,49 @@ char *mutt_quote_string (const char *s)
   return (r);
 }
 
-int mutt_send_message (HEADER *msg, const char *fcc)
+void mutt_prepare_envelope (ENVELOPE *env)
 {
-  char tempfile[_POSIX_PATH_MAX], buffer[STRING];
-  FILE *tempfp;
-  int i;
+  char buffer[LONG_STRING];
+
+  if (env->bcc && !(env->to || env->cc))
+  {
+    /* some MTA's will put an Apparently-To: header field showing the Bcc:
+     * recipients if there is no To: or Cc: field, so attempt to suppress
+     * it by using an empty To: field.
+     */
+    env->to = rfc822_new_address ();
+    env->to->group = 1;
+    env->to->next = rfc822_new_address ();
+
+    buffer[0] = 0;
+    rfc822_cat (buffer, sizeof (buffer), "undisclosed-recipients",
+		RFC822Specials);
+
+    env->to->mailbox = safe_strdup (buffer);
+  }
+
+  mutt_set_followup_to (env);
 
   /* Take care of 8-bit => 7-bit conversion. */
-  rfc2047_encode_adrlist (msg->env->to);
-  rfc2047_encode_adrlist (msg->env->cc);
-  rfc2047_encode_adrlist (msg->env->from);
-  rfc2047_encode_adrlist (msg->env->mail_followup_to);
-  if (msg->env->subject)
+  rfc2047_encode_adrlist (env->to);
+  rfc2047_encode_adrlist (env->cc);
+  rfc2047_encode_adrlist (env->from);
+  rfc2047_encode_adrlist (env->mail_followup_to);
+  rfc2047_encode_adrlist (env->reply_to);
+
+  if (env->subject)
   {
     rfc2047_encode_string (buffer, sizeof (buffer) - 1,
-			   (unsigned char *) msg->env->subject);
-    safe_free ((void **) &msg->env->subject);
-    msg->env->subject = safe_strdup (buffer);
+			   (unsigned char *) env->subject);
+    safe_free ((void **) &env->subject);
+    env->subject = safe_strdup (buffer);
   }
-  encode_headers (msg->env->userhdrs);
-  encode_descriptions (msg->content);
+  encode_headers (env->userhdrs);
 
-  if (!msg->env->message_id)
-    msg->env->message_id = mutt_gen_msgid ();
-
-  /* Write out the message in MIME form. */
-  mutt_mktemp (tempfile);
-  if ((tempfp = safe_fopen (tempfile, "w")) == NULL)
-    return (-1);
-
-  mutt_write_rfc822_header (tempfp, msg->env, msg->content, 0);
-  fputc ('\n', tempfp); /* tie off the header. */
-
-  mutt_write_mime_body (msg->content, tempfp);
-  if (fclose (tempfp) != 0)
-  {
-    mutt_perror (tempfile);
-    unlink (tempfile);
-    return (-1);
-  }
-
-  /* save a copy of the message, if necessary. */
-  if (*fcc && strcmp ("/dev/null", fcc) != 0)
-  {
-    BODY *tmpbody = msg->content;
-
-    /* check to see if the user wants copies of all attachments */
-    if (msg->content->type == TYPEMULTIPART &&
-	strcmp (msg->content->subtype, "encrypted") &&
-	strcmp (msg->content->subtype, "signed") &&
-	!option (OPTFCCATTACH))
-      msg->content = msg->content->parts;
-
-    mutt_write_fcc (fcc, msg, NULL, 0);
-    msg->content = tmpbody;
-  }
-
-  i = invoke_sendmail (msg->env->to, msg->env->cc, msg->env->bcc,
-		       tempfile,
-		       (msg->content->encoding == ENC8BIT));
-  return (i ? -1 : 0);
+  if (!env->message_id)
+    env->message_id = mutt_gen_msgid ();
 }
-
+  
 void mutt_bounce_message (HEADER *h, ADDRESS *to)
 {
   int i;
@@ -1576,19 +1812,21 @@ void mutt_bounce_message (HEADER *h, ADDRESS *to)
     mutt_mktemp (tempfile);
     if ((f = safe_fopen (tempfile, "w")) != NULL)
     {
+      const char *fqdn;
+
       fseek (msg->fp, h->offset, 0);
       mutt_copy_header (msg->fp, h, f, CH_XMIT | CH_NONEWLINE, NULL);
-      fprintf (f, "Resent-From: %s", Username);
-      if (Fqdn[0] != '@')
-	fprintf (f, "@%s", Fqdn);
-      fprintf (f, "\nResent-%s", mutt_make_date (date));
+      fprintf (f, "Resent-From: %s", NONULL(Username));
+      if((fqdn = mutt_fqdn(1)))
+	fprintf (f, "@%s", fqdn);
+      fprintf (f, "\nResent-%s", mutt_make_date (date, sizeof(date)));
       fputs ("Resent-To: ", f);
       mutt_write_address_list (to, f, 11);
       fputc ('\n', f);
       mutt_copy_bytes (msg->fp, f, h->content->length);
       fclose (f);
 
-      invoke_sendmail (to, NULL, NULL, tempfile, h->content->encoding == ENC8BIT);
+      mutt_invoke_sendmail (to, NULL, NULL, tempfile, h->content->encoding == ENC8BIT);
     }
     mx_close_message (&msg);
   }
@@ -1609,7 +1847,7 @@ ADDRESS *mutt_remove_duplicates (ADDRESS *addr)
     tmp = top;
     do {
       if (addr->mailbox && tmp->mailbox &&
-	  !strcasecmp (addr->mailbox, tmp->mailbox))
+	  !mutt_strcasecmp (addr->mailbox, tmp->mailbox))
       {
 	/* duplicate address, just ignore it */
 	tmp = addr;
@@ -1634,7 +1872,7 @@ ADDRESS *mutt_remove_duplicates (ADDRESS *addr)
   return (top);
 }
 
-int mutt_write_fcc (const char *path, HEADER *hdr, const char *msgid, int post)
+int mutt_write_fcc (const char *path, HEADER *hdr, const char *msgid, int post, char *fcc)
 {
   CONTEXT f;
   MESSAGE *msg;
@@ -1652,7 +1890,7 @@ int mutt_write_fcc (const char *path, HEADER *hdr, const char *msgid, int post)
   /* We need to add a Content-Length field to avoid problems where a line in
    * the message body begins with "From "   
    */
-  if (f.magic == M_MMDF || f.magic == M_MBOX)
+  if (f.magic == M_MMDF || f.magic == M_MBOX || f.magic == M_KENDRA)
   {
     mutt_mktemp (tempfile);
     if ((tempfp = safe_fopen (tempfile, "w+")) == NULL)
@@ -1663,14 +1901,17 @@ int mutt_write_fcc (const char *path, HEADER *hdr, const char *msgid, int post)
     }
   }
 
-  hdr->read = 1; /* make sure to put it in the `cur' directory (maildir) */
+  hdr->read = !post; /* make sure to put it in the `cur' directory (maildir) */
   if ((msg = mx_open_new_message (&f, hdr, M_ADD_FROM)) == NULL)
   {
     mx_close_mailbox (&f);
     return (-1);
   }
 
-  mutt_write_rfc822_header (msg->fp, hdr->env, hdr->content, (post ? -1 : 0));
+  /* post == 1 => postpone message. Set mode = -1 in mutt_write_rfc822_header()
+   * post == 0 => Normal mode. Set mode = 0 in mutt_write_rfc822_header() 
+   * */
+  mutt_write_rfc822_header (msg->fp, hdr->env, hdr->content, post ? -post : 0);
 
   /* (postponment) if this was a reply of some sort, <msgid> contians the
    * Message-ID: of message replied to.  Save it using a special X-Mutt-
@@ -1680,10 +1921,34 @@ int mutt_write_fcc (const char *path, HEADER *hdr, const char *msgid, int post)
    */
   if (post && msgid)
     fprintf (msg->fp, "X-Mutt-References: %s\n", msgid);
+  
+  /* (postponment) save the Fcc: using a special X-Mutt- header so that
+   * it can be picked up when the message is recalled 
+   */
+  if (post && fcc)
+    fprintf (msg->fp, "X-Mutt-Fcc: %s\n", fcc);
   fprintf (msg->fp, "Status: RO\n");
 
 
 
+#ifdef _PGPPATH
+  /* (postponment) if the mail is to be signed or encrypted, save this info */
+  if (post && (hdr->pgp & (PGPENCRYPT | PGPSIGN)))
+  {
+    fputs ("Pgp: ", msg->fp);
+    if (hdr->pgp & PGPENCRYPT) 
+      fputc ('E', msg->fp);
+    if (hdr->pgp & PGPSIGN)
+    {
+      fputc ('S', msg->fp);
+      if (PgpSignAs && *PgpSignAs)
+        fprintf (msg->fp, "<%s>", PgpSignAs);
+      if (*PgpSignMicalg)
+	fprintf (msg->fp, "M<%s>", PgpSignMicalg);
+    }
+    fputc ('\n', msg->fp);
+  }
+#endif /* _PGPPATH */
 
 
 
