@@ -78,7 +78,7 @@ int compare_subject (const void *a, const void *b)
   else if (!(*pb)->env->real_subj)
     rc = 1;
   else
-    rc = strcasecmp ((*pa)->env->real_subj, (*pb)->env->real_subj);
+    rc = mutt_strcasecmp ((*pa)->env->real_subj, (*pb)->env->real_subj);
   AUXSORT(rc,a,b);
   return (SORTCODE (rc));
 }
@@ -109,7 +109,7 @@ int compare_to (const void *a, const void *b)
 
   fa = mutt_get_name ((*ppa)->env->to);
   fb = mutt_get_name ((*ppb)->env->to);
-  result = strcasecmp (fa, fb);
+  result = mutt_strcasecmp (fa, fb);
   AUXSORT(result,a,b);
   return (SORTCODE (result));
 }
@@ -123,7 +123,7 @@ int compare_from (const void *a, const void *b)
 
   fa = mutt_get_name ((*ppa)->env->from);
   fb = mutt_get_name ((*ppb)->env->from);
-  result = strcasecmp (fa, fb);
+  result = mutt_strcasecmp (fa, fb);
   AUXSORT(result,a,b);
   return (SORTCODE (result));
 }
@@ -175,6 +175,7 @@ sort_t *mutt_get_sort_func (int method)
 void mutt_sort_headers (CONTEXT *ctx, int init)
 {
   int i;
+  HEADER *h;
   sort_t *sortfunc;
   
   unset_option (OPTNEEDRESORT);
@@ -194,7 +195,7 @@ void mutt_sort_headers (CONTEXT *ctx, int init)
   }
 
   if (!ctx->quiet)
-    mutt_message ("Sorting mailbox...");
+    mutt_message _("Sorting mailbox...");
 
   /* threads may be bogus, so clear the links */
   if (init)
@@ -222,7 +223,7 @@ void mutt_sort_headers (CONTEXT *ctx, int init)
   else if ((sortfunc = mutt_get_sort_func (Sort)) == NULL ||
 	   (AuxSort = mutt_get_sort_func (SortAux)) == NULL)
   {
-    mutt_error ("Could not find sorting function! [report this bug]");
+    mutt_error _("Could not find sorting function! [report this bug]");
     sleep (1);
     return;
   }
@@ -238,13 +239,27 @@ void mutt_sort_headers (CONTEXT *ctx, int init)
   ctx->vcount = 0;
   for (i = 0; i < ctx->msgcount; i++)
   {
-    if (ctx->hdrs[i]->virtual != -1)
+    HEADER *cur = ctx->hdrs[i];
+    if (cur->virtual != -1 || (cur->collapsed && (!ctx->pattern || cur->limited)))
     {
-      ctx->hdrs[i]->virtual = ctx->vcount;
+      cur->virtual = ctx->vcount;
       ctx->v2r[ctx->vcount] = i;
       ctx->vcount++;
     }
-    ctx->hdrs[i]->msgno = i;
+    cur->msgno = i;
+  }
+
+  /* re-collapse threads marked as collapsed */
+  if ((Sort & SORT_MASK) == SORT_THREADS)
+  {
+    h = ctx->tree;
+    while (h)
+    {
+      if (h->collapsed)
+	mutt_collapse_thread (ctx, h);
+      h = h->next;
+    }
+    mutt_set_virtual (ctx);
   }
 
   if (!ctx->quiet)
